unit CheckPgopher;
{ Read a pgopher output file, and compare to a calculation of the same thing }

{$I DFS.inc}

interface

uses
  Classes, SysUtils, Mixture, TextStrm;

function CheckPgopherFile(Generated: Boolean; Mixture: TMixture; S: TFreeFormatStream; var nTried: Integer; Tolerance: Double; var Failures: Integer):Double;
function CheckEigenvalueFile(Mixture: TMixture; S: TFreeFormatStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;

implementation

uses MathSup, MatrixVec;

function CheckEigenvalueFile(Mixture: TMixture; S: TFreeFormatStream; var nTried: Integer; Tolerance: Double; var Failures: Integer): Double;
var
  J, sym, n: Integer; { State Details }
  E: Double;
  Manifold: TManifold;
  Molecule: TMolecule;
  Eigenvalues: TEigenvalues;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn('Line ', S.Line, ':', Name);
      WriteLn(J:5, sym:2, n:2, E:12:4);
    end;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: Read=%g Calc=%g %g', [Name, n1, n2, n1-n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
  function FindMolecule(AName: string): TMolecule;
  var
    i, j: Integer;
  begin
    Result := nil;
    for i := 0 to Mixture.Count-1 do
      with Mixture.Species[i] do
        for j := 0 to Count-1 do
          if Isotopomer[i].Name = AName then begin
            Result := Isotopomer[i];
            Exit;
          end;
  end;
begin
  nTried := 0;
  Molecule := Mixture.Species[0].Isotopomer[0];
  Manifold := Molecule.Manifold[0];
  Result := 0;
  while not S.AtEndOfFile do begin
    S.NextIntegerVar(J);  S.NextIntegerVar(sym);  S.NextIntegerVar(n); S.NextDoubleVar(E);
    Inc(nTried);
    Eigenvalues := Manifold.Eigenvalues(J, sym, [meHrot, meWidth]);
    if n > Eigenvalues.Size then
      Fail(Format('Eigenvalue (%d) > %d', [n, Eigenvalues.Size]))
    else
      CheckNo(E, Eigenvalues.E[n], Tolerance, 'E');
  end;
  S.NextLine;
end;

function CheckPgopherFile(Generated: Boolean; Mixture: TMixture; S: TFreeFormatStream; var nTried: Integer; Tolerance: Double; var Failures: Integer):Double;
var
  MoleculeName: string;
  Manp: string;
  Jp, symp, np: Integer; { Upper State }
  Man: string;
  J,  sym,  n : Integer; { Lower State }
  f, str, cstr, linestrength, w, Ep, E, Sum: Double;
  ManifoldL, ManifoldU: TManifold;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  i: Integer;
  D: TMatrix;
  Line: TLine;
  Failed: Boolean;
  TxDetails: string;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn('Line ', S.Line, ':', Name);
      WriteLn(MoleculeName, ' ', Manp, ' ', Jp:3, symp:2, np:2, J:5, sym:2, n:2, f:12:4, str:8:4, w:8:3, Ep:12:4, E:12:4);
    end;
    Failed := True;
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else if Abs(n1-n2) > tol then
      Fail(Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2]));
    if Abs(n1-n2) > Result then
      Result := Abs(n1-n2);
  end;
  function FindMolecule(AName: string): TMolecule;
  var
    i, j: Integer;
  begin
    Result := nil;
    for i := 0 to Mixture.Count-1 do
      with Mixture.Species[i] do
        for j := 0 to Count-1 do
          if Isotopomer[i].Name = AName then begin
            Result := Isotopomer[i];
            Exit;
          end;
  end;
begin
  nTried := 0;
  Result := 0;
  while not S.AtEndOfFile and (Pos('Frequency', S.RestOfLine) <= 0) do
    S.NextLine;
  S.NextLine;
  i := -1;
  Sum := 0;
  while not S.AtEndOfFile do begin
    Failed := False;
    S.NextStringVar(MoleculeName);
    if CompareText(MoleculeName, 'stop') = 0 then begin
      S.NextLine;
      Continue;
    end;
    if TryStrToInt(MoleculeName,i) then
      Break;
    S.NextStringVar(Manp); S.NextIntegerVar(Jp); S.NextIntegerVar(symp); S.NextIntegerVar(np);
    S.NextStringVar(Man); S.NextIntegerVar(J);  S.NextIntegerVar(sym);  S.NextIntegerVar(n);
    S.NextDoubleVar(f);
    if Generated then begin
      S.NextDoubleVar(w); { Actually weight }
      S.NextDoubleVar(str);
      S.NextLine;
      if w <= 0 then
        Continue; { Skip }
    end else begin
      S.NextDoubleVar(str);  S.NextDoubleVar(Ep);    S.NextDoubleVar(E);
      S.NextDoubleVar(linestrength); S.NextDoubleVar(w);
      TxDetails := Trim(S.RestOfLine);
      if (Length(TxDetails) >0) and (TxDetails[1] = ':') then
        TxDetails := Trim(Copy(TxDetails,2,Length(TxDetails)));
      { May need to do more trimming if format changes in future }
    end;
    Inc(nTried);
    Molecule := FindMolecule(MoleculeName);
    if not Assigned(Molecule) then Fail('Can''t find Molecule:'+MoleculeName);
    if not Failed then begin
      Line.Width := 0;
      ManifoldU := Molecule.ManifoldWithName(Manp);
      if not Assigned(ManifoldU) then
        Fail('Can''t find Upper Manifold:'+Manp)
      else begin
        Line.EigenvaluesU := ManifoldU.Eigenvalues(Jp, symp, [meHrot, meWidth]);
        Line.IndexU := np;
        if Line.IndexU > Line.EigenvaluesU.Size then
          Fail(Format('Upper eigenvalue (%d) > %d', [np, Line.EigenvaluesU.Size]))
        else begin
          if not Generated then
            CheckNo(Ep, Line.EigenvaluesU.E[Line.IndexU], Tolerance, 'E''');
          if Assigned(Line.EigenValuesU.Widths) then
            Line.Width := Line.Width + Line.EigenValuesU.Widths[np];
        end;
      end;
      ManifoldL := Molecule.ManifoldWithName(Man);
      if not Assigned(ManifoldL) then
        Fail('Can''t find Lower Manifold:'+Man)
      else begin
        Line.EigenvaluesL := ManifoldL.Eigenvalues(J, sym, [meHrot, meWidth]);
        Line.IndexL := n;
        if Line.IndexL > Line.EigenvaluesL.Size then
          Fail(Format('Lower eigenvalue (%d) > %d', [n, Line.EigenvaluesL.Size]))
        else begin
          if Assigned(Line.EigenValuesL.Widths) then
            Line.Width := Line.Width + Line.EigenValuesL.Widths[n];
          if not Generated then
            CheckNo(E, Line.EigenvaluesL.E[Line.IndexL], Tolerance, 'E"');
        end;
      end;
      if not failed then begin
        Line.Frequency := Line.EigenvaluesU.E[Line.IndexU]-Line.EigenvaluesL.E[Line.IndexL];
        CheckNo(f, Line.Frequency, Tolerance, 'Frequency');
        TransitionMoments := Molecule.FindTransitionMoments(ManifoldU,ManifoldL);
        if not Assigned(TransitionMoments) then
          Fail('Can''t find transition moment')
        else begin
          D := TransitionMoments.TransitionMatrix(Line.EigenValuesL, Line.EigenValuesU, nil);
          try
            Line.Strength := D[Line.IndexL, Line.IndexU]*Line.EigenvaluesL.StatisticalWeight;
            if not Generated then
              CheckNo(linestrength, Line.Strength, Tolerance, 'Line Strength');
            cstr := Mixture.LineIntensity(Line, Mixture.Temperature, Mixture.Tvib);
            CheckNo(str, cstr, Tolerance, 'Intensity');
            Sum := Sum + cstr;
            if not Generated then begin
              CheckNo(w, Line.Width, Tolerance, 'Width');
              Line.CachedName := '';
              if (TxDetails <> '') and (MakeTxName(Line) <> TxDetails) then begin
                { Format Adjust }
                // Line.CachedName := AnsiReplaceStr(Line.CachedName,'( ','(');
                // if Line.CachedName <> TxDetails then
                  Fail('TxDetails mismatch|'+Line.CachedName+'|'+TxDetails+'| ');
              end;
            end;
          finally
            D.Free;
          end;
          // WriteLn(Line.Strength:8:4, LineIntensity(Line, Mixture.Temperature, Mixture.Tvib)/5:8:4, str:8:4);
        end;
      end;
    end;
    S.NextLine;
  end;
  if i <> -1 then begin
    CheckNo(i, nTried, Tolerance, 'Line Count');
    while not S.AtEndOfFile and (S.NextString('') <> '=') do
      ;
    CheckNo(S.NextDouble(Sum), Sum, 1e-4, 'Line intensity sum');
  end;
end;

end.

