unit CommandLine;
{$I DFS.inc}
{
PGOPHER - a program for simulating rotational structure
Copyright (C) 2005 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
interface

uses
  Classes, SysUtils; 

function HandleCommandLine: Boolean;

implementation

uses {$IFDEF CONSOLE} Profile, {$ELSE} Main, LogUnit, Forms, OvrlyFrm, {$ENDIF}
     {$IFDEF VCL_} StackDump, {$ENDIF}
     Mixture, D3hImport, PgopherImport, CheckPgopher, CheckOverlay,
     TextStrm, Overlay, MNode, JPLCat;

procedure CreateLogForms;
begin
  {$IFNDEF CONSOLE}
  {$IFNDEF MSWINDOWS}
  Application.CreateForm(TLogForm, LogForm); { Only divert output after testing }
  if not Assigned(OverlayForm) then
    Application.CreateForm(TOverlayForm, OverlayForm);
  {$ENDIF}
  {$ENDIF}
end;

function HandleCommandLine: Boolean;
var
  Stream: TFreeFormatStream;
  FStream: TFileStream;
  i, d, nTried: Integer;
  Params, MixtureChanges: array of string;
  Tolerance, LargestError, Offset: Double;
  Mixture: TMixture;
  Overlays: TOverlays;
  nFailures: Integer;
  SortBoth, Interpolate, CvtToPert: Boolean;
  FileName: string;
  {$IFNDEF VCL_}
  VersionInfo: string;
  {$ENDIF}
  procedure ApplyMixtureChanges;
  var
    i: Integer;
  begin
    for i := 0 to High(MixtureChanges) do
      Mixture.HandleAssignment(MixtureChanges[i]);
  end;
begin
  Stream := nil;
  FStream := nil;
  {$IFDEF CONSOLE}
  VersionInfo := {$I commandlinever.inc} ;
  {$ELSE}
    with MainForm.Project do
      VersionInfo := 'Pgopher ' + Version + '.' + IntToStr(Build) + ' ' + BuildDate;
  {$ENDIF}
  {$IFDEF VCL_}
  ExpectedException := EReportException;
  {$ENDIF}
  if (ParamCount >= 1) {$IFNDEF CONSOLE} and (Copy(ParamStr(1),1,2) = '--') {$ENDIF} then begin
   try
    Result := False;
    Tolerance := 1e-4;
    nFailures := 0;
    SortBoth := False;
    Interpolate := False;
    CvtToPert := False;
    Offset := 0;
    SetLength(Params, ParamCount);
    d := 0;
    i := 1;
    while i <= ParamCount do begin
      Params[d] := ParamStr(i);
      if Copy(Paramstr(i), 1, 8) = 'Mixture.' then begin
        SetLength(MixtureChanges, Length(MixtureChanges)+1);
        MixtureChanges[High(MixtureChanges)] := Copy(Paramstr(i), 9, Length(Paramstr(i)));
      end else if ParamStr(i) = '--tolerance' then begin
        Tolerance := StrToFloat(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--fail' then begin
        nFailures := -StrToInt(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--sort' then
        SortBoth := True
      else if ParamStr(i) = '--offset' then begin
        Offset := StrToFloat(ParamStr(i+1));
        Inc(i);
      end else if ParamStr(i) = '--interpolate' then
        Interpolate := True
      else if ParamStr(i) = '--cvttopert' then
        CvtToPert := True
      else
        Inc(d);
      Inc(i);
    end;
    SetLength(Params, d);
    if d = 3 then begin
      { Automated testing }
      {$IFNDEF CONSOLE}
        {$IFDEF MSWINDOWS}
        LogForm.Show;
        {$ENDIF}
        MainForm.EmptyMixtureItemClick(nil);
        MainForm.LoadGeneralFile(Params[1],False);
        Mixture := MainForm.Mixture;
        Overlays := MainForm.Overlays;
      {$ELSE}
        ShowTimingAtEnd := False;
        Mixture := nil;
        Overlays := nil;
        if LowerCase(ExtractFileExt(Params[1])) = '.pgo' then begin
          Mixture := TMixture.Create(nil);
          Mixture.LoadFromFile(Params[1]);
        end else
          AddToOverlay(Overlays, Params[1],True);
      {$ENDIF}
      if CvtToPert then
        Mixture.ConvertToPerturbations;
      ApplyMixtureChanges;
      FStream := TFileStream.Create(Params[2], fmOpenRead);
      Stream := TFreeFormatStream.Create(FStream);
      nTried := 0;
      if Params[0] = '--testd3h' then
        LargestError := CheckD3hFile(Mixture, Stream, nTried, Tolerance, nFailures)
      else if Params[0] = '--testold' then
        LargestError := CheckOldPgopherFile(Mixture, Stream, nTried, Tolerance, nFailures, False)
      else if Params[0] = '--testoldnoom' then
        LargestError := CheckOldPgopherFile(Mixture, Stream, nTried, Tolerance, nFailures, True)
      else if Params[0] = '--testpgo' then
        LargestError := CheckPgopherFile(False, Mixture, Stream, nTried, Tolerance, nFailures)
      else if Params[0] = '--testgen' then
        LargestError := CheckPgopherFile(True, Mixture, Stream, nTried, Tolerance, nFailures)
      else if Params[0] = '--testeig' then
        LargestError := CheckEigenvalueFile(Mixture, Stream, nTried, Tolerance, nFailures)
      else if Params[0] = '--testovr' then
        LargestError := CheckOverlays(Overlays, Stream, nTried, Tolerance, nFailures, SortBoth, Interpolate)
      else if Params[0] = '--testjplcat' then
        LargestError := CheckJPLCatFile(Mixture, Stream, nTried, Tolerance, nFailures, Offset)
      else
        raise Exception.Create('Bad test:'+Params[0]);
      if nFailures <> 0 then begin
        Write(nFailures, ' failure(s) out of ', nTried);
        ExitCode := 1;
      end else
        Write('OK (', nTried:5, ')');
      WriteLn(' Largest Error = ', LargestError:20, '  : ', Params[1], ' ', Params[2]);
    end else if d = 1 then begin
      { Simple Calculation }
      {$IFNDEF CONSOLE}
        MainForm.LoadGeneralFile(Params[0],False);
        Mixture := MainForm.Mixture;
      {$ELSE}
        WriteLn(VersionInfo);
        Mixture := TMixture.Create(nil);
        Mixture.LoadFromFile(Params[0]);
        if Mixture.PrintLevel < 0 then
          Mixture.PrintLevel := 0;
      {$ENDIF}
      ApplyMixtureChanges;
      Mixture.MakeLineList;
    end;
   except
     on E:Exception do
       WriteLn('****Exception:', E.message);
   end;
   Stream.Free;
   FStream.Free;
   CreateLogForms;
  end else begin
    {$IFNDEF CONSOLE}
    CreateLogForms;
    if not Assigned(MainForm.Mixture) then
      MainForm.EmptyMixtureItemClick(nil);
    try
      for i := 1 to ParamCount do
        MainForm.LoadGeneralFile(ParamStr(i),i>1)
    except
      on E:Exception do
        Application.ShowException(E);
    end;
    with Application do begin
      FileName := HelpFile;
      if not FileExists(FileName) then begin
        FileName := ExtractFilePath(Application.ExeName) + FileName;
        if not FileExists(FileName) then
          FileName := ExtractFilePath(Application.ExeName) + 'Help' + PathDelim + ExtractFileName(FileName);
      end;
      HelpFile := ExpandFileName(FileName);
    end;

    Result := True;
    {$ELSE}
    WriteLn(VersionInfo,' - no files');
    {$ENDIF}
  end;
end;

end.

