unit Fitter;

interface

uses Vector, Mixture, Overlay, Obs;

type
  OnShowContour = procedure (Spectra: TOverlays; Simulation: TColumnVector) of object;

procedure FitLines(Parameters: TParameters; Obs: array of TObservation; Precision: Integer);
procedure FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; Parameters: TParameters);
function CalculateObs(var Observation: TObservation; out Value: Double): Boolean;

implementation

uses Progress, MNode, Graphics, SysUtils, MatrixVec, Nlfit, SpectrumPlot, ExpPlot, MathSup;

type
  TFitter = class
  public
    nObs, nPar: Integer;
    a: TMatrix;   { d(obs)/d(par) }
    omc: TColumnVector; { normally obs - calc }
    par: TColumnVector; { Parameters - changed by this routine }
    weights: TColumnVector; { nil or The relative standard deviation of each value in obs }
    OldSD, NewSD: Double; { Previous and estimated next standard deviations }
    UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
    b: TColumnVector;  { Standard Deviations of the parameters}
    ada: TMatrix; { Correlation Matrix }
    SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
    constructor Create(Parameters: array of TParameter; AnObs: Integer);
    destructor Destroy; override;
    procedure DoLeastSQ(Parameters: TParameters; Precision: Integer);
  end;

{ TFitter }

constructor TFitter.Create(Parameters: array of TParameter;  AnObs: Integer);
begin
  nObs := AnObs;
  nPar := Length(Parameters);
  a := TMatrix.Create(nObs, nPar);
  omc := TColumnVector.Create(nObs);
  weights := TColumnVector.Create(nObs); { Also used as workspace }
end;

destructor TFitter.Destroy;
begin
  a.Free;
  omc.Free;
  par.Free;
  b.Free;
  ada.Free;
  weights.Free;
  SVD.Free;
  SVDw.Free;
  inherited;
end;

procedure TFitter.DoLeastSQ(Parameters: TParameters; Precision: Integer);
var
  i, j: Integer;
begin
  { Fill in parameter info }
  par := TColumnVector.Create(nPar);
  for i := 0 to nPar-1 do
    par[i+1] := Parameters[i].Value;
  b := TColumnVector.Create(nPar);
  ada := TMatrix.Create(nPar, nPar);
  SVD := TMatrix.Create(nPar, nPar); { TODO: Make use of these }
  SVDw := TColumnVector.Create(nPar);
  LeastSQ(a, omc, par, weights, OldSD, NewSD, UnWtOldSD, UnWtNewSD, b, ada, SVD, SVDw, 0, 1);
  WriteLn;
  WriteLn(nObs, ' Observations,  ', nPar, ' Parameters');
  Write('Initial Average Error: ', OldSD:10:Precision);
  WriteLn('   (Unweighted: ', UnWtOldSD:10:Precision,')');
  if nPar > 0 then begin
    Write('Predicted New Error:   ', NewSD:10:Precision);
    WriteLn('   (Unweighted: ', UnWtNewSD:10:Precision,')');
    WriteLn('Parameters:');
    WriteLn('  # Old                 New                 Std Dev   Summary              Name');
    for i := 0 to nPar-1 do
      WriteLn(Format('%3d%s%s%s %-20s %s', [i+1, DtosMA(Parameters[i].Value,20,-1), DtosMA(par[i+1],20,-1), DtosMA(b[i+1],10,-1),
                     DtosErr(par[i+1],b[i+1],2), Parameters[i].Name]));
    if nPar > 1 then begin
      WriteLn;
      WriteLn('Correlation Matrix');
      for i := 1 to nPar do
        Write(i:8);
      WriteLn;
      for i := 1 to nPar do begin
        Write(i:3);
        for j := 1 to i do
          Write(ada[i,j]:8:3);
        WriteLn;
      end;
    end;
    { Copy Back Parameters }
    for i := 0 to nPar-1 do
      Parameters[i].MNode.SetParameter(Parameters[i].Index, par[i+1]);
  end;
end;

procedure FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; Parameters: TParameters);
var
  f: TFitter;
  nObs: Integer;
  i, j, l: Integer;
  Simulation: TFitPlot;
  vo, vn: Double;
  procedure CalculateObs(v: TColumnVector);
  var
    i, j, l: Integer;
  begin
    { Copy in required frequencies }
    l := 0;
    for i := 0 to Spectra.Count-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        for j := StartPoint to EndPoint do begin
          Simulation.Frequency[l] := AdjustFrequency(Frequency[j]);
          Inc(l);
        end;
    Simulation.MakeFromMixture(Mixture);
    l := 0; { Copy simluation, doing vertical scaling as we go }
    for i := 0 to Spectra.Count-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        for j := StartPoint to EndPoint do begin
          // v[l+1] := (Simulation.Plot[l]-Spectra[i].Baseline)*Spectra[i].Scale;
          v[l+1] := Simulation.Plot[l]/Scale + Baseline;
          Inc(l);
        end;
  end;
begin
  nObs := 0;
  { Find total number of observed points }
  for i := 0 to Spectra.Count-1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        Inc(nObs, EndPoint-StartPoint+1);
  f := TFitter.Create(Parameters, nObs);
  Simulation := nil;
  try
    Simulation := TFitPlot.Create(Mixture.PlotUnits);
    Simulation.ndf := nObs;
    CalculateObs(f.omc);
    ShowContour(Spectra, f.omc);
    {  Calculate Derivatives wrt each parameter }
    { TODO: Optimize for simple derivatives }
    PushProgress('Derivatives', High(Parameters));
    for i := 0 to High(Parameters) do begin
      SetProgress(i);
      vo := Parameters[i].Value;
      if Parameters[i].Increment = 0 then begin
        vn := vo * 1.0001;
        if vn = 0 then
          vn := 0.001; { Random Guess }
      end else
        vn := vo + Parameters[i].Increment;
      Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
      try
        Parameters[i].MNode.CheckProperties;
        CalculateObs(f.weights);
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
        Parameters[i].MNode.CheckProperties;
      end;
      for j := 1 to nObs do
        f.a[j,i+1] := (f.omc[j] - f.weights[j])/(vn-vo)
    end;
    PopProgress;
    FreeAndNil(f.Weights);
    l := 0; { Copy simluation, doing vertical scaling as we go }
    for i := 0 to Spectra.Count-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
       for j := StartPoint to EndPoint do begin
         // v[l+1] := (Simulation.Plot[l]-Spectra[i].Baseline)*Spectra[i].Scale;
         f.omc[l+1] := f.omc[l+1] - Intensity[j];
         Inc(l);
       end;
    f.DoLeastSQ(Parameters, Mixture.Precision);
  finally
    f.Free;
    Simulation.Free;
  end;
end;

function CalculateObs(var Observation: TObservation; out Value: Double): Boolean;
var
  Eigenvalues: TEigenvalues;
begin
  Result := True;
  with Observation do begin
    AIndexL := FindEigenvaluesAndIndex(Eigenvalues, ManifoldL, JL, SymL, IndexL, Name, True, LineNumber, FileName);
    if AIndexL > 0 then begin
      Value := - Eigenvalues.E[AIndexL];
      Colour := Eigenvalues.DominantBasis[AIndexL].State.FindColour;
    end else
      Result := False;
    AIndexU := FindEigenvaluesAndIndex(Eigenvalues, ManifoldU, JU, SymU, IndexU, Name, False, LineNumber, FileName);
    if AIndexU > 0 then begin
      Value := Value + Eigenvalues.E[AIndexU];
      if Colour = clNone then
        Colour := Eigenvalues.DominantBasis[AIndexU].State.FindColour;
    end else
      Result := False;
    if Colour = clNone then
      Colour := clBlack;
  end;
end;

procedure FitLines(Parameters: TParameters; Obs: array of TObservation; Precision: Integer);
var
  f: TFitter;
  nObs: Integer;
  i, j: Integer;
  vo, vn: Double;
  { For determining printout format }
  LargestMolecule, LargestLower, LargestUpper: string;  { For Determining field lengths }
  MoleculeChanged, LowerChanged, UpperChanged: Boolean; { For Determining if field needs printing }
  function CalculateAllObs(v: TColumnVector): Boolean;
  var
    i: Integer;
    val: Double;
    // Eigenvalues: TEigenvalues;
  begin
    Result := True;
    for i := 0 to nObs-1 do begin
      Result := Result and CalculateObs(Observations[i], val);
      v[i+1] := val;
    end;
//      with  do begin
//        AIndexL := FindEigenvaluesAndIndex(Eigenvalues, ManifoldL, JL, SymL, IndexL, Name, True, LineNumber, FileName);
//        if AIndexL > 0 then begin
//          v[i+1] := - Eigenvalues.E[AIndexL];
//          Colour := Eigenvalues.DominantBasis[AIndexL].State.FindColour;
//        end else
//          Result := False;
//        AIndexU := FindEigenvaluesAndIndex(Eigenvalues, ManifoldU, JU, SymU, IndexU, Name, False, LineNumber, FileName);
//        if AIndexU > 0 then begin
//          v[i+1] := v[i+1] + Eigenvalues.E[AIndexU];
//          if Colour = clNone then
//            Colour := Eigenvalues.DominantBasis[AIndexU].State.FindColour;
//        end else
//          Result := False;
//        if Colour = clNone then
//          Colour := clBlack;
//      end;
  end;
begin
  nObs := Length(Obs);
  if nObs <= 0 then
    raise EReportException.Create('No observations');
  if nObs < Length(Parameters) then
    raise EReportException.CreateFmt('More observations (%d) than parameters (%d)', [nObs,Length(Parameters)]);
  { Work out widths of fields for printout }
  with Observations[0] do begin
    LargestMolecule := ManifoldU.Molecule.Name;
    LargestLower := ManifoldL.Name;
    LargestUpper := ManifoldU.Name;
  end;
  MoleculeChanged := False;
  LowerChanged := False;
  UpperChanged := False;
  for i := 0 to nObs-1 do
    with Observations[i] do begin
      if LargestMolecule <> ManifoldU.Molecule.Name then begin
        MoleculeChanged := True;
        if Length(LargestMolecule) < Length(ManifoldU.Molecule.Name) then
          LargestMolecule := ManifoldU.Molecule.Name
      end;
      if LargestLower <> ManifoldL.Name then begin
        LowerChanged := True;
        if Length(LargestLower) < Length(ManifoldL.Name) then
          LargestLower := ManifoldL.Name
      end;
      if LargestUpper <> ManifoldU.Name then begin
        UpperChanged := True;
        if Length(LargestUpper) < Length(ManifoldU.Name) then
          LargestUpper := ManifoldU.Name
      end;
    end;

  f := TFitter.Create(Parameters, nObs);
  PushProgress('Derivatives', High(Parameters));
  try
    { Determine obs - calc }
    if not CalculateAllObs(f.omc) then
      raise EReportException.Create('Transitions not found');
    {  Calculate Derivatives wrt each parameter }
    for i := 0 to High(Parameters) do begin
      SetProgress(i);
      vo := Parameters[i].Value;
      if Parameters[i].Increment = 0 then begin
        vn := vo * 1.0001;
        if vn = 0 then
          vn := 0.001; { Random Guess }
      end else
        vn := vo + Parameters[i].Increment;
      Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
      try
        Parameters[i].MNode.CheckProperties;
        if not CalculateAllObs(f.weights) then
          raise EReportException.Create('Transitions not found');
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
        Parameters[i].MNode.CheckProperties;
      end;
      for j := 1 to nObs do
        f.a[j,i+1] := (f.omc[j] - f.weights[j])/(vn-vo)
    end;
    { Write out headers }
    WriteLn;
    Write('Residuals before fit');
    if MoleculeChanged then begin
      if Length(LargestMolecule) < 3 then
        LargestMolecule := 'Mol';
    end else begin
      Write(' for ', LargestMolecule);
      LargestMolecule := '';
    end;
    if not (LowerChanged or UpperChanged) then
      if LargestUpper = LargestLower then
        Write(' for the ', LargestUpper, ' state')
      else
        Write(' for the ', LargestUpper, ' - ', LargestLower, ' Transition');
    WriteLn;
    if Length(LargestMolecule) > 0 then
      Write(Copy('Molecule', 1, Length(LargestMolecule)), ' ');
    Write(' ':Length(LargestUpper), '   J''S'' #''');
    Write(' ':Length(LargestLower), '   J"S" #"');
    Writeln('     Observed    Calculated      Obs-Calc  Weight (O-C)/Wt');
    { Work out obs-calc, weights }
    for i := 0 to nObs-1 do begin
      with Observations[i] do begin
        if Length(LargestMolecule) > 0 then
          Write(ManifoldU.Molecule.Name:Length(LargestMolecule), ' ');
        WriteLn(ManifoldU.Name, ' ', JU:3, SymU:2, AIndexU:3, ' ',
                ManifoldL.Name, ' ', JL:3, SymL:2, AIndexL:3,
                AtLeast(Frequency,14,Precision), AtLeast(f.omc[i+1],14,Precision),
                AtLeast(Frequency-f.omc[i+1],14,Precision),
                Weight:8:4, DTosMA((Frequency-f.omc[i+1])/Weight,10,2), ' ', Name);
      end;
      f.omc[i+1] := f.omc[i+1] - Observations[i].Frequency;
      f.weights[i+1] := Observations[i].Weight;
    end;
    f.DoLeastSQ(Parameters, Precision);
  finally
    f.Free;
    PopProgress;
  end;
end;


end.
