unit ImportCALPGM;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

{ Import data file from Herb Pickett's CALPGM program suite }
interface

uses TextStrm, Mixture;

function ImportCALPGMFile(S: TFreeFormatStream): TMixture;

implementation

uses {$IFNDEF CONSOLE} Dialogs, {$ENDIF}
     HundA, OAsymT, SymT, SysUtils, Units, MNode, QuantumNumber;

type
  TCALPGMReader = class
  public
    Title, Comment: string;
    CHR: Char;
    SPIND, NVIB, KNMIN, KNMAX, IXX, IAX, WTPL, WTMN, VSYM, EWT, DIAG: Integer;
    { Parameter details }
    IDPAR, FF, I2, I1, NS, TYP, KSQ, NSQ, V2, V1: Integer;
    PAR, ERRPAR: Double;
    ID: Integer; { IDPAR less vibrational number }

    S: TFreeFormatStream;
    Mixture: TMixture;
    constructor Create(Stream: TFreeFormatStream);
    function ReadParameterLine: Boolean;
  end;

  { TMoleculeReader }

  TMoleculeReader = class
  public
    R: TCALPGMReader;
    nProblems: Integer;
    ElectronSpin: TQuantumNumber;
    ElectronSpinIndex: Integer;
    NuclearSpins: array of TQuantumNumber;
    SpinIndex: array of Integer;
    constructor Create(AReader: TCALPGMReader);
    procedure Read; virtual;
    function AddState(Manifold: TManifold):TState; virtual; abstract;
    function ConvertParameter(State: TState; PAR: Double):Boolean; virtual; abstract;
    function SParameter(IDI: Integer; State: TState; PAR: Double):Boolean; virtual; abstract;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; PAR: Double):Boolean; virtual; abstract;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; PAR: Double):Boolean; virtual; abstract;
    function ConvertPerturbation(State1, State2: TState; PAR: Double):Boolean; virtual; abstract;
    procedure AddSpins;
    function ConvertParametersWithSpin(State: TState; PAR: Double):Boolean;
  end;

  TAsymmetricMoleculeReader = class(TMoleculeReader)
  public
    Manifold: TAsymmetricManifold;
    Molecule: TAsymmetricMolecule;
    RepKnown: Boolean;
    procedure Read; override;
    function AddState(AManifold: TManifold):TState; override;
    function ConvertParameter(State: TState; PAR: Double):Boolean; override;
    function SParameter(IDI: Integer; State: TState; PAR: Double):Boolean; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; PAR: Double):Boolean; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; PAR: Double):Boolean; override;
    function ConvertPerturbation(State1, State2: TState; PAR: Double):Boolean; override;
  end;

  { TLinearMoleculeReader }

  TLinearMoleculeReader = class(TMoleculeReader)
  public
    Manifold: TLinearManifold;
    Molecule: TLinearMolecule;
    procedure Read; override;
    function AddState(AManifold: TManifold):TState; override;
    function ConvertParameter(State: TState; PAR: Double):Boolean; override;
    function SParameter(IDI: Integer; State: TState; PAR: Double):Boolean; override;
    function ISParameter(INuc: Integer; IDI: Integer; State: TState; PAR: Double):Boolean; override;
    function IParameter(INuc: Integer; IDI: Integer; State: TState; PAR: Double):Boolean; override;
    function ConvertPerturbation(State1, State2: TState; PAR: Double):Boolean; override;
  end;

{ TMoleculeReader }

constructor TMoleculeReader.Create(AReader: TCALPGMReader);
begin
  R := AReader;
  nProblems := 0;
end;

procedure TMoleculeReader.Read;
var
  Manifold: TManifold;
  i, j, StateMin, StateMax: Integer;
  State: TState;
  procedure ShowProblem;
  begin
    Inc(nProblems);
    WriteLn('Problem with parameter on line ', R.S.Line, ': ',
             R.IDPAR, ' ', R.PAR, ' ', R.ERRPAR, ' ', R.Comment);
  end;
begin
  R.Mixture.Species[0].Isotopomer[0].Name := Trim(Copy(R.Title,1,50));
  R.Mixture.Species[0].Isotopomer[0].Comment := R.Title;
  Manifold := R.Mixture.Species[0].Isotopomer[0].Manifold[0];
  Manifold.Name := 'X';
  Manifold.Initial := True;
  Manifold.Molecule.nNuclei := Length(NuclearSpins);
  for i := 1 to Abs(R.NVIB) do begin
    State := AddState(Manifold);
    State.Name := IntToStr(i-1);
    State.AdjustNuclei;
    for j := 0 to Manifold.Molecule.nNuclei-1 do
      State.Nucleus(j).Spin := NuclearSpins[j];
  end;
  while R.ReadParameterLine do begin
    StateMin := R.V1;
    StateMax := R.V2;
    if ((Abs(R.NVIB) < 10) and (R.V1 = 9) and (R.V2 = 9))
       or ((R.V1 = 99) and (R.V2 = 99)) then begin
      StateMin := 0;
      StateMax := Abs(R.NVIB)-1;
    end else if R.V1 <> R.V2 then begin
      if ConvertPerturbation(Manifold.State[R.V1], Manifold.State[R.V2], R.PAR) then
        ShowProblem;
      Continue;
    end;
    for i := StateMin to StateMax do
      if ConvertParametersWithSpin(Manifold.State[i], R.PAR) then
        ShowProblem;
  end;
end;

procedure TMoleculeReader.AddSpins;
var
  t: Integer;
  SpinDigits: Integer;
  Index, SIndex: Integer;
  Buffer: string;
  function NextSpin: TQuantumNumber;
  begin
    Result := (t mod SpinDigits) - 1;
    t := t div SpinDigits;
  end;
begin
  ElectronSpin := 0;
  ElectronSpinIndex := -1;
  if R.SPIND <> 0 then begin
    Buffer := '1';
    if not InputQuery('Select Electron Spin', 'Enter spin number for electron (0 for none)', Buffer) then
      Abort;
    ElectronSpinIndex := StrToInt(Buffer);
    t := Abs(R.SPIND);
    if t mod 10 = 0 then begin
      SpinDigits := 100;
      t := t div 10;
    end else
      SpinDigits := 10;
    Index := 0;
    SIndex := 0;
    while t <> 0 do begin
      SetLength(SpinIndex, Index+2); { Ignore first entry }
      if Index = ElectronSpinIndex-1 then begin
        ElectronSpin := NextSpin;
        WriteLn('Assuming electron is spin number ', ElectronSpinIndex+1, ' (', ElectronSpin, '/2)');
        SpinIndex[High(SpinIndex)] := -1;
      end else begin
        SetLength(NuclearSpins, Length(NuclearSpins)+1);
        NuclearSpins[High(NuclearSpins)] := NextSpin;
        SpinIndex[High(SpinIndex)] := SIndex;
        Inc(SIndex);
        WriteLn('Assuming spin number ', Index, ' is a nuclear spin of ', NuclearSpins[High(NuclearSpins)], '/2');
      end;
      Inc(Index);
    end
  end;
end;

function ImportCALPGMFile(S: TFreeFormatStream): TMixture;
var
  Reader: TCALPGMReader;
  MoleculeReader: TMoleculeReader;
  Species: TSpecies;
begin
  Reader := TCALPGMReader.Create(S);
  try
    Result := TMixture.Create(nil);
    try
      Reader.Mixture := Result;
      Result.Name := 'Mixture';
      Result.Units := MHz;
      Result.IntensityUnits := nm2MHzperMolecule;
      Species := TSpecies.Create(Result);
      Species.Name := 'S';
      if Reader.KNMIN = Reader.KNMAX then
        MoleculeReader := TLinearMoleculeReader.Create(Reader)
      else if Reader.SPIND >= 0 then
        MoleculeReader := TAsymmetricMoleculeReader.Create(Reader)
      else
        MoleculeReader := nil; // TSymmetricMolecule.Create(Species);
      try
        MoleculeReader.Read;
        {$IFNDEF CONSOLE}
        if MoleculeReader.nProblems > 0 then
          ShowMessage('Problems with import - see log window');
        {$ENDIF}
      finally
        MoleculeReader.Free;
      end;
      Result.FixupProperties;
    except
      Result.Free;
      Raise;
    end;
  finally
    Reader.Free;
  end;
end;

function TMoleculeReader.ConvertParametersWithSpin(State: TState; PAR: Double): Boolean;
var
  IDI: Integer;
begin
 if (R.I1 <> 0) or (R.I2 <> 0) then begin
   IDI := (((R.FF*1000 + R.NS)*100 + R.TYP)*10 + R.KSQ)*10 + R.NSQ;
   if R.I1 = ElectronSpinIndex then
     if R.I2 = 0 then
       Result := SParameter(IDI, State, PAR)
     else if R.I2 = ElectronSpinIndex then
       Result := SParameter(IDI + 1100000, State, PAR)
     else
       Result := ISParameter(R.I2, IDI, State, PAR)
   else if R.I2 = ElectronSpinIndex then
     if R.I1 = 0 then
       Result := SParameter(IDI, State, PAR)
     else
       Result := ISParameter(R.I1, IDI, State, PAR)
   else if R.I2 = 0 then
     Result := IParameter(R.I1, IDI, State, PAR)
   else if R.I1 = 0 then
     Result := IParameter(R.I2, IDI, State, PAR)
   else if R.I1 = R.I2 then
     Result := IParameter(R.I2, IDI + 1100000, State, PAR)
   else
     Result := True;
 end else
   Result := ConvertParameter(State, PAR)
end;

{ TCALPGMReader }

constructor TCALPGMReader.Create(Stream: TFreeFormatStream);
begin
  S := Stream;
  Title := S.RestOfLine;
  S.NextLine;
  S.NextLine; { Skip NPAR, NLINE, NITR, NXPAR,  THRESH, ERRTST, FRAC, CAL }
  S.FlowToNextLine := False;
  { Line 3 }
  CHR := S.NextString('g')[1];
  S.NextIntegerVar(SPIND);
  S.NextIntegerVar(NVIB);
  if NVIB = 0 then begin
    NVIB := 1;
    WriteLn('NVIB read as 0 - assuming 1');
  end;
  S.NextIntegerVar(KNMIN);
  S.NextIntegerVar(KNMAX);
  IXX := S.NextInteger(0); if IXX <> 0 then WriteLn('IXX =', IXX, ' but all interactions included');
  S.NextIntegerVar(IAX);
  S.NextIntegerVar(WTPL);
  S.NextIntegerVar(WTMN);
  S.NextIntegerVar(VSYM);
  S.NextIntegerVar(EWT);
  DIAG := S.NextInteger(1); // ???
  S.NextLine;
end;

function TCALPGMReader.ReadParameterLine: Boolean;
var
  t: Integer;
  idstring: string;
  function ND(n: Integer = 10): Integer;
  begin
    Result := t mod n;
    t := t div n;
  end;
begin
  Result := False;
  if not S.GetFreeItem or (S.Token = '') then
    Exit;
  idstring := S.Token;
  if idstring[Length(idstring)] = '.' then
    Delete(idstring, Length(idstring), 1);
  IDPAR := StrToInt(idstring);
  if IDPAR = MaxInt then
    Exit;
  t := Abs(IDPAR);
  if Abs(NVIB) < 10 then begin
    V1 := ND;
    V2 := ND;
  end else if Abs(NVIB) < 100 then begin
    V1 := ND(100);
    V2 := ND(100);
  end else begin
    V1 := ND(1000);
    V2 := ND(1000);
  end;
  ID := t;
  NSQ := ND;
  KSQ := ND;
  TYP := ND(100);
  NS := ND;
  I1 := ND;
  I2 := ND;
  FF := ND;
  S.NextDoubleVar(PAR);
  S.NextDoubleVar(ERRPAR);
  Comment := S.RestOfLine;
  S.NextLine;
  Result := True; { Read a line }
end;

{ TLinearMoleculeReader }

procedure TLinearMoleculeReader.Read;
begin
  Molecule := TLinearMolecule.Create(R.Mixture.Species[0]);
  Molecule.Symmetric := R.WTMN <> R.WTPL;
  AddSpins;
  Manifold := TLinearManifold.Create(Molecule);
  inherited Read;
end;

function TLinearMoleculeReader.AddState(AManifold: TManifold): TState;
var
  State: TLinear;
begin
  State := TLinear.Create(AManifold);
  State.S := ElectronSpin;
  State.Lambda := R.KNMIN*2;
  Result := State;
end;

function TLinearMoleculeReader.ConvertParameter(State: TState; PAR: Double): Boolean;
begin
 Result := False; { assume OK }
 with State as TLinear do
   case R.ID of
     0: Origin := Origin + PAR;
     1: B := B + PAR;
     2: D := D - PAR;
     3: H := H + PAR;
     4: L := L + PAR;
     5: M := M + PAR;
     6: PP := PP + PAR;

   400: q := q + 2*PAR;
   401: qD := qD + 2*PAR;
   402: qH := qH + 2*PAR;
   403: qL := qL + 2*PAR;
   else
     Result := True;
   end;
end;

function TLinearMoleculeReader.ConvertPerturbation(State1, State2: TState; PAR: Double): Boolean;
begin
  Result := True;
end;

function TLinearMoleculeReader.SParameter(IDI: Integer; State: TState; PAR: Double): Boolean;
begin
  Result := False;
  with State as TLinear do
    case IDI of
      0: gamma := gamma + PAR;
      1: gammaD := gammaD + PAR;
      2: gammaH := gammaH + PAR;
      3: gammaL := gammaL + PAR;

    400: p := p - 2*PAR;
    401: pD := pD - 2*PAR;
    402: pH := pH - 2*PAR;
    403: pL := pL - 2*PAR;

    100: A := A + PAR;

    1100100: LambdaSS := LambdaSS + PAR/(S*(2*S-2));
    1100101: LambdaD :=  LambdaD  + PAR/(S*(2*S-2));
    1100400: o := o + PAR;
    1100401: oD := oD + PAR;
    else
      Result := True;
    end;
end;

function TLinearMoleculeReader.ISParameter(INuc, IDI: Integer; State: TState; PAR: Double): Boolean;
begin
  Result := False;
  with State.Nucleus(SpinIndex[INuc]) as TLinearNucleus do
    case IDI of
      0: b := b + PAR;
    100: begin
           c := c + PAR;
           b := b - c/3;
         end;
    400: d := d - 2*PAR;
    401: dJ := dJ - 2*PAR;
    else
      Result := True;
    end;
end;

function TLinearMoleculeReader.IParameter(INuc, IDI: Integer; State: TState; PAR: Double): Boolean;
begin
  Result := False;
  with State.Nucleus(SpinIndex[INuc]) as TLinearNucleus do
    case IDI of
            0: cI := cI + PAR;
          100: a := a + PAR;
      1100100: eQq0 := eQq0 + PAR/1.5;
      1100400: eQq2 := eQq2 - 4*PAR;
    else
      Result := True;
    end;
end;

{ TAsymmetricMoleculeReader }

function TAsymmetricMoleculeReader.ConvertParameter(State: TState; PAR: Double):Boolean;
  function SetRep(SRep: Boolean):Double;
  var
    i: Integer;
  begin
    Result := PAR;
    if RepKnown then begin
      if TAsymmetricTop(State).Molecule.SReduction <> SRep then
        raise Exception.Create('Inconsistent reduction');
    end else begin
      TAsymmetricTop(State).Molecule.SReduction := SRep;
      if SRep then
        for i := 0 to Abs(R.NVIB)-1 do
          with Manifold.State[i] as TAsymmetricTop do
            deltaJ := - deltaJ;
    end;
    RepKnown := True;
  end;
begin
 Result := False; { assume OK }
 with State as TAsymmetricTop do
  case R.ID of
     0: Origin := Origin + PAR;
   100: A := A + PAR;
   200: B := B + PAR;
   300: C := C + PAR;

     1: begin
          BBar := BBar + PAR;
          SetParameter(zIndex,GetParameter(zIndex)+PAR);
        end;
    10: SetParameter(zIndex,GetParameter(zIndex)+PAR);
   400: BDelta := BDelta + 4*PAR;

     2: DJ := DJ - PAR;
    11: DJK := DJK - PAR;
    20: DK := DK - PAR;
//           401: delJ := -SetRep(False); ?? Needs adjusting
   410: deltaK := deltaK -SetRep(False);
   401: if Molecule.SReduction then
          deltaJ := deltaJ + PAR
        else
          deltaJ := deltaJ - PAR;
   500: deltaK := deltaK + SetRep(True);

     3: HJ := HJ + PAR;
    12: HJK := HJK + PAR;
    21: HKJ := HKJ + PAR;
    30: HK := HK + PAR;
   501: phiJK := phiJK + SetRep(True);
   600: phiK := phiK + SetRep(True);
   402: phiJ := phiJ + PAR;
   411: phiJK := phiJK + SetRep(False);
   420: phiK := phiK + SetRep(False);

     4: LJ := LJ + PAR;
    13: LJJK := LJJK + PAR;
    22: LJK := LJK + PAR;
    31: LKKJ := LKKJ + PAR;
    40: LK := LK + PAR;
   403: llJ := llJ + PAR;
   412: llJK := llJK + SetRep(False);
   421: llKJ := llKJ + SetRep(False);
   430: llK := llK + SetRep(False);
   502: llJK := llJK + SetRep(True);
   601: llKJ := llKJ + SetRep(True);
   700: llK := llK + SetRep(True);

 10001: DsN := DsN + PAR;
 50001: DsNK := DsNK + PAR; { A Reduction }
 10010: DsKN := DsKN + PAR;
 50010: DsK := DsK + PAR; { A Reduction }
 10400: ds1 := ds1 + PAR;
 50400: ds2 := ds2 + PAR; { A Reduction }

  else
    Result := True;
  end;
  if Result then
    Result := ConvertPerturbation(State, State, PAR);
end;

function TAsymmetricMoleculeReader.ConvertPerturbation(State1, State2: TState; PAR: Double): Boolean;
begin
  Result := True;
  if Result and (R.FF = 0) and (R.I2 = 0) and (R.I1 = 0) and (R.NS = 0) then
    with TAsymmetricTopPerturbation.Create(Manifold) do begin
      SetBraKetName(State1.Name, State2.Name);
      Value := PAR;
      JPower := R.NSQ*2;
      SetAxisPower(zAxis, 2*R.KSQ);
      Result := False;
      Comment := IntToStr(R.IDPAR);
      case R.TYP of
        0: ;
        1: aPower := aPower + 2;
        2: bPower := bPower + 2;
        3: cPower := cPower + 2;
        4..10: pPower := 2*(R.TYP-3);
        20: aPower := aPower + 1;
        21: begin bPower := bPower + 1; cPower := cPower + 1; end;
        40: bPower := bPower + 1;
        41: begin aPower := aPower + 1; cPower := cPower + 1; end;
        60: cPower := cPower + 1;
        61: begin aPower := aPower + 1; bPower := bPower + 1; end;
      else
        Free;
      end;
    end;
end;

procedure TAsymmetricMoleculeReader.Read;
begin
  Molecule := TAsymmetricMolecule.Create(R.Mixture.Species[0]);
  AddSpins;
  if R.NVIB > 0 then
    Molecule.Representation := Ir
  else
    Molecule.Representation := IIIl;
  Molecule.PointGroup := C2v;
  Molecule.C2zAxis := TabcAxis(Abs(R.IAX)-1);
  Molecule.eeWt := R.WTPL; { TODO: Fix next four lines }
  Molecule.eoWt := R.WTPL;
  Molecule.oeWt := R.WTMN;
  Molecule.ooWt := R.WTMN;
  Manifold := TAsymmetricManifold.Create(Molecule);
  RepKnown := False;
  inherited;
end;

function TAsymmetricMoleculeReader.AddState(AManifold: TManifold): TState;
var
  State: TAsymmetricTop;
  t: Integer;
begin
  State := TAsymmetricTop.Create(AManifold);
  State.S := ElectronSpin;
  Result := State;
  { This bizarre construct avoids a bug in Delphi }
  t := 2*R.KNMIN;
  State.Kmin := t;
  State.Kmax := 2*R.KNMAX;
end;

function TAsymmetricMoleculeReader.SParameter(IDI: Integer; State: TState; PAR: Double): Boolean;
begin
  Result := False;
  with State as TAsymmetricTop do
    case IDI of
    100: eaa := eaa + PAR;
    200: ebb := ebb + PAR;
    300: ecc := ecc + PAR;
   6100: eabbar := eabbar + PAR;
    101: DsNK := DsNK + PAR; { S Reduction }
    110: DsK := DsK + PAR;   { S Reduction }
    500: ds2 := ds2 + PAR;
1100100: alpha := alpha + 2*PAR/(3*S*(2*S-2));

    else
      Result := True;
    end;
end;

function TAsymmetricMoleculeReader.ISParameter(INuc, IDI: Integer; State: TState; PAR: Double): Boolean;
  procedure SetTqq(abcAxis: TabcAxis);
  var
    xyzAxis: TxyzAxis;
  begin
    xyzAxis := (State.Molecule as TAsymmetricMolecule).xyzAxis(abcAxis);
    with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
      case xyzAxis of
        xAxis: Txxmyy := Txxmyy + PAR;
        yAxis: Txxmyy := Txxmyy - PAR;
        zAxis: Tzz := Tzz + Par;
      end;
  end;
begin
  Result := False;
  with State.Nucleus(SpinIndex[INuc]) as TAsymmetricTopNucleus do
    case IDI of
      0: aF := aF + PAR;
    100: SetTqq(aAxis);
    200: SetTqq(bAxis);
    300: SetTqq(cAxis);
   2100: Tbc := Tbc + PAR;
   4100: Tac := Tac + PAR;
   6100: Tab := Tab + PAR;
    else
      Result := True;
    end;
end;

function TAsymmetricMoleculeReader.IParameter(INuc, IDI: Integer; State: TState; PAR: Double): Boolean;
begin
  Result := True;
end;

end.
