unit Overlay;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Graphs, GraphRange, {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$ENDIF}
     Units, Graphics, SysUtils, Classes, MNode, Mixture, DOM;

type

  { TPlot }

  TPlot = class(TStreamableMNode)
  protected
    function GetHMax: Double; virtual; abstract;
    function GetHMin: Double; virtual; abstract;
    function GetIMin: Double; virtual; abstract;
    function GetIMax: Double; virtual; abstract;
  public
    Peaks: TDoubleArray;
    constructor Create(AOwner: TComponent); override;
    procedure AddPeak(Position: Double; DisplayUnits: TResonanceUnits; DoCopyToClipboard: Boolean = True);
    function AdjustFrequency(Freq: Double):Double; virtual;
    function DeAdjustFrequency(Freq: Double):Double; virtual;
    procedure ClearPeaks;
    procedure ClearSelection; virtual; abstract;
    procedure CycleUnits; override;
    procedure DeleteLastPeak;
    function Empty: Boolean; virtual; abstract;
    function FindPeak(AFmin, AFmax: Double; var Position: Double; PlotUnits: TResonanceUnits): Boolean; virtual; abstract;
    procedure Invert; virtual; abstract;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    class function OwnerType: TComponentClass; override;
    // procedure LoadFromFile(FileName: string); virtual; abstract;
    function nDF: Integer; virtual; abstract;
    function nChildTypes: Integer; override;
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    procedure SelectFrequencyRange(AFmin, AFmax: Double); virtual; abstract;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; AImin, AImax: Double; PlotUnits: TResonanceUnits; var PlotRanges: TPlotRanges; ShowName: Boolean); virtual;
    procedure ShowPeaks(G: TGraphD; AImin, AImax: Double; PlotUnits: TResonanceUnits);
    {$ENDIF}
    procedure ReadPeaks(Child: TDOMNode);
    procedure WritePeaks(Parent, Result: TDOMNode);
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    property HMin: Double read GetHMin;
    property HMax: Double read GetHMax;
    property IMin: Double read GetIMin;
    property IMax: Double read GetIMax;
  published
    property FrequencyOffset: Double index 1 read GetParameter write SetParameter; { Frequency Offset }

    property Units:  TResonanceUnits index 1 read GetUnits  write SetUnits;
    property Colour: TColor          index 3 read GetColour write SetColour;
    property Hide:   Boolean         index 4 read GetFlag   write SetFlag;
  end;

  TPlotClass = class of TPlot;

  TOverlays = class(TStreamableMNode)
  private
    function GetPlot(AIndex: Integer): TPlot;
  public
    HMin, HMax: Double; { Units as set below }
    Imin, Imax: Double;
    LastShowOffset: Boolean;
    constructor Create(AOwner: TComponent); override;
    procedure CycleUnits; override;
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; PlotUnits: TResonanceUnits; var PlotRanges: TPlotRanges; Offset: Boolean = false);
    function FindPeak(AArea: TDRect; var Position: Double; var Plot: TPlot; PlotUnits: TResonanceUnits): Boolean;
    {$ENDIF}
    function FixupProperties:Exception; override;
    procedure Invert;
    procedure ChooseUniqueColourFor(Plot: TPlot);
    function nVisible: Integer; { Number of non-hidden plots }
    property Items[AIndex: Integer]: TPlot read GetPlot; default;
    function nChildTypes: Integer; override;
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    procedure SaveToFile(AFileName: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Units:         TResonanceUnits index 1 read GetUnits write SetUnits;
    property PeakThreshold: Double index 1 read GetParameter write SetParameter;
  end;

{ This routine adds to the given overlay, or creates the overlay if necessary;
  It return true if a new overlay hes been created }
function AddToOverlay(var Overlay: TOverlays; FileName: string; Merge: Boolean):Boolean; overload;

implementation

uses {$IFNDEF CONSOLE} Huge, Clipbrd, ExpBitmap, MNodeClip, {$ENDIF}
     Math, MathSup, MNodeColour, ExpPlot;

function AddToOverlay(var Overlay: TOverlays; FileName: string; Merge: Boolean):Boolean;
var
  Plot: TPlot;
  i: Integer;
  Stream: TFileStream;
  NewOverlay: TOverlays;
  function TryLoading(PlotClass: TPlotClass): TPlot;
  begin
    Result := PlotClass.Create(NewOverlay);
    try
      Result.LoadFromFile(FileName);
    except
      FreeAndNil(Result);
    end;
  end;
begin
  Result := not Merge or not Assigned(Overlay);
  Plot := nil;
  try
    if Result then begin
      NewOverlay := TOverlays.Create(nil);
      NewOverlay.Name := 'Overlays';
      // if Assigned(Mixture) then
      //  Overlays.Units := Mixture.Units;
    end else
      NewOverlay := Overlay;
    if LowerCase(ExtractFileExt(FileName)) = '.ovr' then begin
      Stream := TFileStream.Create(FileName, fmOpenRead);
      try
        if Result then
          NewOverlay.LoadFromStream(Stream, FileName)
        else
          NewOverlay.MergeFromStream(Stream, FileName)
      finally
        Stream.Free;
      end;
    end else {$IFNDEF CONSOLE} if (Filename <> '') or (PasteXMLNodeOnto(NewOverlay) = nil) then {$ENDIF} begin
      Plot := TryLoading(TExperimentalPlot);
      {$IFNDEF CONSOLE}
      if not Assigned(Plot) then
        Plot := TryLoading(TBitmapPlot);
      {$ENDIF}
      if not Assigned(Plot) then begin
        if FileName = '' then
          FileName := 'clipboard';
        raise EReportException.Create('Overlay Load from ' + FileName + ' failed');
      end;
      Plot.Name := ExtractFileName(FileName);
      if Plot.Name = '' then
        Plot.Name := IntToStr(NewOverlay.Count)
      else if Plot.Comment = '' then
        Plot.Comment := FileName;
      for i := Plot.ComponentIndex to NewOverlay.Count-1 do
        NewOverlay.ChooseUniqueColourFor(NewOverlay.Items[i]);
    end;
    NewOverlay.CheckProperties;
    if Result then begin
      Overlay.Free;
      Overlay := NewOverlay;
      PerformMNodeNotification(NewOverlay, mnNewBaseNode, 0);
    end;
  except
    if Result then
      FreeAndNil(NewOverlay)
    else
      Plot.Free;
    raise;
  end;
end;

{ TPlot }

function TPlot.DeAdjustFrequency(Freq: Double): Double;
begin
  Result := Freq - FrequencyOffset;
end;

function TPlot.AdjustFrequency(Freq: Double): Double;
begin
  Result := Freq + FrequencyOffset;
end;

{$IFNDEF CONSOLE}
procedure TPlot.Show(G: TGraphD; AImin, AImax: Double; PlotUnits: TResonanceUnits; var PlotRanges: TPlotRanges; ShowName: Boolean);
var
  xl, xr: Double;
begin
  if ShowName then begin
    xl := Convert(HMin,Units,PlotUnits);
    xr := Convert(HMax,Units,PlotUnits);
    InOrder(xl,xr);
    if xl - G.Xmin > G.Xmax - xr then
      { More space on left }
      G.TextJ(xl, (AImax+AImin)/2, taRightJustify, Name)
    else { More space on right }
      G.Text(xr, (AImax+AImin)/2, Name)
  end;
  ShowPeaks(G, AImin, AImax, PlotUnits);
end;

procedure TPlot.ShowPeaks(G: TGraphD; AImin, AImax: Double; PlotUnits: TResonanceUnits);
var
  i: Integer;
  Mi, Mx, Fi, Fx, s, ds: Double;
begin
  G.MapArg(Fi, Fx, Mi, Mx);
  if Fi > Fx then begin
    s := Fi;
    Fi := Fx;
    Fx := s;
  end;
  with Owner as TOverlays do
    if LastShowOffset or (Count = 1) then
      G.Style(4);
  ds := (Fx-Fi)/G.Width;
  for i := 0 to High(Peaks) do begin
    s := Convert(AdjustFrequency(Peaks[i]), Units, PlotUnits);
    if (s >= Fi) and (S <= Fx) then begin
      G.ClipPoint(s, AImin);
      G.ClipJoin(s, AImax);
      G.TextAngle(90);
      G.TextJ(S, AImax, taRightJustify, RoundedStr(s, ds));
    end;
  end;
  G.TextAngle(0);
end;

var
  CF_Measurement: DWORD;
{$ENDIF}

procedure TPlot.AddPeak(Position: Double; DisplayUnits: TResonanceUnits; DoCopyToClipboard: Boolean);
{$IFNDEF CONSOLE}
var
  Stream: THugeMemoryStream;
  Lab1, Lab2, Lab: string;
{$ENDIF}
begin
  Setlength(Peaks, Length(Peaks)+1);
  Peaks[Length(Peaks)-1] := DeAdjustFrequency(Position);
  {$IFNDEF CONSOLE}
  if not DoCopyToClipboard then
    Exit;
  if CF_Measurement = 0 then
    CF_Measurement := RegisterClipboardFormat('Measurement');
  Position := Convert(Position, Units, DisplayUnits);
  Stream := THugeMemoryStream.Create;
  (* Mimic the following C structure:
    typedef struct {
      double value;
      char signature[4];    /* Normally MEAS */
      char filename[1];     /* Null terminated file name; can be longer */
    } ClipboardMeasurement ;
  *)
  Stream.Write(Position, Sizeof(Double));
  Stream.Write('MEAS', 4);
  Lab1 := ExtractFileName(Name);
  if (Lab1 = '1') and (Owner.ComponentCount = 1) then
    Lab1 := '';
  Lab2 := ExtractFileName(TMNode(Owner).Name);
  if (Lab1 <> '') and (Lab2 <> '') then
    Lab := Lab1 + ' - ' + Lab2
  else
    Lab := Lab1 + Lab2;
  if Lab <> '' then
    Stream.Write(Lab[1], Length(Lab)+1)
  else
    Stream.Write(#0#0, 1);
  { Copy to clipboard }
  Clipboard.Open;
  try
    Clipboard.AsText := RoundedStr(Position, Abs(HMax-Hmin)/nDF/10);
    Stream.CopyToClipboard(CF_Measurement);
  finally
    Clipboard.Close;
    Stream.Free;
  end;
  {$ENDIF}
end;

procedure TPlot.ClearPeaks;
begin
  SetLength(Peaks,0);
end;

constructor TPlot.Create(AOwner: TComponent);
begin
  SavedProperties := 'Colour=None,' + SavedProperties;
  inherited;
end;

procedure TPlot.DeleteLastPeak;
begin
  if Length(Peaks) > 0 then
    SetLength(Peaks, Length(Peaks)-1);
end;

function TPlot.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

class function TPlot.OwnerType: TComponentClass;
begin
  Result := TOverlays;
end;

function TPlot.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := 'Horizontal Axis units.';
    3: Result := 'Colour for plot.';
    4: Result := 'Set to hide this plot.'
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TPlot.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Frequency offset used when displaying plot.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TPlot.WritePeaks(Parent, Result: TDOMNode);
var
  Strings: TStringList;
  i: Integer;
begin
  if Length(Peaks) = 0 then
    Exit;
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Peaks)+1;
    Strings.Add('');
    for i := 0 to High(Peaks) do
      Strings.Add(FloatToStr(Peaks[i]));
    Result.AppendChild(Parent.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

function TPlot.nChildTypes: Integer;
begin
  Result := 0;
end;

function TPlot.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  Result :=nil;
end;

procedure TPlot.ReadPeaks(Child: TDOMNode);
var
  Strings: TStringList;
  i: Integer;
  v: Double;
begin
  while Assigned(Child) do begin
    if Child is TDOMCDATASection then begin
      Strings := TStringList.Create;
      try
        Strings.Text := TDOMCDATASection(Child).Data;
        for i := 0 to Strings.Count-1 do
          if TryStrToFloat(Strings[i], v) then
            AddPeak(v, Units, False);
      finally
        Strings.Free;
      end;
    end;
    Child := Child.NextSibling;
  end;
end;

procedure TPlot.CycleUnits;
var
  NewUnits: TResonanceUnits;
begin
  if Units = High(TResonanceUnits) then
    NewUnits := Low(TResonanceUnits)
  else
    NewUnits := Succ(Units);
  ScaleParameters(Units, NewUnits);
  Units := NewUnits;
end;

procedure TPlot.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
begin
  for i := 0 to High(Peaks) do
    Peaks[i] := Convert(Peaks[i], UFrom, UTo);
end;

{ TOverlays }

constructor TOverlays.Create(AOwner: TComponent);
begin
  SavedProperties := 'PeakThreshold=0.5,' + SavedProperties;
  DimensionlessParameters := 'PeakThreshold=0.5,' + DimensionlessParameters;
  inherited;
end;

function TOverlays.FixupProperties:Exception;
var
  Started: Boolean;
  v: Double;
  i: Integer;
begin
  Result := inherited FixupProperties;
  if Assigned(Result) or (Count = 0) then Exit;
  Started := False;
  for i := 0 to Count-1 do
    if not Items[i].Empty and not Items[i].Hide then begin
      v := Convert(Items[i].HMin, Items[i].Units, Units);
      if not Started or (HMin > v) then HMin := v;
      v := Convert(Items[i].HMax, Items[i].Units, Units);
      if not Started or (HMax < v) then HMax := v;
      if not Started or (IMin > Items[i].IMin) then IMin := Items[i].IMin;
      if not Started or (IMax < Items[i].IMax) then IMax := Items[i].IMax;
      Started := True;
    end;
end;

function TOverlays.GetPlot(AIndex: Integer): TPlot;
begin
  Result := Components[AIndex] as TPlot;
end;

{$IFNDEF CONSOLE}
function TOverlays.FindPeak(AArea: TDRect; var Position: Double; var Plot: TPlot; PlotUnits: TResonanceUnits): Boolean;
var
  i, nFound: Integer;
  xl, dx, p, Frac, CFrac, Left, Right: Double;
begin
  { Track fraction of plot area covered }
  Result := False;
  if nVisible = 0 then
    Exit;
  dx := 1/nVisible;
  xl := -dx;
  Frac := 0;
  nFound := 0;
  for i := 0 to Count-1 do begin
    if Items[i].Hide then
      Continue;
    xl := xl + dx;
    if LastShowOffset and
       ( (Max(AArea.Top, AArea.Bottom) < xl)
         or (Min(AArea.Top, AArea.Bottom) > xl+dx) ) then
      Continue;
    with Items[i] do begin
      Left := Convert(AArea.Left, PlotUnits, Units);
      Right := Convert(AArea.Right, PlotUnits, Units);
      InOrder(Left, Right);
      if FindPeak(Left, Right, p, Units) then begin
        CFrac := (Min(Right,HMax) - Max(HMin,Left))/(Right - Left);
        if CFrac > Frac then begin
          Position := p;
          Plot := GetPlot(i);
          if CFrac > 1.5*Frac then
            nFound := 1
          else
            Inc(nFound);
          Frac := CFrac;
        end;
      end;
    end;
  end;
  Result := (nFound = 1) or ((nFound > 1) and LastShowOffset);
  if Result then
    Plot.AddPeak(Position, Units);
end;

procedure TOverlays.Show(G: TGraphD; PlotUnits: TResonanceUnits; var PlotRanges: TPlotRanges; Offset: Boolean = false);
var
  i: Integer;
  xl, dx: Double;
begin
  LastShowOffset := Offset;
  if nVisible = 0 then Exit;
  xl := 0;
  if Offset then
    dx := 1/nVisible
  else
    dx := 1;
  for i := 0 to Count-1 do
    if not Items[i].Hide then begin
      Items[i].Show(G, xl, xl+dx, PlotUnits, PlotRanges, Offset and (nVisible > 1));
      if Offset then
        xl := xl + dx;
    end;
end;
{$ENDIF}

procedure TOverlays.ChooseUniqueColourFor(Plot: TPlot);
var
  i, ci: Integer;
  Clash: Boolean;
begin
  for ci := 1 to High(MNodeColourValues)-2 do begin
    Plot.Colour := MNodeColourValues[ci];
    Clash := False;
    for i := 0 to Count-1 do
      if (Items[i] <> Plot)
         and (Items[i].Colour = Plot.Colour) then begin
        Clash := True;
        Break;
      end;
    if not Clash then Exit;
  end;
end;

procedure TOverlays.SaveToFile(AFileName: string);
var
  Stream: TStream;
  i: Integer;
begin
  if LowerCase(ExtractFileExt(AFileName)) = '.dat' then begin
    Stream := TFileStream.Create(AFileName, fmCreate);
    try
      for i := 0 to Count-1 do
        if Items[i] is TExperimentalPlot then
        TExperimentalPlot(Items[i]).WriteDataToStream(Stream);
    finally
      Stream.Free;
    end;
  end else
    inherited;
end;

procedure TOverlays.Invert;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    Items[i].Invert;
end;

function TOverlays.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Units for plot.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TOverlays.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Threshold for peak finder.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TOverlays.nChildTypes: Integer;
begin
  Result := {$IFDEF CONSOLE} 1 {$ELSE} 2 {$ENDIF};
end;

function TOverlays.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case AIndex of
    0: Result := TExperimentalPlot;
    {$IFNDEF CONSOLE}
    1: Result := TBitmapPlot;
    {$ENDIF}
  else
    Result := nil;
  end;
end;

procedure TOverlays.CycleUnits;
begin
  { Note - just change top level - do not change units of individual plots }
  if Units = High(TResonanceUnits) then
    Units := Low(TResonanceUnits)
  else
    Units := Succ(Units);
end;

function TOverlays.nVisible: Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to Count-1 do
    if not Items[i].Hide then
      Inc(Result);
end;

end.

