unit SpectrumPlot;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Graphs, GraphRange, {$ENDIF}
     Units, Graphics, DOM, MNode, Mixture, Classes, Life, SysUtils;

type

  TAbstractPlot = class(TSpectrum)
  private
    FFOffset: Double;
    FRefWidth: Double;
    FLifeModel: TLifeModel;
  protected
    procedure AdjustWidth(var Width, Strength: Double);
  public
    IMin, IMax: Double; { Vertical limits used in plotting }
    procedure InitializeFromMixture(Mixture: TMixture); virtual;
    procedure FindLimits; virtual;
    function MakeFromMixture(Mixture: TMixture):Boolean;
    property FOffset: Double read FFOffset write FFOffset;
    property RefWidth: Double read FRefWidth write FRefWidth;
    property LifeModel: TLifeModel read FLifeModel write FLifeModel;
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; var PlotRanges:TPlotRanges; Offset: Boolean = False; YLabel: string = ''); overload; virtual;
    procedure Show(G: TGraphD; AImin, AIMax: Double; YLabel: string; var PlotRanges:TPlotRanges) overload;  virtual; abstract;
    {$ENDIF}
  end;

  { TSpectrumPlot }

  TSpectrumPlot = class(TAbstractPlot)
  private
    FIntensityInRange: Double;
    FnDF: Integer;
    FGaussian: Double;
    FLorentzian: Double;
    FnLinesInRange: Integer;
    FSMargin: Double;
    LineShape: TDoubleArray;
    FWidthMult: Integer;
    FConvoluted: Boolean;
    procedure SetGaussian(const Value: Double);
    procedure SetLorentzian(const Value: Double);
    procedure SetSMargin(const Value: Double);
    procedure SetWidthMult(const Value: Integer);
  protected
    OThreshold: Double;
    procedure SetnDF(const Value: Integer); virtual;
  public
    WidthSet: Boolean;
    Plot: TDoubleArray; { The Spectrum }
    procedure InitializeFromMixture(Mixture: TMixture); override;
    property nDF: Integer read Fndf write Setndf; { Number of points to simulate spectrum at }
    property Lorentzian: Double read FLorentzian write SetLorentzian;
    property Gaussian: Double read FGaussian write SetGaussian;
    property SMargin: Double read FSMargin write SetSMargin;
    property WidthMult: Integer read FWidthMult write SetWidthMult;
    property Convoluted: Boolean read FConvoluted;
    { TODO: Tau, Sine / Generalize }
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits); overload; override;
    procedure AddLine(const Line: TLine; Mixture: TMixture); overload; override;
    procedure Clear; override;
    procedure Convolute;
    procedure FindLimits; override;
    procedure MakeLineShape(Lor, Gau: Double);
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; AImin, AIMax: Double; YLabel: string; var PlotRanges:TPlotRanges); overload; override;
    {$ENDIF}
    property nLinesInRange: Integer read FnLinesInRange;
    property IntensityInRange: Double read FIntensityInRange;
  end;

  TMultiColouredSpectrumPlot = class(TSpectrumPlot)
  private
    FMixture: TMixture;
    Plots: array of TSpectrumPlot;
  public
    constructor Create(AMixture: TMixture);
    destructor Destroy; override;
    procedure Clear; override;
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits); override;
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; var PlotRanges:TPlotRanges; Offset: Boolean = False; YLabel: string = ''); override;
    {$ENDIF}
  end;

  TFortratPlot = class(TAbstractPlot)
  private
    JMin, JMax: Integer;
    Stream: TMemoryStream;
    MaxIntensity: Double;
    MaxSym: Integer;
    MaxDeltaJ: Integer;
    ColoursUsed: array of TColor;
    MoleculeType: TMoleculeClass; { If only one molecue, use that for symmetry labels }
    SampleMolecule: TMolecule;
  public
    MinI: Double;
    ShowSymmetry: Boolean;
    ShowDeltaJ: Boolean;
    ScaleMarkSize: Boolean;
    UseUpper: Boolean;
    destructor Destroy; override;
    procedure Clear; override;
    procedure AddLine(const Line: TLine; Mixture: TMixture); override;
    procedure InitializeFromMixture(Mixture: TMixture); override;
    {$IFNDEF CONSOLE}
    procedure Show(G: TGraphD; var PlotRanges:TPlotRanges; Offset: Boolean = False; YLabel: string = ''); override;
    {$ENDIF}
  end;

  TFitPlot = class(TSpectrumPlot)
  private
    FnDFLockout: Boolean;
  protected
    procedure SetnDF(const Value: Integer); override;
  public
    MaxIntensity: Double;
    Frequency: TDoubleArray; { The required frequencies }
    procedure Clear; override;
    procedure AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits); override;
    procedure InitializeFromMixture(Mixture: TMixture); override;
  end;


implementation

uses {$IFNDEF CONSOLE} Gstyle, Gmarks, ClipBrd, Huge, {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$ENDIF}
     Progress, TextStrm, Math, MathSup, Wofz, QuantumNumber;

{ TAbstractPlot }

procedure TAbstractPlot.FindLimits;
begin
  { Would be abstract, only no action might well be OK }
end;

procedure TAbstractPlot.InitializeFromMixture(Mixture: TMixture);
begin
  FOffset := Mixture.FOffset;
  FMin := Mixture.FMin;
  FMax := Mixture.FMax;
  Temperature := Mixture.Temperature;
  Tvib := Mixture.Tvib;
  RefWidth := Mixture.RefWidth;
  LifeModel := Mixture.LifeModel;
  Units := Mixture.PlotUnits;
  Clear;
end;

function TAbstractPlot.MakeFromMixture(Mixture: TMixture):Boolean;
begin
  PushProgress('Simulating Spectrum');
  if not Assigned(Mixture.LineList) then
    Result := Mixture.MakeLineList <> nil
  else
    Result := True;
  InitializeFromMixture(Mixture);
  AddIn(Mixture.LineList, Mixture);
  PopProgress;
end;

{$IFNDEF CONSOLE}
procedure TAbstractPlot.Show(G: TGraphD; var PlotRanges:TPlotRanges; Offset: Boolean = False; YLabel: string = '');
begin
  FindLimits;
  Show(G, IMin, IMax, YLabel, PlotRanges);
end;
{$ENDIF}

procedure TAbstractPlot.AdjustWidth(var Width, Strength: Double);
begin
  case LifeModel of
    lmProduct, lmProductWidth:
      { Scale Peak intensity as width/(width + RefWidth) }
      if RefWidth > 0 then
        Strength := Strength * Width / (Width + RefWidth)
      else
        Strength := Strength * Width;
    lmParent, lmParentWidth:
      { Scale Peak intensity as 1/(width + RefWidth) }
      if Width + RefWidth > 0 then
        Strength := Strength / ( Width + RefWidth);
    lmGate:
      { Scale Peak intensity as exp(-width * RefWidth) }
      Strength := Strength * Exp(-Width*RefWidth);
  end;
  if LifeModel in [lmNone, lmProduct, lmParent, lmGate] then
    Width := 0; { Discard width if rerquired }
end;

{ TSpectrumPlot }

procedure TSpectrumPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  Centre, nl, i: Integer;
  df, a, w1: Double;
begin
  AdjustWidth(AWidth, AStrength);
  AWidth := ConvertWidth(AWidth, AFrequency, AUnits, Units);
  AFrequency := Convert(AFrequency, AUnits, Units);
  df := (FMax-FMin)/(nDF-1);
  a := (AFrequency - FMin)/df;
  { Prevent overflows on conversion to integer }
  if Abs(a) < High(Integer) div 2 then
    Centre := Round(a)
  else
    Centre := High(Integer) div 2;
  if AWidth > Abs(3*df) then begin
    { Add in Lorentzian to plot }
    nl := Round(WidthMult * 2 * AWidth / Abs(df) + 4);
    if (Centre >= -nl) and (Centre-nl < nDF) then begin
      WidthSet := True;
      a := AStrength * AWidth / 2 / Pi / Abs(df) { / df };
      w1 := Sqr(AWidth) / 4 / df / df;
      for i := Max(Centre-nl,0) to Min(Centre+nl,nDF-1) do
        Plot[i] := Plot[i] + a/(Sqr(i-Centre) + w1);
    end else
      Exit;
    WidthSet := True;
  end else if (Centre >= 0) and (Centre < nDF) then
    Plot[Centre] := Plot[Centre] + AStrength { /Abs(df) } { TODO: Check normalization compared to above }
  else
    Exit;
  Inc(FnLinesInRange);
  FIntensityInRange := FIntensityInRange + AStrength;
end;

procedure TSpectrumPlot.AddLine(const Line: TLine; Mixture: TMixture);
begin
  AddLine(Line.Frequency, Mixture.LineIntensity(Line, Temperature, Tvib), Line.Width, Line.Colour, Mixture.Units);
end;

procedure TSpectrumPlot.Clear;
begin
  inherited;
  FConvoluted := False;
  if Length(Plot) > 0 then
    FillChar(Plot[0], Length(Plot)*SizeOf(Double), 0);
  WidthSet := False;
  FnLinesInRange := 0;
  FIntensityInRange := 0;
end;

procedure TSpectrumPlot.Convolute;
var
  x: TDoubleArray;
  i, k, nLw: Integer;
begin
  if Convoluted then Exit;
  MakeLineShape(Lorentzian, Gaussian);
  nLw := (Length(LineShape)-1) div 2;
  SetLength(x, Length(Plot));
  FillChar(x[0], Length(Plot)*SizeOf(Double), 0);
  for i := 0 to High(Plot) do
    if Plot[i] <> 0 then
      for k := Max(i-nLw,0) to Min(i+nLw, High(Plot)) do
        x[k] := x[k] + Plot[i]*LineShape[nLW+k-i];
  for i := 0 to High(Plot) do
    Plot[i] := x[i];
  FConvoluted := True;
end;

procedure TSpectrumPlot.FindLimits;
begin
  Convolute;
  Limits(Plot, IMin, IMax);
  If Imax = IMin then
    IMax := IMin + 1; { Flat spectrum - quick fix to avoid divide by zero }
end;

procedure TSpectrumPlot.InitializeFromMixture(Mixture: TMixture);
begin
  inherited;
  OThreshold := Mixture.OThreshold;
  SMargin := Mixture.SMargin;
  Lorentzian := Mixture.Lorentzian;
  Gaussian := Mixture.Gaussian;
  WidthMult := Mixture.WidthMult;
  nDF := Mixture.nDF;
end;

procedure TSpectrumPlot.MakeLineShape(Lor, Gau: Double);
var
  n, i: Integer;
  df: Double;
begin
  df := (FMax-FMin)/(nDF-1);
  if Lor+Gau <= Abs(3*df) then begin
   SetLength(LineShape, 1);
   LineShape[0] := 1/ABS(df)
  end else begin
{  TODO: IF (sine .NE. 0) THEN
     n = MIN((dfmult*(2*lor+gau)+ABS(sine))/ABS(df)+4, DBLE(nmax))
   ELSE }
   n := Round(WidthMult*(lor+gau)/ABS(df)+4);
   SetLength(LineShape, 2*n+1);
   for i := 0 to n do begin
     LineShape[i+n] := Voigt(i*df, Lor, Gau);
     LineShape[n-i] := LineShape[i+n];
   end;
   WidthSet := True;
  end;
end;

procedure TSpectrumPlot.SetGaussian(const Value: Double);
begin
  FGaussian := Value;
end;

procedure TSpectrumPlot.SetLorentzian(const Value: Double);
begin
  FLorentzian := Value;
end;

procedure TSpectrumPlot.SetnDF(const Value: Integer);
begin
  FnDF := Value;
  SetLength(Plot, nDF);
end;

procedure TSpectrumPlot.SetSMargin(const Value: Double);
begin
  FSMargin := Value;
end;

procedure TSpectrumPlot.SetWidthMult(const Value: Integer);
begin
  FWidthMult := Value;
end;

{$IFNDEF CONSOLE}
procedure TSpectrumPlot.Show(G: TGraphD; AImin, AIMax: Double; YLabel: string; var PlotRanges:TPlotRanges);
var
  i: Integer;
  df, x, s: Double;
begin
  Convolute;
  if Colour <> clNone then
    G.StyleP(Colour, G.LineStyle.Width, G.LineStyle.Pattern);
  df := (FMax-FMin)/(nDF-1);
  s := 1/(AIMax-AIMin);
  x := FMin-Foffset;
  if WidthSet then begin
    G.Point(x, s*(Plot[0]-AIMin));
    for i := 1 to nDF-1 do begin
      x := x + df;
      G.Join(x, s*(Plot[i]-AIMin));
    end;
  end else begin
    { Special case for stick spectra - make sure lines are vertical and single }
    G.Point(x, -s*AImin);
    for i := 0 to nDF-1 do begin
      if Plot[i] > 0 then begin
        G.Join(x, -s*AImin);
        G.Join(x, s*(Plot[i]-AIMin));
        G.Join(x, -s*AImin);
      end;
      x := x + df;
    end;
    G.Join(x-df, -s*AImin);
  end;
  if YLabel <> '' then
    G.ScaledAxis(AImin, AIMax, 0, 1, G.Xmin, False, YLabel);
  AddPlotRange(PlotRanges, G, 0, 1, AImin, AIMax);
end;
{$ENDIF}

{ TFitPlot }

procedure TFitPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  i: Integer;
  x: Double; { Smallest offset from peak centre }
  y: Double; { Correspondig height }
begin
  AdjustWidth(AWidth, AStrength);
  AWidth := ConvertWidth(AWidth, AFrequency, AUnits, Units);
  AFrequency := Convert(AFrequency, AUnits, Units);
  if AStrength <= MaxIntensity then
    Exit;
  if AFrequency <= FMin then
    x := FMin - AFrequency
  else if AFrequency >= FMax then
    x := AFrequency - FMax
  else
    x := 0;
  y := AStrength*Voigt(x, AWidth + Lorentzian, Gaussian);
  if y <= MaxIntensity then
    Exit;
  if y*OThreshold > MaxIntensity then
    MaxIntensity := y*OThreshold;
  { TODO: Lots of possible optimizations here }
  for i := 0 to High(Plot) do
    Plot[i] := Plot[i] + AStrength*Voigt(Frequency[i]-AFrequency, AWidth + Lorentzian, Gaussian);
end;

procedure TFitPlot.Clear;
begin
  inherited Clear;
  Limits(Frequency, Fmin, FMax);
  MaxIntensity := 0;
end;

procedure TFitPlot.InitializeFromMixture(Mixture: TMixture);
begin
  FnDFLockout := True;
  try
    inherited;
  finally
    FnDFLockout := False;
  end;
end;

procedure TFitPlot.SetnDF(const Value: Integer);
begin
  if FnDFLockout then Exit;
  inherited;
  SetLength(Frequency, nDF);
end;

{ TMultiColouredSpectrumPlot }

procedure TMultiColouredSpectrumPlot.AddLine(AFrequency, AStrength, AWidth: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  i: Integer;
  CurrentPlot: TSpectrumPlot;
begin
  inherited; { Always calculate sum, even if not displayed so that scaling is consistent }
  if AColour = clNone then
    Exit;
  CurrentPlot := nil;
  for i := 0 to High(Plots) do
    if Plots[i].Colour = AColour then begin
      CurrentPlot := Plots[i];
      Break;
    end;
  if not Assigned(CurrentPlot) then begin
    SetLength(Plots, Length(Plots)+1);
    CurrentPlot := TSpectrumPlot.Create(FMixture.PlotUnits);
    CurrentPlot.InitializeFromMixture(FMixture);
    CurrentPlot.Colour := AColour;
    Plots[Length(Plots)-1] := CurrentPlot;
  end;
  CurrentPlot.AddLine(AFrequency, AStrength, AWidth, AColour, AUnits);
end;

procedure TMultiColouredSpectrumPlot.Clear;
var
  i: Integer;
begin
  inherited;
  for i := 0 to High(Plots) do
    Plots[i].Free;
  SetLength(Plots,0);
end;

constructor TMultiColouredSpectrumPlot.Create(AMixture: TMixture);
begin
  inherited Create(AMixture.PlotUnits);
  FMixture := AMixture;
end;

destructor TMultiColouredSpectrumPlot.Destroy;
begin
  Clear;
  inherited;
end;

{$IFNDEF CONSOLE}
procedure TMultiColouredSpectrumPlot.Show(G: TGraphD; var PlotRanges:TPlotRanges; Offset: Boolean = False; YLabel: string = '');
var
  i, nPlots, nP: Integer;
  SubYlabel: string;
  function Lower(nPlot: Integer):Double;
  begin
    Result := IMin;
    if Offset then
      Result := Result - (IMax-IMin)*(nPlot-1)
  end;
  function Upper(nPlot: Integer):Double;
  begin
    if Offset then
      Result := Lower(nPlot) + (IMax-IMin)*nPlots
    else
      Result := IMax;
  end;
begin
  FindLimits;
  { Calculate number of plots }
  nPlots := Length(Plots);
  if FMixture.ShowSum then
    Inc(nPlots);
  if nPlots = 0 then Exit;
  nP := 0;
  for i := 0 to High(Plots) do begin
    Inc(nP);
    SubYlabel := YLabel;
    if not Offset and (FMixture.ShowSum or (i > 0)) then
      SubYlabel := '';
    Plots[i].Show(G, Lower(nP), Upper(nP), SubYlabel, PlotRanges);
  end;
  if FMixture.ShowSum then
    Show(G, Lower(nPlots), Upper(nPlots), YLabel, PlotRanges);
end;
{$ENDIF}

{ TFortratPlot }

type
  TFortratRec = packed record
    Frequency: Double;
    Strength:  Double;
    J:         Integer;
    Colour:    TColor;
    DeltaJ:    Shortint;
    Sym:       Byte;
  end;

procedure TFortratPlot.AddLine(const Line: TLine; Mixture: TMixture);
var
  c: Double;
  FortratRec: TFortratRec;
  Width, f: Double;
  i: Integer;
  Molecule: TMolecule;
begin
  Inc(FnLines);
  f := Convert(Line.Frequency, Mixture.Units, Units);
  c := (f - FMin)/(FMax - FMin);
  if (c>=0) and (c <=1) and Assigned(Line.EigenvaluesL) then begin
    FortratRec.Frequency := f - FOffset;
    if UseUpper then begin
      FortratRec.J := Line.EigenvaluesU.J;
      Molecule := Line.EigenvaluesU.Manifold.Molecule;
      FortratRec.Sym := Molecule.JAdjustedSym(Line.EigenvaluesU.J, Line.EigenvaluesU.Sym);
    end else begin
      FortratRec.J := Line.EigenvaluesL.J;
      Molecule := Line.EigenvaluesL.Manifold.Molecule;
      FortratRec.Sym := Molecule.JAdjustedSym(Line.EigenvaluesL.J, Line.EigenvaluesL.Sym);
    end;
    if MoleculeType = nil then begin
      MoleculeType := TMoleculeClass(Molecule.ClassType);
      SampleMolecule := Molecule;
    end else if Assigned(SampleMolecule) and (MoleculeType <> Molecule.ClassType) then
      SampleMolecule := nil;
    FortratRec.DeltaJ := Line.EigenvaluesU.J - FortratRec.J;
    if Abs(FortratRec.DeltaJ) > MaxDeltaJ then MaxDeltaJ := Abs(FortratRec.DeltaJ);
    FortratRec.Colour := Line.Colour;
    if FortratRec.Colour = clNone then
      FortratRec.Colour := clBlack;
    if MaxSym < FortratRec.Sym then
      MaxSym := FortratRec.Sym;
    if not Assigned(Mixture) then
      Mixture := Line.EigenvaluesL.Manifold.Mixture;
    FortratRec.Strength := Mixture.LineIntensity(Line, Temperature, Tvib);
    Width := Line.Width;
    AdjustWidth(Width, FortratRec.Strength);
    // Not needed as only Strength required: Width := ConvertWidth(Width, Line.Frequency, Mixture.Units, Units);
    Stream.Write(FortratRec, SizeOf(FortratRec));
    if (Jmax < 0) or (JMin > FortratRec.J) then begin
      JMin := FortratRec.J;
      IMin := FortratRec.J / 2.;
    end;
    if FortratRec.J > JMax then begin
      JMax := FortratRec.J;
      IMax := JMax / 2.;
    end;
    if MaxIntensity < FortratRec.Strength then
      MaxIntensity := FortratRec.Strength;
    for i := 0 to High(ColoursUsed) do
      if FortratRec.Colour = ColoursUsed[i] then
        Exit;
    SetLength(ColoursUsed, Length(ColoursUsed)+1);
    ColoursUsed[Length(ColoursUsed)-1] := FortratRec.Colour;
  end;
end;

procedure TFortratPlot.Clear;
begin
  inherited;
  IMin := 0;
  IMax := 0;
  JMax := -1;
  MaxIntensity := 0;
  MaxDeltaJ := 0;
  MaxSym := -1;
  Stream.Free;
  Stream := TMemoryStream.Create;
  SetLength(ColoursUsed, 0);
  MoleculeType := nil;
  SampleMolecule := nil;
end;

destructor TFortratPlot.Destroy;
begin
  Clear;
  FreeAndNil(Stream);
  inherited;
end;

procedure TFortratPlot.InitializeFromMixture(Mixture: TMixture);
begin
  inherited;
  MinI := Mixture.MinI;
  ShowSymmetry := Mixture.ShowSymmetry;
  ShowDeltaJ := Mixture.ShowDeltaJ;
  ScaleMarkSize := Mixture.ScaleMarkSize;
  UseUpper := Mixture.UseUpper;
end;

{$IFNDEF CONSOLE}
procedure TFortratPlot.Show(G: TGraphD; var PlotRanges:TPlotRanges; Offset: Boolean = False; Ylabel: string = '');
var
  FortratRec: TFortratRec;
  MarkStyle: TMarkStyle;
  BaseSize: Single;
  i, m, iSym, MarkNumber, nPlots: Integer;
  o: Double;
  s: string;
  ColourOffset: array of Double;
  function SymLabel(Sym: Integer): string;
  begin
    if Assigned(SampleMolecule) then begin
      Result := SampleMolecule.JAdjustedSymName(0,Sym);
      if Result <> IntToStr(Sym) then
        Result := IntToStr(Sym) + '/' + Result;
    end else
      Result := IntToStr(Sym);
  end;
begin
  if nLines = 0 then
    Exit;
  m := 0;
  if ShowSymmetry then begin
    s := 'Symmetry: ';
    for iSym := 0 to MaxSym do
      if ShowDeltaJ then
        for i := -MaxDeltaJ div 2 to MaxDeltaJ div 2 do begin
          s := s + #1 + Char(Ord(FirstMark)+m) + #0':' + SymLabel(iSym) + DeltaJName(i*2) + ' ';
          Inc(m);
        end
      else
        s := s + #1 + Char(Ord(FirstMark)+iSym) + #0':' + SymLabel(iSym) + ' ';
  end else if ShowDeltaJ then begin
    s := 'DeltaJ: ';
    for i := -MaxDeltaJ div 2 to MaxDeltaJ div 2 do begin
      s := s + #1 + Char(Ord(FirstMark)+m) + #0':' + DeltaJName(i*2) + ' ';
      Inc(m);
    end;
  end;
  G.Title2(s);
  Stream.Position := 0;
  MarkStyle := TMarkStyle.Create;
  MarkStyle.DotAtCentre := True;
  MarkStyle.Size := 2;
  BaseSize := G.MarkStyles[0].Size;
  nPlots := 1;
  if Offset then
    nPlots := Length(ColoursUsed);
  SetLength(ColourOffset, nPlots);
  o := G.Ymin;
  for i := 0 to High(ColourOffset) do begin
    ColourOffset[i] := o;
    o := o + (G.Ymax-G.Ymin)/nPlots;
    G.ScaledAxis(IMin, G.Ymax, ColourOffset[i] + (IMin-G.Ymin)/nPlots, o, G.Xmin, False, 'J');
    AddPlotRange(PlotRanges, G, ColourOffset[i] + (IMin-G.Ymin)/nPlots, o, IMin, G.Ymax);
  end;
  o := G.Ymin;
  while Stream.Read(FortratRec, SizeOf(FortratRec)) = SizeOf(FortratRec) do
    with FortratRec do
      if Strength >= MinI*MaxIntensity then begin
        MarkStyle.Colour := Colour;
        if ShowDeltaJ then begin
          MarkNumber := (DeltaJ + MaxDeltaJ) div 2;
          if ShowSymmetry then
            Inc(MarkNumber, Sym * (MaxDeltaJ + 1));
        end else if ShowSymmetry then
          MarkNumber := Sym
        else
          MarkNumber := 0;
        MarkStyle.MarkType := TMarkType(Ord(FirstMark) + MarkNumber mod (Ord(LastMark)-Ord(FirstMark)+1));
        if ScaleMarkSize and (MaxIntensity > 0) then
          MarkStyle.Size := 2 * BaseSize * Sqrt(Strength / MaxIntensity);
        if nPlots > 1 then
          for i := 0 to High(ColoursUsed) do
            if Colour = ColoursUsed[i] then begin
              o := ColourOffset[i];
              Break;
            end;
        G.SymbolT(Frequency, (J/2.-G.Ymin)/nPlots+o, MarkStyle);
      end;
  MarkStyle.Free;
end;
{$ENDIF}

end.
