unit HundA;

{ Hund's Case (a) basis functions }

interface
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

uses SysUtils, Classes, MNode, Mixture, QuantumNumber;

type

  TLinearMolecule = class;

  TLinear = class(TState)
  private
    FMolecule: TLinearMolecule;
    FInverted: Boolean;
  public { Functions and properties specific to TLinear }
    function HRot(J, Lamp, Sigp, Omp, Lam, Sig, Om: TQuantumNumber): Double;
    function NRn(n: Integer; J, Lam, Spin, Omp, Om: TQuantumNumber): Double;
    property Molecule: TLinearMolecule read FMolecule;
    property Inverted: Boolean read FInverted;
  public
    constructor Create(AOwner: TComponent); override;
    class function OwnerType: TComponentClass; override;
    class function NucleusType: TNucleusClass; override;
    function FixupProperties: Exception; override;
    function OddSpin: TQuantumNumber; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer; override;
    function GetSettingsNames(AIndex: Integer; var LowestValue:Integer): PPChar; override;
    function HyperfineMatrixElement(A, B: TBasisState; Signp, Sign: Integer; MNode: TMNode; F: TQuantumNumber): Double; override;
    function RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double; override;
    function RotationalTransitionMatrixElement(A, B: TBasisState; SignA, SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber):Double; override;
    function RotationalPerturbationMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Double; override;
    function StateOrigin: Double; override; { Estimate lowest J=0 energy - Used to stop Boltzman factors blowing up at low temperatures }
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure ConvertToPerturbations; override;
  published
    property B:        Double index  LSP   read GetParameter write SetParameter;
    property A:        Double index  LSP+1 read GetParameter write SetParameter;
    property LambdaSS: Double index  LSP+2 read GetParameter write SetParameter;
    property o:        Double index  LSP+3 read GetParameter write SetParameter;
    property theta:    Double index  LSP+4 read GetParameter write SetParameter; { S >= 2 only }
    property Gamma:    Double index  LSP+5 read GetParameter write SetParameter;
    property p:        Double index  LSP+6 read GetParameter write SetParameter;
    property q:        Double index  LSP+7 read GetParameter write SetParameter;
    { Centrifugal Distortion }
    property D:        Double index  LSP+8 read GetParameter write SetParameter;
    property H:        Double index  LSP+9 read GetParameter write SetParameter;
    property L:        Double index  LSP+10 read GetParameter write SetParameter;
    property M:        Double index  LSP+11 read GetParameter write SetParameter;
    property PP: { P } Double index  LSP+12 read GetParameter write SetParameter;
    { Centrifugal Distortion of A }
    property AD:       Double index LSP+13 read GetParameter write SetParameter;
    { Centrifugal Distortion of Lambda }
    property LambdaD:  Double index  LSP+14 read GetParameter write SetParameter;
    property LambdaH:  Double index  LSP+15 read GetParameter write SetParameter;
    { Centrifugal Distortion of lambda doubling }
    property oD:       Double index  LSP+16 read GetParameter write SetParameter;
    property oH:       Double index  LSP+17 read GetParameter write SetParameter;
    property oL:       Double index  LSP+18 read GetParameter write SetParameter;
    property pD:       Double index  LSP+19 read GetParameter write SetParameter;
    property pH:       Double index  LSP+20 read GetParameter write SetParameter;
    property pL:       Double index  LSP+21 read GetParameter write SetParameter;
    property qD:       Double index  LSP+22 read GetParameter write SetParameter;
    property qH:       Double index  LSP+23 read GetParameter write SetParameter;
    property qL:       Double index  LSP+24 read GetParameter write SetParameter;
    { Centrifugal distortion of gamma }
    property gammaD:   Double index  LSP+25 read GetParameter write SetParameter;
    property gammaH:   Double index  LSP+26 read GetParameter write SetParameter;
    property gammaL:   Double index  LSP+27 read GetParameter write SetParameter;

    property Lambda:    TQuantumNumber  index LSI read GetSetting write SetSetting; { Use -1 for sigma- }
    property gerade:    Boolean         index LSI+1 read GetFlag write SetFlag;
  end;

  TLinearNucleus = class(TNucleus)
  public
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property a:        Double index  1 read GetParameter write SetParameter;
    property b:        Double index  2 read GetParameter write SetParameter;
    property c:        Double index  3 read GetParameter write SetParameter;
    property d:        Double index  4 read GetParameter write SetParameter;
    property eQq0:     Double index  5 read GetParameter write SetParameter;
    property eQq2:     Double index  6 read GetParameter write SetParameter;
    property cI:       Double index  7 read GetParameter write SetParameter;
    { Centrifugal distortion of the above.
      Note that these only have an outline implementation, and may
      not be the same as other programs }
    property aJ:       Double index  8 read GetParameter write SetParameter;
    property bJ:       Double index  9 read GetParameter write SetParameter;
    property cJ:       Double index 10 read GetParameter write SetParameter;
    property dJ:       Double index 11 read GetParameter write SetParameter;
    property eQq0J:    Double index 12 read GetParameter write SetParameter;
    property eQq2J:    Double index 13 read GetParameter write SetParameter;
  end;

  { TLinearBasisState }

  TLinearBasisState = class(TBasisState)
  public
    Omega: TQuantumNumber;
    Fn: Integer; { Indicates F1, F2, F3 .... Not guaranteed for J < S }
    function N: TQuantumNumber; { Best guess at N - computed from Fn }
    function Parity: Integer;
    function S: TLinear;
    function efLabel: string;

    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel: string; override; { State name that goes with the given basis state
                                            (assuming match between basis state and final state is OK) }
    function StateLabelKey: string; override; { Explanation of State Name }
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as R(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
  end;

  TLinearManifold = class(TManifold)
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
  public
    class function OwnerType: TComponentClass; override;
    function FixupProperties: Exception; override;
    function FirstSym(J: TQuantumNumber):Integer; override;
    function NextSym(J: TQuantumNumber; Sym: Integer):Integer; override;
    function nSym: Integer; override; { Number of different symmetries }
  end;

  TLinearMolecule = class(TMolecule)
  public
    constructor Create(AOwner: TComponent); override;
    class function ManifoldType: TMNodeClass; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    function SymName(Sym: Integer):string; override;
    function JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Symmetric: Boolean index LMI   read GetFlag write SetFlag; { True for D infinity h }
    { Statistical Weights for symmetric molecules }
    property SymWt:     Integer index LMI+1 read GetSetting write SetSetting;
    property AsymWt:    Integer index LMI+2 read GetSetting write SetSetting;
    { If true, used the R^2 Hamiltonian rather than the default N^2 form }
    property RSquaredH: Boolean index LMI+3 read GetFlag write SetFlag;
    { What quantum numbers to (try and) show }
    property ShowJ:       Boolean index LMI+4 read GetFlag write SetFlag;
    property ShowOmega:   Boolean index LMI+5 read GetFlag write SetFlag;
    property ShowN:       Boolean index LMI+6 read GetFlag write SetFlag;
    property ShowFNumber: Boolean index LMI+7 read GetFlag write SetFlag;
    property Showef:      Boolean index LMI+8 read GetFlag write SetFlag;
  end;

  TLinearOperator = (LS, Luncouple, Homog, aL, e, eQq1);

  TLinearPerturbation = class (TPerturbation)
  private
    procedure SetOperator(const Index: Integer; const AValue: TLinearOperator);
    function GetOperator(const Index: Integer): TLinearOperator;
  public
    constructor Create(AOwner: TComponent); override;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Op:TLinearOperator index LPI read GetOperator write SetOperator;
    property n:Integer         index LPI+1 read GetSetting write SetSetting;
    property Npower:Integer    index LPI+2 read GetSetting write SetSetting;
  end;

implementation

uses Math, Life, nj;

{ TLinearBasisState }

function TLinearBasisState.BasisLabel: string;
begin
  Result := Format('%s %2s', [inherited BasisLabel, State.Manifold.Mixture.QuantumNumberToStr(Omega)]);
end;

function TLinearBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey + ' Omega';
end;

function TLinearBasisState.N: TQuantumNumber;
var
  OmLow, nSkip: TQuantumNumber;
begin
  with S do begin
    if S = 0 then
      Result := J
    else if Lambda > 0 then begin
      if Inverted then
        OmLow := S+Lambda
      else
        OmLow := S-Lambda;
      nSkip := OmLow - Min(OmLow, J);
      Result := Max(Abs(J-S),Lambda) + 2*Fn - 2 - nSkip;
    end else
      { Sigma States: Note that these two expressions are equivalent, unless J < S }
      if Inverted then
        Result := J + S + 2*(Fn - (S+1))
      else
        Result := Abs(J-S) + 2*(Fn-1);
        // Result := Max(Abs(J-S),Lambda) + 2*(Fn-1)
    {$IFOPT R+}
    if not IsTriangle(J, S, Result) then
      raise Exception.CreateFmt('TLinearBasisState.N: %d %d %d', [J, S, Result]);
    {$ENDIF}
  end;
end;

function TLinearBasisState.Parity: Integer;
begin
  Result := 1 - Sym*2;
end;

function TLinearBasisState.S: TLinear;
begin
  Result := State as TLinear;
end;

function TLinearBasisState.efLabel: string;
begin
  if ((J and 2) <> 0) xor (Sym = 1) then
    Result := 'f'
  else
    Result := 'e';
end;

function TLinearBasisState.StateLabel: string;
var
  Molecule: TLinearMolecule;
begin
  Molecule := State.Manifold.Molecule as TLinearMolecule;
  Result := inherited StateLabel;
  if Molecule.ShowJ then
    Result := Result + Format(' %2s', [Molecule.Mixture.QuantumNumberToStr(J)]);
  if Molecule.ShowN then
    Result := Result + Format(' %2s', [Molecule.Mixture.QuantumNumberToStr(N)]);
  if Molecule.ShowOmega then
    Result := Result + Format(' %2s', [Molecule.Mixture.QuantumNumberToStr(Omega)]);
  if Molecule.ShowFNumber then
    Result := Result + Format(' F%d', [Fn]);
  if Molecule.Showef then begin
    if not Molecule.ShowFNumber then
      Result := Result + ' ';
    Result := Result + efLabel;
  end;
  Result := Result + HyperfineStateLabel;
end;

function TLinearBasisState.StateLabelKey: string;
var
  Molecule: TLinearMolecule;
begin
  Molecule := State.Manifold.Molecule as TLinearMolecule;
  Result := inherited StateLabelKey;
  if Molecule.ShowJ then
    Result := Result + '  J';
  if Molecule.ShowN then
    Result := Result + '  N';
  if Molecule.ShowOmega then
    Result := Result + ' Om';
  if Molecule.ShowFNumber then
    Result := Result + ' Fn';
  if Molecule.Showef then
    Result := Result + ' p';
  Result := Result +  HyperfineStateLabelKey;
end;

function TLinearBasisState.TxName(Upper, Lower: TBasisState): string;
var
  U, L: TLinearBasisState;
  US, LS: TLinear;
  Molecule: TLinearMolecule;
begin
  Molecule := State.Manifold.Molecule as TLinearMolecule;
  U := Upper as TLinearBasisState;
  US := U.State as TLinear;
  L := Lower as TLinearBasisState;
  LS := L.State as TLinear;
  Result := DeltaJName(U.J-L.J);
  if Molecule.ShowN then
    Result := LowerCase(DeltaJName(U.N-L.N)) + Result;
  if (US.S > 0) or (LS.S > 0) then begin
    if U.Fn <> L.Fn then
      Result := Result + IntToStr(U.Fn);
    Result := Result + IntToStr(L.Fn);
  end;
  if (US.Lambda > 0) and (LS.Lambda > 0) then
    Result := Result + L.efLabel;
  Result := Result + Format('(%2s)',[Molecule.QuantumNumberToStr(L.J)]);
  { TODO: Add hyperfine }
end;

function TLinearBasisState.TxNameKey(Upper, Lower: TBasisState): string;
var
  U, L: TLinearBasisState;
  US, LS: TLinear;
  Molecule: TLinearMolecule;
begin
  Molecule := State.Manifold.Molecule as TLinearMolecule;
  U := Upper as TLinearBasisState;
  US := U.State as TLinear;
  L := Lower as TLinearBasisState;
  LS := L.State as TLinear;
  Result := 'deltaJ';
  if Molecule.ShowN then
    Result := 'deltaN ' + Result;
  if (US.S > 0) or (LS.S > 0) then begin
    Result := Result + ' (F'') F"';
  end;
  if (US.Lambda > 0) and (LS.Lambda > 0) then
    Result := Result + ' e/f"';
  Result := Result + '(J")';
end;

function TLinearBasisState.TrackIndex: Integer;
begin
  Result := Fn-1;
end;

{ TLinear }

constructor TLinear.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TLinearMolecule;
  SavedProperties := 'gerade=True,' + SavedProperties;
  inherited;
end;

class function TLinear.OwnerType: TComponentClass;
begin
  Result := TLinearManifold;
end;

function TLinear.OddSpin: TQuantumNumber;
begin
  Result := (S + Molecule.MaxNuclearSpin) and 1;
end;

function TLinear.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer;
var
  Sigma, Dir, Nmin: TQuantumNumber;
  FNumber, FNumberInc, StartIndex, nAdded, i: Integer;
  procedure Add(Omega: TQuantumNumber);
  var
    BasisState: TLinearBasisState;
  begin
    if J < Abs(Omega) then Exit;
    BasisState := TLinearBasisState.Create(Eigenvalues);
    BasisState.State := Self;
    BasisState.J := J;
    BasisState.Omega := Omega;
    BasisState.HyperfineState := HyperfineState;
    BasisState.StateIndex := Eigenvalues.LastStateIndex;
    if (Lambda <= 0) and (Omega = 0) then
      BasisState.Phase := 0
    else begin
      BasisState.Phase := Minus(J-S+Eigenvalues.Sym*2);
      if Lambda < 0 then
        BasisState.Phase := - BasisState.Phase;
    end;
  end;
begin
  { Statistical Weight }
  if Molecule.Symmetric then begin
    if gerade xor (Eigenvalues.Sym > 0) then
      Result := Molecule.SymWt
    else
      Result := Molecule.AsymWt;
  end else
    Result := 1;
  if Result = 0 then
    Exit;
  MatchOdd(S, J);
  StartIndex := Eigenvalues.Size;
  if Lambda <= 0 then begin
    { Sigma states }
    { In general have N = J-S, J-S+1, ... J+S.
                 and Fn = F1,  F2,    ... F(2S+1) }
    Nmin := Abs(J-S);
    { However, for any parity, only Even or Odd N will be present }
    if (Minus(Nmin) < 0) xor (Eigenvalues.Sym = 1) xor (Lambda < 0) then
      Inc(Nmin,2);
    Sigma := SmallestProjection(S);
    if not Odd(S) and ((Lambda < 0) xor (Minus(J-S) < 0) xor (Eigenvalues.Sym = 1)) then
      { Even multiplicity - may be missing Sigma = 0 }
      Sigma := 2;
    nAdded := (S-Sigma) div 2 + 1; { The number added for all but the lowest J }
    while Sigma <= S do begin
      Add(Sigma);
      Inc(Sigma, 2);
    end;
    { In general, lowest F number will be 1 or 2 - work it out from lowest N }
    FNumber := (Nmin-J-S) div 2 + S + 1;
    { For the case where not all 2S+1 N values are present,
      the above will omit the lowest F numbers; this is correct if we
      want to strictly derive the F number from N. This gives an odd
      correlation with energy though; for regular states the missing levels
      are the higher energy levels, which correspond to the higher F numbers. }
    if not Inverted and (nAdded > Eigenvalues.Size - StartIndex) { Some missing } then
       { Want F1 or F2 as the lowest F number }
       FNumber := 2 - FNumber mod 2;
    FNumberInc := 2;
  end else begin
    { As a speed up, try adding in rising energy order }
    if Inverted then
      Dir := -1
    else
      Dir := 1;
    Sigma := -Dir*S - 2*Dir;
    FNumber := 1;
    FNumberInc := 1;
    { Note special case here - normally have
      F1, F2, ... Fn
      and loose either the lowest or the highest F numbers for J < Lambda + S.
      However, For S > Lambda actually loose some at both ends }
    repeat
      Inc(Sigma, 2*Dir);
      Add(Lambda+Sigma);
      nAdded := Eigenvalues.Size - StartIndex;
      if nAdded < 1 then
        Inc(FNumber);
    until Sigma = Dir*S;
  end;
  Eigenvalues.SortFrom(StartIndex);
  for i := StartIndex to Eigenvalues.Size-1 do begin
    (Eigenvalues.Basis[i+1] as TLinearBasisState).Fn := FNumber;
    Inc(FNumber, FNumberInc);
  end;
  Inc(Eigenvalues.LastStateIndex);
end;

{ SetupRmat calculates the matrices of even powers of R or N in rn.
  To allow efficient use of these, the most recent result is cached }

const
  MaxRmatPower = 12; { Maximum power of R/N to calculate }
type
  TRMat = array of array of Double;
threadvar
  JCurrent, LambdaCurrent, SCurrent:TQuantumNumber;
  RSquaredHCurrent, RMatInitialised: Boolean;
  OmegaMin: TQuantumNumber; { Omega range covered by current matrix }
  nOmegas: Integer;
  rn: array [1..MaxRmatPower div 2] of TRMat;

procedure MatSymMul(a, b, c: TRMat);
{ Matrix product a = b*c of SYMMETRIC matrices }
var
  i, j, k: Integer;
  Sum: Double;
begin
  for i := 0 to High(a) do
    for j := 0 to i do begin
      Sum := 0;
      for k := 0 to High(a) do
        Sum := Sum + b[i,k]*c[j,k];
      a[j,i] := Sum;
      a[i,j] := Sum;
    end;
end;

procedure SetupRmat(J, Lambda, S: TQuantumNumber; RSquaredH: Boolean);
var
  Omega, OmegaMax: TQuantumNumber;
  i, ii: Integer;
  sjl: Double;
begin
  if not RMatInitialised or (RSquaredHCurrent <> RSquaredH)
     or (JCurrent <> J) or (LambdaCurrent <> Lambda) or (SCurrent <> S) then begin
    { Recalculation needed }
    RMatInitialised := True;
    RSquaredHCurrent := RSquaredH;
    JCurrent := J;
    LambdaCurrent :=Lambda;
    SCurrent := S;
    OmegaMin := Lambda - S;
    if OmegaMin < - J then
      OmegaMin := -J;
    OmegaMax := Lambda + S;
    if OmegaMax > J then
      OmegaMax := J;
    nOmegas := (OmegaMax - OmegaMin) div 2 + 1;
    if { nOmegas > Length(rn[1]) } True { TODO: use expression }then
      for i := 1 to High(rn) do
        SetLength(rn[i], nOmegas, nOmegas);
    { Form matrix of R**2 }
    { Diagonal Elements }
    Omega := OmegaMin;
    sjl := j*(j+2) + s*(s+2);
    if not RSquaredH then
      sjl := sjl + lambda*lambda;
    for i := 0 to nOmegas-1 do begin
      rn[1,i,i] := (sjl - omega*omega - Sqr(omega-lambda)) / 4;
      Inc(Omega,2);
    end;
    { Off diagonal by 1 }
    Omega := OmegaMin + 2;
    for i := 1 to nOmegas-1 do begin
      rn[1,i,i-1] :=
        - Sqrt( (J*(J+2.0) - Omega * (Omega-2))
                *(S*(S+2) - (Omega-Lambda) * (Omega-Lambda-2)) )/4;
      rn[1,i-1,i] := rn[1,i,i-1];
      Inc(Omega,2);
    end;
    { All other elements zero }
    for i := 2 to nOmegas-1 do
      for ii := 0 to i-2 do begin
        rn[1,i,ii] := 0;
        rn[1,ii,i] := 0;
      end;
    { Now generate higher powers }
    for i := 2 to MaxRmatPower div 2 do
      MatSymMul(rn[i],rn[i-1],rn[1]);
  end;
end;

function TLinear.NRn(n: Integer; J, Lam, Spin, Omp, Om: TQuantumNumber): Double;
{ Matrix Elements of N^n or R^n, depending on state of RSquaredH }
var
  Ompp, OmMax: TQuantumNumber;
begin
  if Abs(Om-Omp) = n then
    Result := Minus(n) * rl(J, Omp, Om) * rl(Spin, Lam-Omp, Lam-Om)
  else if (Om = Omp) and (n = 2) then begin
    Result := J*(J+2) + Spin*(Spin+2) - Om*Om - Sqr(Lam-Om);
    if not Molecule.RSquaredH then
      Result := Result + Lam*Lam;
    Result := Result / 4;
  end else if Abs(Omp-Om) < n then begin
    { Handle using recursion. There may be scope for caching results, as is done
      for SetupRmat above }
    Result := 0;
    Ompp := Max(Omp - 2, Lam-Spin);
    OmMax := Min(Omp + 2, Lam+Spin);
    while Ompp <= OmMax do begin
      if Abs(Ompp) <= J then
        Result := Result + NRn(2,J,Lam,Spin,Omp,Ompp)*NRn(n-2,J,Lam,Spin,Ompp,Om);
      Inc(Ompp,2);
    end;
  end else
    Result := 0;
end;

function TLinear.RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double;
var
  Bra, Ket: TLinearBasisState;
  Lam, Lamp, Om, Omp: TQuantumNumber;
begin
  Bra := A as TLinearBasisState;
  Ket := B as TLinearBasisState;
  Result := 0;
  if (Bra.Sym <> Ket.Sym) or (Bra.J <> Ket.J) then
    raise Exception.Create('Bad TLinear.RotationalMatrixElement call');
  if SignA < 0 then begin
    if SignB > 0 then begin
      Bra := TLinearBasisState(B);
      Ket := TLinearBasisState(A);
    end;
    SignA := -SignA;
    SignB := -SignB;
  end;
  if Lambda < 0 then begin
    Lamp := 0;
    Lam := 0;
  end else begin
    Lamp := Lambda*SignA;
    Lam := Lambda*SignB;
  end;
  Omp := Bra.Omega*SignA;
  Om := Ket.Omega*SignB;
  case TMatElemType(Flag) of
    meHrot: Result := HRot(Bra.J, Lamp, Omp-Lamp, Omp, Lam, Om-Lam, Om);
    meWidth: begin { TODO: Add more operators? Make a perturbation type?
                     If adding more operators, adjust HasWidth }
               if (Bra = Ket) and (SignA = SignB) then
                 Result := Width
             end;
  else
    raise NotYet;
  end;
end;

function TLinear.HRot(J, Lamp, Sigp, Omp, Lam, Sig, Om: TQuantumNumber): Double;
{ Matrix elements in an unsymmetrised case (a) basis. }
var
  i, ii, ir, iir, irmax: Integer;
begin
  Result := 0;
  if Lamp < 0 then
    raise Exception.Create('HRot: states out of order');
  if Lamp = Lam then begin
    { Prepare R^2n and the indices into it }
    SetupRmat(J, Lamp, S, Molecule.RSquaredH);
    i := (Omp-OmegaMin) div 2;
    ii := (Om-OmegaMin) div 2;
    if Sig = Sigp then begin
      Result := Origin
                + A*Lamp*sig/4 { ALzSz }
                { 2/3 lambda (3Sz^2-S^2) }
      	        + 2/3 * LambdaSS*(3*Sig*Sig - S*(S+2)) / 4
                { 1/12 theta (35Sz^4-30S^2Sz^2+25Sz^2-6S^2+3S^4) }
                + theta *(35*Sig*Sig*Sig*Sig/16
                          -30*S*(S+2)*Sig*Sig/16
                          +25*Sig*Sig/4 - 6*S*(S+2)/4
                          + 3*S*S*(S+2)*(S+2)/16)/12;
    end;
    { BR^2 - DR^4 + HR^6 + LR^8 + MR^10 + PR^12 }
    Result := Result + B*rn[1,i,ii] - D*rn[2,i,ii]
                     + H*rn[3,i,ii] + L*rn[4,i,ii]
                     + M*rn[5,i,ii] + PP*rn[6,i,ii];
    { 1/2 AD (LzSzR^2 + R^2LzSz) }
    Result := Result + AD*rn[1,i,ii]*(Lam*Sig + Lamp*Sigp)/8;
    { gamma N . S }
    if Sig = Sigp then
      Result := Result + gamma * (Om*Sig - S*(S+2))/4
    else if Abs(Sig-Sigp) = 2 then
      Result := Result + gamma * rl(J,Omp,Om)*rl(S,Sigp,Sig)/2;
    { 1/2 [NzS,R**2 + ...]+ }
    Result := Result + (gammaD*rn[1,i,ii] + gammaH*rn[2,i,ii] + gammaL*rn[3,i,ii])
                       * ((Om*Sig+Omp*Sigp)/2 - S*(S+2))/4;
    { 1/2 N+-R**2 + ... }
    if i > 0 then
      Result := Result + (gammaD*rn[1,i-1,ii] + gammaH*rn[2,i-1,ii] + gammaL*rn[3,i-1,ii])
                         * rl(J,Omp,Omp-2) * rl(S,Sigp,Sigp-2) / 4;
    if i < nOmegas-1 then
      Result := Result + (gammaD*rn[1,i+1,ii] + gammaH*rn[2,i+1,ii] + gammaL*rn[3,i+1,ii])
                         * rl(J,Omp,Omp+2) * rl(S,Sigp,Sigp+2) / 4;
    { 1/2 (R**2 + ...) N+- }
    if ii > 0 then
      Result := Result + (gammaD*rn[1,i,ii-1] + gammaH*rn[2,i,ii-1] + gammaL*rn[3,i,ii-1])
                         * rl(J,Om-2,Om) * rl(S,Sig-2,Sig) / 4;
    if ii < nOmegas-1 then
      Result := Result + (gammaD*rn[1,i,ii+1] + gammaH*rn[2,i,ii+1] + gammaL*rn[3,i,ii+1])
                         * rl(J,Om+2,Om) * rl(S,Sig+2,Sig) / 4;
    { 2/3 lambdaD 1/2 (3Sz^2-s^2) + similar higher terms }
    Result := Result + 2/3 * (LambdaD * rn[1,i,ii] + LambdaH * rn[2,i,ii])
                       * (3*(Sig*Sig+Sigp*Sigp)/2- S*(S+2)) / 4;
  end else if Abs(Lamp-Lam) = 4 then begin
    { Prepare R^2n and the indices into it }
    SetupRmat(J, Lamp, S, Molecule.RSquaredH);
    { To be able to use rn for lambda < 0, must look up opposite
      sign of omega. This assumes lambdap >= 0 }
    { indices to use with lambdap }
    i := (Omp-OmegaMin) div 2;
    { for -lambda=lambdap, sigma, omega+4 }
    ii := (Lamp+Sig-OmegaMin) div 2;
    { indices to use with lambda }
    ir := (-(Lam+Sigp)-OmegaMin) div 2;
    iir := (-Om-OmegaMin) div 2;
    { imax := (OmegaMax-OmegaMin) div 2; = nOmegas-1 }
    irmax := (-Max(-J,Min(J,Lam-S))-OmegaMin) div 2;
    { 1/2 (o+p+q) (S+**2 + S-**2) }
    if Om = Omp then
      Result := Result + (o+p+q)*rl(S,Sigp,Sig) / 2;
    { 1/4 (oD+pD+qD) [R**2, S+**2 + S-**2] }
    { (S+**2 does not contribute as this requires lambdap=lambda-2) }
    { R**2 S+**2
      if ii < nOmegas then
        Result := Result + alphaD*rn[1,i,ii+2]*rl(S, Sig+4, Sig)/4; }
    { (R**2 + R**4 + R**6) S-**2 }
    if (ii > 1) and (ii-1 <= nOmegas) then
      Result := Result + ((oD+pD+qD)*rn[1,i,ii-2] + (oH+pH+qH)*rn[2,i,ii-2] + (oL+pL+qL)*rn[3,i,ii-2])
                         *rl(S, Sig-4, Sig)/4;
    { S-**2 (R**2 + R**4 + R**6) }
    if (ir > 1) and (ir-2 <= irmax) then
      Result := Result + ((oD+pD+qD)*rn[1,ir-2,iir] + (oH+pH+qH)*rn[2,ir-2,iir] + (oL+pL+qL)*rn[3,ir-2,iir])
                         * rl(S,Sigp,Sigp+4) / 4;
    { S+**2 R**2
      if ir+2 < nOmegas then
        Result := Result + alphaD*rl(S,Sigp,Sigp-4)*rn[1,ir+2,iir]/4; }
    { -1/2 (p+2q) (J-S- + J-S-) }
    { Only S- contributes }
    if Sigp = Sig-2 then
      Result := Result - 0.5 * rl(S,Sigp,Sig)*(p+2*q)*rl(J,Omp,Om);
    { -1/2 (pD+2qD) (R**2 + R**4 + R**6) J-S- }
    if (ii > 0) and (ii <= nOmegas) then
      Result := Result - 0.25 * rl(J,Om+2,Om) * rl(S,Sig-2,Sig)
                         * ( (pD+2*qD)*rn[1,i,ii-1] + (pH+2*qH)*rn[2,i,ii-1] + (pL+2*qL)*rn[3,i,ii-1] );
    { -1/2 (pD+2qD) J-S- (R**2 + R**4 + R**6) }
    if (ir > 0) and (ir <= nOmegas) then
      Result := Result - 0.25 * rl(J,Omp,Omp-2) * rl(S,Sigp,Sigp+2)
                         * ( (pD+2*qD)*rn[1,ir-1,iir] + (pH+2*qH)*rn[2,ir-1,iir] + (pL+2*qL)*rn[3,ir-1,iir] );
    { 1/2 q (J+**2 + J-**2) }
    { Only J- contributes }
    if Sig = Sigp then
      Result := Result + 0.5 * q * rl(J, Omp, Om);
    { 1/4 (R**2 + R**4 + R**8) J-**2 }
    if (ii >= 0) and (ii < nOmegas) then
      Result := Result + 0.25 * rl(J,Om+4,Om) * ( qD*rn[1,i,ii] + qH*rn[2,i,ii] + qL*rn[3,i,ii] );
    { 1/4 J-**2 (R**2 + R**4 + R**8) }
    if (ir >= 0) and (ir <= irmax) then
      Result := Result + 0.25 * rl(J,Omp,Omp-4) * ( qD*rn[1,ir,iir] + qH*rn[2,ir,iir] + qL*rn[3,ir,iir] );
  end else if Abs(Lam-Lamp) = 8 then begin
    { Lambda Doubling in a Delta state
      As this rountine is always called with lambdap positive, only need exp(4i phi) term }
    if Sigp = Sig-8 then
      { 1/2 m(S+**4 + S-**4) }
      Result :=  0.5 * o { = m } * rl(S,Sig,SigP)
    else if Sigp = Sig-6 then
      { -1/2 n(S+**3J+ + S-**3J-) }
      Result := - 0.5 * oD { = n } * rl(S,Sig,SigP) * rl(j,Om+2,Om)
    else if Sigp = Sig-4 then
      { 1/2 o(S+**2J+**2 + S-**2J-^**2) }
      Result := 0.5 * oH { = o } * rl(S,Sig,SigP) * rl(j,Om+4,Om)
    else if Sigp = Sig-2 then
      { -1/2 p(S+J+**3 + S-J-**3) }
      Result := - 0.5 * p * rl(S,Sig,SigP) * rl(j,Om+6,Om)
    else if Sigp = Sig then
      { 1/2 q(J+**4 + J-**4) }
      Result := 0.5 * q * rl(j,Omp,Om)
  end;
end;

function TLinear.StateOrigin: Double;
var
  SigmaStep, Sigma: TQuantumNumber;
  E: Double;
begin
  { TODO: Call Hrot, rather than a partial evaluation }
  Result := Origin;
  { Estimate lowest spin orbit + spin-spin energy;
    ( Hrot(J=Lambda+S) may also be suitable, but more complicated )
    Just consider sigma = -s, 0 and s states }
  SigmaStep := Max(S,2);
  Sigma := -S;
  while Sigma <= S do begin
    E := Origin + A*Lambda*Sigma/4 { ALzSz }
                { 2/3 lambda (3Sz^2-S^2) }
      	        + 2/3 * LambdaSS*(3*Sigma*Sigma - S*(S+2)) / 4;
    if E < Result then
      Result := E;
    Inc(Sigma, SigmaStep);
  end;
end;

function TLinear.GetSettingsNames(AIndex: Integer; var LowestValue:Integer): PPChar;
const
  LambdaNames: array [0..14] of PChar = ('Sigma-', 'Sigma+', '', 'Pi', '', 'Delta', '', 'Phi', '', 'Gamma', '', 'H', '', 'I', nil);
begin
  if AIndex = LSI then begin
    Result := @LambdaNames;
    LowestValue := -1;
  end else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

class function TLinear.NucleusType: TNucleusClass;
begin
  Result := TLinearNucleus;
end;

function TLinear.HyperfineMatrixElement(A, B: TBasisState;
  Signp, Sign: Integer; MNode: TMNode; F: TQuantumNumber): Double;
var
  Bra, Ket: TLinearBasisState;
  JP, OmegaP, LamP, SigmaP, J, Omega, Lam, Sigma, dOm: TQuantumNumber;
  Nuc: TLinearNucleus;
  jjp1: Double;
  RankWanted: TQuantumNumber;
begin
  Bra := A as TLinearBasisState;
  Ket := B as TLinearBasisState;
  Nuc := MNode as TLinearNucleus;
  { This routine works for any sign of Lambda.
    Note that swapping trick used for Hrot will not work here, as it would
    require swapping the F quantum numbers which are evaluated elsewhere.
    Can't use sigma(xz) either, as we only have a part of the overall
    Hamiltonian operator here, so we would need to reflect that also }
{ This code will do the swap and adjust the phase for some cases,
  but may not be completely general
  Phase := 1;
  if Signp < 0 then begin
    if Sign > 0 then begin
      Bra := TLinearBasisState(B);
      Ket := TLinearBasisState(A);
    end;
    Phase := Minus(Bra.J-Ket.J);
    SignP := -SignP;
    Sign  := -Sign;
  end; }
  JP := Bra.J;
  OmegaP := SignP*Bra.Omega;
  LamP := SignP*Max(Lambda,0);
  SigmaP := OmegaP - LamP;
  J := Ket.J;
  Omega := Sign*Ket.Omega;
  Lam := Sign*Max(Lambda,0);
  Sigma := Omega - Lam;
  Result := 0;
  RankWanted := F mod 10;
  F := F div 10;
  if RankWanted = 2 then begin
    { Magnetic Hyperfine }
    { Work out average J for use in centrifugal distortion }
    { TODO: Use [,N^2]+ instead }
    jjp1 := (J + Jp)*(J + JP + 2) / 16;
    if LamP = Lam then begin
       dOm := OmegaP - Omega;
       Result := 0;
       if SigmaP = Sigma then
         Result := (Nuc.a + Nuc.aJ*jjp1)*Lam/2;
       Result := Result + (Nuc.b + Nuc.c/3 + (Nuc.bJ+Nuc.cJ/3)*jjp1)
                                  *Minus(S - SigmaP)
                                  *threej(S, -SigmaP, 2, dOm, S, Sigma)
                                  *Sqrt(S*(S+2)*(S+1)/4);
       Result := Result + (Nuc.c + Nuc.cJ*jjp1)
                          *Sqrt(30.0)/3.0*Minus(dOm + S - SigmaP)
                          *threej(S, -SigmaP, 2, dOm, S, Sigma)
                          *threej(2, -dOm, 4, 0, 2, dOm)
                          *Sqrt(S*(S+2)*(S+1)/4.0);
       Result := Result*Minus(Jp - OmegaP) * threej(Jp, -OmegaP, 2, dOm, J, Omega);
       if (Sigma = SigmaP) and (J = JP) then
         Result := Result + Nuc.cI*Sqrt( J * (J+2) / (Jp+1) / 4 );
    end else if Abs(LamP-Lam) = 4 then begin
       dOm := SigmaP - Sigma;
       Result := (Nuc.d + Nuc.dJ*jjp1)
                 *Minus(JP - OmegaP + S - SigmaP)
                 *threej(JP, -OmegaP, 2, -dOm, J, Omega)
                 *threej(S, -SigmaP, 2, dOm, S, Sigma)
                 *Sqrt(S*(S+2)*(S+1)/4.0);
    end else
      Exit;
    Result := Result * Sqrt(Nuc.Spin*(Nuc.Spin+2)*(Nuc.Spin+1)*(J+1.0)*(JP+1.0)/4);
  end else if RankWanted = 4 then begin
    { Quadrupole }
    if Sigma <> SigmaP then
      Exit;
    { Work out average J for use in centrifugal distortion }
    jjp1 := (J + Jp)*(J + JP + 2) / 16;
    dOm := OmegaP - Omega;
    if dOm = 0 then
      Result := Nuc.eQq0/2 + Nuc.eQq0J*jjp1
    else if Abs(dOm) = 4 then
      Result := Nuc.eQq2*Sqrt(6)/12 + Nuc.eQq2J*jjp1;
    if Result <> 0 then
      Result := Result { * minus(j + i + f) * sixj(f, i, j, 4, j p, i) }
                * Minus(J + Omega + dOm) * Sqrt((J + 1.0) * (JP + 1))
                * threej(4, -dOm, J, -Omega, JP, OmegaP)
                /( 2 * threej(Nuc.Spin, -Nuc.Spin, 4, 0, Nuc.Spin, Nuc.Spin) );
  end else
    raise NotYet;
end;

function TLinear.RotationalTransitionMatrixElement(A, B: TBasisState; SignA,
  SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber): Double;
var
  Comp: TQuantumNumber;
  Bra, Ket: TLinearBasisState;
  JP, OmegaP, LamP, SigmaP, J, Omega, Lam, Sigma: TQuantumNumber;
  Switch: Boolean;
  mm: Double;
begin
  Bra := A as TLinearBasisState;
  Ket := B as TLinearBasisState;
  JP := Bra.J;
  OmegaP := SignA*Bra.Omega;
  LamP := SignA*Max(Bra.S.Lambda,0);
  SigmaP := OmegaP - LamP;
  J := Ket.J;
  Omega := SignB*Ket.Omega;
  Lam := SignB*Max(Ket.S.Lambda,0);
  Sigma := Omega - Lam;
  Result := 0;

//  if Sigma <> Sigmap then
//    Exit; { Spin forbidden }
  { TODO: Spin forbidden transitions }
  if TransitionMoment is TSphericalTransitionMoment then begin
    Comp := TSphericalTransitionMoment(TransitionMoment).Component;
    if TransitionMoment is TSpinTransitionMoment then begin
      if (Abs(SigmaP-Sigma) <> Abs(Comp)) or (Lam <> Lamp) then
        Exit;
      Switch := (Comp <> OmegaP-Omega) or ((Comp = 0) and (Sigma < 0));
      case Abs(Comp) of
        0: Result := Abs(Sigma)/2; { Sign of simga is taken care of in switching }
        2: if (Bra.S.S = Ket.S.S) or (Bra.S.S <= 2) then
             Result := -rl(Bra.S.S,SigmaP,Sigma)/Sqrt(2) { S+-; }
           else
             raise NotYet; { TODO: Use Wigner-Eckart to get right scaling Sigma dependence for off diagonal in Spin }
      else
        raise NotYet;
      end;
    end else begin
      if Abs(Omega-OmegaP) <> Comp then
        Exit;
      if (Bra.S.S = Ket.S.S) and (Sigma <> Sigmap) then
        Exit; { Only enforce Delta Sigma = 0 rule for spin allowed transitions }
      Switch := (Comp <> OmegaP-Omega) or ((Comp = 0) and (Lam < 0));
      Result := 1;
      { Note that Lambda dependence is not included in matrix element }
    end;
    { If the other sign of Comp = q is required, apply the following symmetry
      relationship to relate the T(k,q) and T(k,-q) matrix elements:
      E* <S' SIGMA' LAMBDA' sym'| T(k,q) | S SIGMA LAMBDA sym>
       = (-1)^(S'-OMEGA'+sym') <S' -SIGMA' -LAMBDA' sym'|
             (-1)^q T(k,-q)
             (-1)^(S-OMEGA+sym) | S -SIGMA -LAMBDA sym>
       = (-1)^(S'-S+sym'-sym) <S' -SIGMA' -LAMBDA' sym'| T(k,-q) | S -SIGMA -LAMBDA sym>
    }
    if Switch then begin
      Result := Result*Minus(Bra.S.S-Ket.S.S);
      if (Bra.S.Lambda < 0) xor (Ket.S.Lambda < 0) then
        Result := -Result;
      if TTransitionMoment(TransitionMoment).Magnetic then
        Result := Result * Minus(TTransitionMoment(TransitionMoment).Rank); { Different rule for reflecting magnetic operators }
    end;
    with TTransitionMoment(TransitionMoment) do begin
      mm := (Jp*(Jp+2)+ J*(J+2))/8;
      Result := Result*Strength*Sqrt(1 + C*mm + D*mm*mm);
    end;
    if TransitionMoment is TRotationalTransitionMoment then begin
      if J = Jp then  { TODO: g N rather than g J }
        Result := Result * Sqrt(J*(J+2)*(J+1)/4)
      else
        Result := 0;
      Exit;
    end;
  end else
    raise NotYet;
  Result := Result*Sqrt((J+1.0)*(Jp+1)/(TTransitionMoment(TransitionMoment).Rank+1))
            *Minus(J-Omega+Comp)
            *threej(Jp, OmegaP, TTransitionMoment(TransitionMoment).Rank, Omega-OmegaP, J, -Omega);
  { TODO: Check phase here carefully }
  { Note that eq 11 in Dixon, Bayley and Ashfold paper is incomplete, as
    it assumes the M and K are projections of the same quantum number, which is
    not the case if hyperfine is included. }
end;

function TLinear.RotationalPerturbationMatrixElement(
   A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber): Double;
var
  Phase, dOm, SignS: Integer;
  Bra, Ket: TLinearBasisState;
  JP, OmP, LamP, J, Om, Lam: TQuantumNumber;
  Perturbation: TLinearPerturbation;
  ChangePhase, Swap: Boolean;
  Nuc: TLinearNucleus;

  MaxOmega, OmPP: TQuantumNumber;

  function BasePerturbation(OmegaP, Omega: TQuantumNumber): Double;
  { Perturbation without any centrifugal distortion }
  var
    SigmaP, Sigma: TQuantumNumber;
  begin
    SigmaP := OmegaP - LamP;
    Sigma := Omega - Lam;
    Result := 0;
    case Perturbation.Op of
      LS: { -9: Spin orbit coupling, with omega dependence from Wigner-Eckart theorem;
            See Lefebvre-Brion and Field, 2nd Edition Equation 3.4.36 }
        if (Omega = OmegaP) and (Abs(SigmaP-Sigma) <= 2) then begin
          if ChangePhase then
            SignS := -1
          else
            SignS := 1;
          if Swap then
            Result := phase * Perturbation.Value * Minus(Ket.S.S-SignS*SigmaP)
                      * threej(Ket.S.S, -SignS*Sigma, 2, SignS*(Sigma-SigmaP), Bra.S.S, SignS*SigmaP)
          else
            Result := phase * Perturbation.Value * Minus(Bra.S.S-SignS*Sigma)
                      * threej(Bra.S.S, -SignS*Sigmap, 2, SignS*(SigmaP-Sigma), Ket.S.S, SignS*Sigma);
          // if Perturbation.Op = LSJ then
          //  Result := Result * J * (J + 2) / 4;
        end;
      Luncouple: { -2 : J+^n L-^n + J-^n L+^n }
        if (SigmaP = Sigma) and (Abs(LamP-Lam) = 2*Perturbation.n) then
          Result := Phase * Perturbation.Value * rl(J, OmegaP, Omega);
      Homog: { 6: Simple homogeneous perturbation with delta omega = 0; deprectated }
        if Omega = OmegaP then
          Result := Phase * Perturbation.Value;
      aL: if (SigmaP = Sigma) and (Abs(LamP-Lam) = 2) then
            Result := Phase * Minus(JP-OmegaP) * threej(JP, -OmegaP, 2, OmegaP-Omega, J, Omega)
                     * (Omega-OmegaP) * Perturbation.Value / sqrt(8.0)
                     * Sqrt(Nuc.Spin*(Nuc.Spin+2)*(Nuc.Spin+1)*(J+1.0)*(JP+1.0)/4.0);
      e:  if Abs(LamP-Lam) = 2 then begin
            dOm := OmegaP - Omega;
            Result := -Phase * Perturbation.Value * Minus(JP - OmegaP + S - SigmaP + dOm)
                      * threej(JP, -OmegaP, 2, -dOm, J, Omega)
                      * threej(S, -SigmaP, 2, SigmaP-Sigma, S, Sigma)
                      * threej(2, SigmaP-Sigma, 4, LamP-Lam, 2, -dOm)
                      * Sqrt(20*S*(S+2)*(S+1)/4)
                      * Sqrt(Nuc.Spin*(Nuc.Spin+2)*(Nuc.Spin+1)*(J+1.0)*(JP+1.0)/4.0)
          end;
      eQq1: if (SigmaP = Sigma) and (Abs(LamP-Lam) = 2) then begin
              Result := -Phase * Perturbation.Value
                        * Minus(J + OmegaP) * Sqrt((J + 1.0) * (JP + 1))
                        * threej(4, Omega-OmegaP, J, -Omega, JP, OmegaP)
                        /( 2 * threej(Nuc.Spin, -Nuc.Spin, 4, 0, Nuc.Spin, Nuc.Spin)
                           * sqrt(6.0));
              if LamP > Lam then begin
                if Lam < 0 then
                  Result := - Result;
              end else if Lam > 0 then
                Result := - Result;
            end;
    else
      raise NotYet;
    end;
  end;

begin
  Result := 0;
  Perturbation := MNode as TLinearPerturbation;
  Bra := A as TLinearBasisState;
  Ket := B as TLinearBasisState;
  JP := Bra.J;
  OmP := SignA*Bra.Omega;
  LamP := SignA*Max(Bra.S.Lambda,0);
  J := Ket.J;
  Om := SignB*Ket.Omega;
  Lam := SignB*Max(Ket.S.Lambda,0);
  { For off diagonal matrix elements define matrix element to be that
    LamP > Lam and LamP > 0 (or SigmaP > Sigma and SigmaP > 0 if Lambda=LambdaP = 0)
    and use symmetry and hermiticity to relate to other elements. }
  if LamP > Lam then begin
    ChangePhase := LamP <= 0;
    Swap := ChangePhase;
  end else if LamP < Lam then begin
    ChangePhase := Lam <= 0;
    Swap := not ChangePhase;
  end else if LamP <> 0 then begin
    { Lam = Lamp }
    ChangePhase := LamP < 0;
    if ChangePhase then
      Swap := OmP < Om
    else
      Swap := OmP > Om;
  end else if OmP > Om then begin { LamP = Lam = 0 }
    ChangePhase := OmP <= 0;
    Swap := ChangePhase;
  end else if OmP < Om then begin
    ChangePhase := Om <= 0;
    Swap := not ChangePhase;
  end else begin
    { LamP = Lam = 0 and SigmaP = Sigma }
    ChangePhase := OmP < 0;
    Swap := False;
  end;
  if ChangePhase then begin
    Phase := Minus(Ket.S.S-Bra.S.S);
    if (Bra.S.Lambda < 0) xor (Ket.S.Lambda < 0) then
      Phase := -Phase;
  end else
    Phase := 1;
  if Perturbation.Nucleus > 0 then
    Nuc := Nucleus(Perturbation.Nucleus-1) as TLinearNucleus
  else
    Nuc := nil;

  if Perturbation.Npower = 0 then
    Result := BasePerturbation(OmP, Om)
  else begin
    OmPP := Max(Om-Perturbation.Npower, Lam-Ket.S.S);
    MaxOmega := Min(Om+Perturbation.Npower, Lam+Ket.S.S);
    while OmPP <= MaxOmega do begin
      if Abs(OmPP) <= J then
        Result := Result + BasePerturbation(OmP,OmPP)
                           * NRn(Perturbation.Npower, J, Lam, Ket.S.S, OmPP, Om);
      Inc(Ompp,2);
    end;
    OmPP := Max(Omp-Perturbation.Npower, LamP-Bra.S.S);
    MaxOmega := Min(Omp+Perturbation.Npower, LamP+Bra.S.S);
    while OmPP <= MaxOmega do begin
      if Abs(OmPP) <= J then
        Result := Result + NRn(Perturbation.Npower, JP, LamP, Bra.S.S, OmP, OmPP)
                           * BasePerturbation(OmPP,Om);
      Inc(Ompp,2);
    end;
    Result := Result / 2;
  end;
end;

const
  LinearSettingNames: array[LSI..LSI+1] of string = (
'Electronic angular momentum - use Sigma-, Sigma+, Pi, Delta, Phi, Gamma, H, I or a numerical value',
'true for gerade states; ignored if molecule is not symmetric.'
);

function TLinear.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearSettingNames))
     and (AIndex <= High(LinearSettingNames)) then
    Result := LinearSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 LinearParameterNames: array [LSP..LSP+27] of string = (
'Rotational constant.',
'Spin-orbit coupling constant.',
'Spin - spin coupling constant (=1.5*epsilon).',
'Spin-spin coupling constant giving lambda doubling for Omega = 0 states.',
'Higher Order Spin-Spin interaction (S>=2 only).',
'Spin rotation coupling constant.',
'Lambda doubling constant.',
'Lambda doubling constant.',
'Quartic centrifugal distortion.',
'Sextic centrifugal distortion.',
'J^8  centrifugal distortion.',
'J^10  centrifugal distortion.',
'J^12 centrifugal distortion. (Not P to avoid confusion with p.)',
'Centrifugal distortion of A.',
'Centrifugal Distortion of Lambda.',
'J^4 Centrifugal Distortion of Lambda.',
'Centrifugal distortion of o.',
'J^4 Centrifugal distortion of o.',
'J^6 Centrifugal distortion of o.',
'Centrifugal distortion of p.',
'J^6 Centrifugal distortion of p.',
'J^8 Centrifugal distortion of p.',
'Centrifugal distortion of q.',
'J^6 Centrifugal distortion of q.',
'J^8 Centrifugal distortion of q.',
'Centrifugal distortion of gamma.',
'J^6 Centrifugal distortion of gamma.',
'J^8 Centrifugal distortion of gamma.'
);

function TLinear.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearParameterNames))
     and (AIndex <= High(LinearParameterNames)) then
    Result := LinearParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TLinear.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result) then
    { TODO: Roll into lowest state origin logic }
    try
      if S = 0 then
        FInverted := False
      else if Lambda > 0 then
        { TODO: Call DiagonalMatrixlemnts somehow, so that everything as perturbations is still correct }
        FInverted := HRot(Lambda+S, Lambda, S, Lambda+S, Lambda, S, Lambda+S)
                     < HRot(Lambda+S, Lambda, -S, Lambda-S, Lambda, -S, Lambda-S)
      else
        FInverted := HRot(S, 0, 0, 0, 0, 0, 0) > HRot(S, 0, S, S, 0, S, S)
    except
      on E: Exception do
        { Copy Exception }
        Result := Exception.Create(E.Message);
    end;
end;

procedure TLinear.ConvertToPerturbations;
var
  i, Pow: Integer;
  procedure DoPert(Scale: Double; Index: Integer; O: TLinearOperator; An: Integer; ANPower: Integer);
  var
    P: TLinearPerturbation;
  begin
    Inc(Pow,2);
    Scale := Scale*GetParameter(Index);
    if Scale = 0 then Exit;
    P := TLinearPerturbation.Create(Owner);
    P.Value := Scale;
    with P do begin
      Op := O;
      n := An;
      Npower := ANPower;
    end;
    P.SetBraKetName(Name,Name);
    SetParameter(Index, 0);
  end;
begin
  Pow := 2;
  DoPert(1, IndexOfParameter('B'), Luncouple, 0, Pow);
  DoPert(-1, IndexOfParameter('D'), Luncouple, 0, Pow);
  for i := IndexOfParameter('H') to IndexOfParameter('PP') do
    DoPert(1, i, Luncouple, 0, Pow);
  if (Lambda > 0) and (S > 0) then begin
    DoPert(Lambda*S/threej(S,-S,2,0,S,S)/4, IndexOfParameter('A'), LS, 1, 0);
    DoPert(Lambda*S/threej(S,-S,2,0,S,S)/4, IndexOfParameter('AD'), LS, 1, 2);
  end;
end;

{ TLinearManifold }

class function TLinearManifold.OwnerType: TComponentClass;
begin
  Result := TLinearMolecule;
end;

function TLinearManifold.FixupProperties: Exception;
var
  i: Integer;
begin
  Result:=inherited FixupProperties;
  if not Assigned(Result) and not (Molecule as TLinearMolecule).Symmetric then
    for i := 0 to nStates-1 do
      if not (State[i] as TLinear).gerade then begin
        Result := EReportException.Create('ungerade state (' + State[i].Name + ') in unsymmetric molecule (' + Molecule.Name + ')');
        Exit;
      end;
end;

function TLinearManifold.FirstSym(J: TQuantumNumber):Integer;
begin
  Result := 0;
end;

function TLinearManifold.NextSym(J: TQuantumNumber; Sym: Integer):Integer;
begin
  if Sym = 0 then
    Result := 1
  else
    Result := -1;
end;

function TLinearManifold.nSym: Integer;
begin
  Result := 2;
end;

function TLinearManifold.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case AIndex of
    0: Result := TLinear;
    1: Result := TLinearPerturbation;
  else
    Result := nil;
  end;
end;

{ TLinearMolecule }

constructor TLinearMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := 'ShowJ=True,ShowN=True,ShowFNumber=True,Showef=True,SymWt=1,AsymWt=1,' + SavedProperties;
  inherited Create(AOwner);
end;

function TLinearMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  if TransitionMoment.Magnetic then
    Result := Sym1 = Sym2
  else
    Result := ((Sym1 xor Sym2) xor (TransitionMoment.Rank div 2) and 1) = 0; { Parity selection rule }
end;

function TLinearMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  if (Mixture.JAdjustSym) and ((J and 2) <> 0) then
    Result := 1 - Sym
  else
    Result := Sym;
end;

function TLinearMolecule.LastSym: Integer;
begin
  Result := 1;
end;

function TLinearMolecule.SymName(Sym: Integer): string;
begin
  if Sym = 0 then
    Result :='+'
  else
    Result := '-';
end;

function TLinearMolecule.JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string;
begin
  if Mixture.JAdjustSym then
    if ((J and 2) <> 0) xor (Sym = 1) then
      Result := 'f'
    else
      Result := 'e'
  else
    Result := inherited JAdjustedSymName(J, Sym);
end;

const
  LinearMoleculeSettingNames: array[LMI..LMI+8] of string = (
'True for D infinity h, False for C infinity v.',
'Statistical weight of symmetric rotational levels. Ignored if no centre of symmetry.',
'Statistical weight of asymmetric rotational levels. Ignored if no centre of symmetry.',
'Set to use the R^2 Hamiltonian rather than the default N^2 form.',
'Show J in state label.',
'Show Omega in state label.',
'Show N in state label.',
'Show F Number in state label.',
'Show e/f symmetry in state label.'
);

function TLinearMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearMoleculeSettingNames))
     and (AIndex <= High(LinearMoleculeSettingNames)) then
    Result := LinearMoleculeSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

class function TLinearMolecule.ManifoldType: TMNodeClass;
begin
  Result := TLinearManifold;
end;

{ TLinearPerturbation }

function TLinearPerturbation.GetOperator(const Index: Integer): TLinearOperator;
begin
  Result := TLinearOperator(GetSetting(Index));
end;

procedure TLinearPerturbation.SetOperator(const Index: Integer; const AValue: TLinearOperator);
begin
  SetSetting(Index, Ord(AValue));
  case AValue of
    aL, e: NuclearRank := 2;
    eQq1:  NuclearRank := 4;
  else
    NuclearRank := 0;
  end;
end;

function TLinearPerturbation.PairName: string;
  function Show(Op: string; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Op;
    else
      Result := Op + IntToStr(Power)
    end;
  end;
begin
  case Op of
    LS: Result := 'LS';
    Luncouple: Result := Show('J+',n) + Show('L-',n);
    aL: Result := 'I-L+';
    e:  Result := 'e';
    eQq1: Result := 'eQq1';
  end;
  if Npower <> 0 then
    Result := Result + ' N^' + IntToStr(Npower);
end;

constructor TLinearPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := 'n=1,' + SavedProperties;
  inherited;
end;

function TLinearPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LPI then
    Result := 'Perturbation type'
  else if AIndex = LPI+1 then
    Result := 'Power of operator in perturbation - not currently used.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TLinearNucleus }

const
 LinearNucleusParameterNames: array [1..13] of string = (
'Nuclear spin - orbit interaction (I.L).',
'Nuclear spin - electron spin (I.S). Note that b = bF - c/3.',
'Nuclear spin - electron spin dipole dipole interaction diagonal in Lambda.',
'Nuclear spin - electron spin dipole dipole interaction off diagonal in Lambda by 2.',
'Nuclear quadrupole coupling constant diagonal in Lambda.',
'Nuclear quadrupole coupling constant off diagonal in Lambda by 2.',
'Nuclear spin rotation (I.J).',
'Centrifugal distortion of nuclear spin - orbit interaction (I.L). (Outline implementation)',
'Centrifugal distortion of nuclear spin - electron spin (I.S). (Outline implementation)',
'Centrifugal distortion of nuclear spin - electron spin dipole dipole interaction diagonal in Lambda. (Outline implementation)',
'Centrifugal distortion of nuclear spin - electron spin dipole dipole interaction off diagonal in Lambda by 2. (Outline implementation)',
'Centrifugal distortion of nuclear quadrupole coupling constant diagonal in Lambda. (Outline implementation)',
'Centrifugal distortion of nuclear quadrupole coupling constant off diagonal in Lambda by 2. (Outline implementation)'
);

function TLinearNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearNucleusParameterNames))
     and (AIndex <= High(LinearNucleusParameterNames)) then
    Result := LinearNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

initialization
  RegisterClasses([TLinearMolecule, TLinearManifold, TLinear, TLinearNucleus, TLinearPerturbation]);
  TSpecies.AddMoleculeType(TLinearMolecule);
end.


