unit OAsymT;

{ Open Shell Asymmetric Top }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses SysUtils, Classes, MNode, Mixture, QuantumNumber;

const
  FirstWidthPar = LSP+40;

type

  TAsymmetricMolecule = class;

  TAsymmetricTop = class (TState)
  private
    FKinc: Integer;
    TSR: array [0..2] of array [-2..2] of Double; { Spherical tensor equivalent of spin rotation terms }
    QuarticSpinRotation: Boolean; { True if any Quartic Spin Rotation term non zero }
    RoSyms: array [0..7] of Integer; { Mapping from overall sym number to rotational eo symmetry;
                                       Implicitly contains more than one value for subgroups of C2v }
    nRotSym: Integer; { Number of different symmetries of rotational wavefunctions }
  public
    property Kinc: Integer read FKinc;
    constructor Create(AOwner: TComponent); override;
    class function OwnerType: TComponentClass; override;
    class function NucleusType: TNucleusClass; override;
    function Molecule: TAsymmetricMolecule;
    function OddSpin: TQuantumNumber; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer; override;
    function AddOneN(Eigenvalues: TEigenvalues; N, J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer;
    function FixupProperties: Exception; override;
    function HRotA(J, Kp, K: TQuantumNumber): Double;
    function HRotS(J, Kp, K: TQuantumNumber): Double;
    function HSR(J, Np, Kp, N, K:TQuantumNumber): Double;
    function HasWidth: Boolean; override;
    function HWidth(J, Kp, K: TQuantumNumber): Double;
    function HDip(TransitionMoment: TTransitionMoment; Jp, Kp, J, K: TQuantumNumber): Double;
    { Key Matrix element functions }
    function RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double; override;
    function RotationalPerturbationMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Double; override;
    function HyperfineMatrixElement(A, B: TBasisState; Signp, Sign: Integer; MNode: TMNode; F: TQuantumNumber): Double; override;
    function RotationalTransitionMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Double; override;

    function GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function zIndex: Integer; { Setting index of zAxis rotational constant }
  published
    property Symmetry: Integer         index LSI   read GetSetting write SetSetting;
    property Kmin:     TQuantumNumber  index LSI+1 read GetSetting write SetSetting;
    property Kmax:     TQuantumNumber  index LSI+2 read GetSetting write SetSetting;

    property A: Double index LSP   read GetParameter write SetParameter;
    property B: Double index LSP+1 read GetParameter write SetParameter;
    property C: Double index LSP+2 read GetParameter write SetParameter;
    { Alternate form for rotational constants }
    property BBar:  Double index LSP+3 read GetParameter write SetParameter;
    property BDelta:Double index LSP+4 read GetParameter write SetParameter;
    { Quartic Centrifugal Distortion }
    property DK:      Double index LSP+ 5 read GetParameter write SetParameter;
    property DJK:     Double index LSP+ 6 read GetParameter write SetParameter;
    property DJ:      Double index LSP+ 7 read GetParameter write SetParameter;
    property deltaK:  Double index LSP+ 8 read GetParameter write SetParameter;
    property deltaJ:  Double index LSP+ 9 read GetParameter write SetParameter;
    { Sextic Centrifugal Distortion }
    property HK:    Double index LSP+10 read GetParameter write SetParameter;
    property HKJ:   Double index LSP+11 read GetParameter write SetParameter;
    property HJK:   Double index LSP+12 read GetParameter write SetParameter;
    property HJ:    Double index LSP+13 read GetParameter write SetParameter;
    property phiK:  Double index LSP+14 read GetParameter write SetParameter;
    property phiJK: Double index LSP+15 read GetParameter write SetParameter;
    property phiJ:  Double index LSP+16 read GetParameter write SetParameter;
    { Octic Centrifugal Distortion }
    property LK:    Double index LSP+17 read GetParameter write SetParameter;
    property LKKJ:  Double index LSP+18 read GetParameter write SetParameter;
    property LJK:   Double index LSP+19 read GetParameter write SetParameter;
    property LJJK:  Double index LSP+20 read GetParameter write SetParameter;
    property LJ:    Double index LSP+21 read GetParameter write SetParameter;
    property llK:   Double index LSP+22 read GetParameter write SetParameter;
    property llKJ:  Double index LSP+23 read GetParameter write SetParameter;
    property llJK:  Double index LSP+24 read GetParameter write SetParameter;
    property llJ:   Double index LSP+25 read GetParameter write SetParameter;
    { Spin-Rotation Parameters }
    property eaa:    Double index LSP+26 read GetParameter write SetParameter;
    property ebb:    Double index LSP+27 read GetParameter write SetParameter;
    property ecc:    Double index LSP+28 read GetParameter write SetParameter;
    property eabbar: Double index LSP+29 read GetParameter write SetParameter; { (eab + eba)/2 }
    property eacbar: Double index LSP+30 read GetParameter write SetParameter; { (eac + eca)/2 }
    property ebcbar: Double index LSP+31 read GetParameter write SetParameter; { (ebc + ebc)/2 }
      { Anti symmetric combinations like eab - eba omitted as they are not dererminable }
    { Quartic spin rotation parameters - use S reduction names for parameters }
    {   A reduction names given as comments }
    property DsN:   { DeltasN }  Double index LSP+32 read GetParameter write SetParameter;
    property DsNK:  { DeltasNK } Double index LSP+33 read GetParameter write SetParameter;
    property DsKN:  { DeltasKN } Double index LSP+34 read GetParameter write SetParameter;
    property DsK:   { DeltasK }  Double index LSP+35 read GetParameter write SetParameter;
    property ds1:   { deltasN }  Double index LSP+36 read GetParameter write SetParameter;
    property ds2:   { deltasK }  Double index LSP+37 read GetParameter write SetParameter;
    { Spin-Spin interaction }
    property alpha: Double index LSP+38 read GetParameter write SetParameter;
    property beta:  Double index LSP+39 read GetParameter write SetParameter;

    { Linewidth parameters - initial stab }
    property wKa:   Double index FirstWidthPar read GetParameter write SetParameter;
    property wKb:   Double index FirstWidthPar+1 read GetParameter write SetParameter;
    property wKc:   Double index FirstWidthPar+2 read GetParameter write SetParameter;
    property wBar:  Double index FirstWidthPar+3 read GetParameter write SetParameter;
    property wDelta:Double index FirstWidthPar+4 read GetParameter write SetParameter;
  end;

  TAsymmetricTopPerturbation = class (TPerturbation)
  public
    function MatrixElement(J, Np, Kp, N, K: TQuantumNumber): Double;
    function Molecule: TAsymmetricMolecule;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure SetAxisPower(Axis: TxyzAxis; AValue: Integer);
  published
    property JPower:Integer index LPI   read GetSetting write SetSetting;
    property aPower:Integer index LPI+1 read GetSetting write SetSetting;
    property bPower:Integer index LPI+2 read GetSetting write SetSetting;
    property cPower:Integer index LPI+3 read GetSetting write SetSetting;
    property pPower:Integer index LPI+4 read GetSetting write SetSetting;
  end;

  { TAsymmetricTopNucleus }

  TAsymmetricTopNucleus = class(TNucleus)
  private
    TT: array[-2..2] of Double; { Spherical tensor equivalent of dipole - dipole terms }
  public
    function FixupProperties: Exception; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property aF:     Double index  1 read GetParameter write SetParameter;
    property Tzz:    Double index  2 read GetParameter write SetParameter;
    property Txxmyy: Double index  3 read GetParameter write SetParameter;
    property Tab:    Double index  4 read GetParameter write SetParameter;
    property Tac:    Double index  5 read GetParameter write SetParameter;
    property Tbc:    Double index  6 read GetParameter write SetParameter;
  end;

  TAsymmetricTopBasisState = class(TBasisState)
  public
    N, K: TQuantumNumber;
    Ka, Kc: TQuantumNumber;
    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel: string; override; { State name, assuming energy ordering preserved }
    function StateLabelKey: string; override; { Explanation of State Name }
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as qRK(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
  end;

  TAsymmetricManifold = class(TManifold)
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
  public
    class function OwnerType: TComponentClass; override;
    function FirstSym(J: TQuantumNumber):Integer; override;
    function NextSym(J: TQuantumNumber; Sym: Integer):Integer; override;
    function nSym: Integer; override; { Number of different symmetries }
  end;

  TRepresentation = (Ir=0, IIr, IIIr, Il, IIl, IIIl);
  TPointGroup = (C1=0, Ci, C2, Cs, D2, C2v, C2h, D2h);

  TAsymmetricMolecule = class(TMolecule)
  private
    FOddRankSym: Integer;
    function GetPointGroup(Index: integer): TPointGroup;
    function GetRepresentation(Index: integer): TRepresentation;
    procedure SetPointGroup(Index: integer; const AValue: TPointGroup);
    procedure SetRepresentation(Index: integer; const AValue: TRepresentation);
    function GetAxis(const Index: Integer): TabcAxis;
    procedure SetAxis(const Index: Integer; const AValue: TabcAxis);
  public
    constructor Create(AOwner: TComponent); override;
    class function ManifoldType: TMNodeClass; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function eoSym(RveSym: Integer): integer;
    function FixupProperties: Exception; override;
    function GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    property OddRankSym: Integer read FOddRankSym; { Symmetry of dipole moment }
    function HelpOnSetting(AIndex: Integer): string; override;
    function xyzAxis(abcAxis: TabcAxis): TxyzAxis; { Given a b or c, return corresponding x y or z axis }
  published
    property PointGroup:     TPointGroup index LMI read GetPointGroup write SetPointGroup;
    property Representation: TRepresentation index LMI+1 read GetRepresentation write SetRepresentation;
    property SReduction:     Boolean  index LMI+2 read GetFlag write SetFlag; { Use Watson's S reduction (as opposed to A) }
    property eeWt:           Integer  index LMI+3 read GetSetting write SetSetting;
    property eoWt:           Integer  index LMI+4 read GetSetting write SetSetting;
    property oeWt:           Integer  index LMI+5 read GetSetting write SetSetting;
    property ooWt:           Integer  index LMI+6 read GetSetting write SetSetting;
    property C2zAxis:        TabcAxis index LMI+7 read GetAxis write SetAxis;
    property C2xAxis:        TabcAxis index LMI+8 read GetAxis write SetAxis;
    property PseudoC2v:      Boolean  index LMI+9 read GetFlag write SetFlag; { Ignore any operators breaking the C2v symmetry }
    property FakeSym:        Boolean  index LMI+10 read GetFlag write SetFlag; { Turn off transition moment test }
  end;


implementation

uses Math, MathSup, nj, Profile;

var
  HRotATime: string = 'HRotA';
  AsymTTransitionMatrixElement: string = 'OAsymT Transition Matrix Element';
  T1Time: string = 'T1';

const
  FirstStatisticalWeight = LMI+3;

{ Two symmetry types are in use here. For the V group, the coding is:
    ee=00 eo=01 oe=10 oo=11
  e and o specify the oddness/evenness of Ka and Kc respectively.
  Multiplication of symmetry expressed like this is simply by bitwise xor.
  
  For the point groups of the molecule as a whole, (the Symmetry variable) the coding
  depends on the point group, but multiplication of symmetries is still expressed by bitwise xor.

  
  The axis for which K is used in the basis is specified by the representation
  Ir: z=a and so on
  To use the same ideas for the vibronic symmetry, we use the identification (for C2v):
  R2z' = C2
  R2x' = sigma(y'z')
  R2y' = sigma(x'z')
  where yet another axis system is being introduced for the point group. This is
  defined by the postion of C2 axis (z') and the out of plane axis (x'). The
  location of these wrt the inertial axes is defined by the C2zAxis and C2xAxis
  variables.
  
  For D2h and D2 the C2zAxis and C2xAxis variable specify the axes corresponding to
  the C2z and C2z operators in the cheracter table

  For Cs we have:
  R2z' = sigma(x'y')
  where C2zAxis is now the out of plane axis, and C2xAxis is unused
  C2 and C2h are effectively the same, the matching point group operator is C2

  For C1, no simplification is possible
  Ci is similar, with the proviso that all rotational functions are g
  
   }

{ TAsymmetricTop }

const
  C2vCharacterTable: array [0..3] of array [0..3] of Integer =
   (    { E C2 s(zx) s*(yz) }
        {   Rz   Ry     Rx  }
   {A1} ( 1, 1,   1,     1),
   {A2} ( 1, 1,  -1,    -1),
   {B1} ( 1,-1,   1,    -1),
   {B2} ( 1,-1,  -1,     1));
  CsCharacterTable: array [0..1] of array [0..1] of Integer =
   (    { E s(xy) } { z out of plane }
        {    Rz   }
   {A1} ( 1,  1   ),
   {A2} ( 1, -1));

  D2hCharacterTable: array [0..7] of array [0..4] of Integer =
   (     { E C2z C2y C2x i }
         {   Rz  Ry  Rx  E* }
   {Ag}  ( 1, 1,  1,  1,  1),
   {B1g} ( 1, 1, -1, -1,  1),  {Rz}
   {B2g} ( 1,-1,  1, -1,  1),  {Ry}
   {B3g} ( 1,-1, -1,  1,  1),  {Rx}
   {Au}  ( 1, 1,  1,  1, -1),
   {B1u} ( 1, 1, -1, -1, -1),  {z}
   {B2u} ( 1,-1,  1, -1, -1),  {y}
   {B3u} ( 1,-1, -1,  1, -1)); {x}
  { D2 is the same as ignoring g/u in D2h }

   C2hCharacterTable: array [0..3] of array [0..2] of Integer =
   (     { E C2z i }
         {   Rz  E* }
   {Ag}  ( 1, 1, 1),
   {Bg}  ( 1,-1, 1),
   {Au}  ( 1, 1,-1),
   {Bu}  ( 1,-1,-1));
  { C2 is the same as ignoring g/u in C2h }
  { Ci only has g/u, and rotations are always g }



  { Antisymmetric wrt rotation about the inertial axis implies the following symmetries: }
  AntiSymWrt: array [TabcAxis] of Integer = ( 2 {a - 10}, 3 {b - 11}, 1 {c - 01} );
  { Non totally symmetric function that is symmetric wrt the given axis: }
  SymWrt: array [TabcAxis] of Integer = ( 1 {a - 01}, 3 {b - 11}, 2 {c - 10} );
  { Mask to switch to functions that are antisymmetric wrt reflection in Cs }
  CsSwitch: array [TabcAxis] of Integer = ( 2 {a - 10}, 1 {b - arbitrary}, 1 {c - 01} );

const
  AxisNames: array [TabcAxis] of string = ('a', 'b', 'c');

{ Mapping between x/y/z and a/b/c axes }
const                                             {   Ir,   IIr,  IIIr,    Il,   IIl,  IIIl}
  ZAxisFor: array [TRepresentation] of TabcAxis = (aAxis, bAxis, cAxis, aAxis, bAxis, cAxis);
  XAxisFor: array [TRepresentation] of TabcAxis = (bAxis, cAxis, aAxis, cAxis, aAxis, bAxis);
  YAxisFor: array [TRepresentation] of TabcAxis = (cAxis, aAxis, bAxis, bAxis, cAxis, aAxis);
  { Reverse Mapping }
  aAxisFor: array [TRepresentation] of TxyzAxis = (zAxis, yAxis, xAxis, zAxis, xAxis, yAxis);
  bAxisFor: array [TRepresentation] of TxyzAxis = (xAxis, zAxis, yAxis, yAxis, zAxis, xAxis);
  cAxisFor: array [TRepresentation] of TxyzAxis = (yAxis, xAxis, zAxis, xAxis, yAxis, zAxis);

function TAsymmetricMolecule.xyzAxis(abcAxis: TabcAxis): TxyzAxis;
{ Given a b or c, return corresponding x y or z axis }
begin
  if ZAxisFor[Representation] = abcAxis then
    Result := zAxis
  else if YAxisFor[Representation] = abcAxis then
    Result := yAxis
  else
    Result := xAxis;
end;

function abcAxis(Representation: TRepresentation; xyzAxis: TxyzAxis): TabcAxis;
{ Given x y or z, return corresponding a b or c axis }
begin
  if aAxisFor[Representation] = xyzAxis then
    Result := aAxis
  else if bAxisFor[Representation] = xyzAxis then
    Result := bAxis
  else
    Result := cAxis;
end;

function TAsymmetricMolecule.eoSym(RveSym: Integer): Integer;
{ Convert between D2, C2v and D2h point group labels and eo symmetry labels }
{ For the other point groups, a particular symmetry will give more than one eo combination }
var
  Sym: Integer;
begin
  Result := 0;
  RveSym := RveSym and 3; { Discard E*/i for D2h }
  { Character table ordering is the same for the rotation operations for D2, C2v and D2h }
  { We know directly from eo label the symmetry wrt Ra and Rc, but not Rb, so
    must use only these two only to match irreducible representation labels }
  Sym := AntiSymWrt[C2zAxis];
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[RveSym][1{=C2}] < 0 then
      Result := Sym;
  Sym := AntiSymWrt[C2zAxis] xor AntiSymWrt[C2xAxis]; { = AntiSymWrt[C2yAxis] }
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[RveSym][2{=sigma(zx)}] < 0 then
      Result := Result or Sym;
  Sym := AntiSymWrt[C2xAxis];
  if (Sym <> 3) then { if not b axis }
    if C2vCharacterTable[RveSym][3{=sigma(yz)}] < 0 then
      Result := Result or Sym;
end;

function TAsymmetricTop.FixupProperties: Exception;
var
  i, j, w, w1: Integer;
  eaaIndex: Integer;
  exx, eyy, ezz, exybar, exzbar, eyzbar: Double;
  function efind(Axis1, Axis2: TabcAxis):Double;
  var
    s: string;
  begin
    if Axis1 < Axis2 then
      s := AxisNames[Axis1] + AxisNames[Axis2]
    else
      s := AxisNames[Axis2] + AxisNames[Axis1];
    Result := GetParameter(IndexOfParameter('e'+s+'bar'))
  end;
begin
  FFullRotationalSymmetry := Molecule.PseudoC2v or (Molecule.PointGroup in [D2, C2v, D2h]);
  { Convert spin rotation terms to spherical tensor form }
  eaaIndex := IndexOfParameter('eaa');
  ezz := GetParameter(eaaIndex+Ord(ZAxisFor[Molecule.Representation]));
  exx := GetParameter(eaaIndex+Ord(XAxisFor[Molecule.Representation]));
  eyy := GetParameter(eaaIndex+Ord(YAxisFor[Molecule.Representation]));
  { Order is ab ac bc }
  exybar := efind(XAxisFor[Molecule.Representation], YAxisFor[Molecule.Representation]);
  exzbar := efind(XAxisFor[Molecule.Representation], ZAxisFor[Molecule.Representation]);
  eyzbar := efind(YAxisFor[Molecule.Representation], ZAxisFor[Molecule.Representation]);

  TSR[0][0] := -(exx+eyy+ezz)/sqrt(3.0);
  { First rank components are all zero, as these are only determined by the antisymetric sums
    which are not determined }
  TSR[1][ 0] := 0;
  TSR[1][-1] := 0;
  TSR[1][ 1] := 0;
  TSR[2][ 0] := (2*ezz - exx - eyy)/Sqrt(6.0);
  TSR[2][ 1] := -(exzbar + eyzbar);
  TSR[2][-1] :=  (exzbar - eyzbar);
  TSR[2][ 2] :=  (exx - eyy)/2 + exybar;
  TSR[2][-2] :=  (exx - eyy)/2 - exybar;

  { Find if any Quartic Spin Rotation }
  QuarticSpinRotation := False;
  for i := IndexOfParameter('DsN') to IndexOfParameter('ds2') do
    QuarticSpinRotation := QuarticSpinRotation or (GetParameter(i) <> 0);

  { Defaults - need adjusting in some cases }
  for i := 0 to 3 do
    RoSyms[i] := i;

  case Molecule.PointGroup of
    C1, Ci: if Molecule.PseudoC2v then
              nRotSym := 4
            else
              nRotSym := 1;
    Cs, C2, C2h: begin
      if Molecule.PseudoC2v then
        nRotSym := 4
      else begin
        nRotSym := 2;
        { Pair of symmetric functions }
        RoSyms[0] := 0;
        RoSyms[2] := SymWrt[Molecule.C2zAxis];
        { Pair of antisymmetric functions }
        RoSyms[1] := RoSyms[0] xor CsSwitch[Molecule.C2zAxis];
        RoSyms[3] := RoSyms[2] xor CsSwitch[Molecule.C2zAxis];
      end;
      if (Symmetry and 1) <> 0 then
        { Antisymmetric (wrt R2z') vibronic state - switch over }
        for i := 0 to 3 do
          RoSyms[i] := RoSyms[i] xor CsSwitch[Molecule.C2zAxis];
    end;
    C2v, D2, D2h: begin
      nRotSym := 4;
      for i := 0 to Molecule.LastSym do
        RoSyms[i] := (Molecule.eoSym(Symmetry) xor i) and 3;
    end;
  else
    raise NotYet;
  end;
  { To generate Ka Kc assignments, must generate K states in order of increasing energy.
    Decide on this ordering here.
    Note that this could generate odd results for if the centrifugal distortion
    is dominant, or if B ~ (A+C)/2. }
  case ZAxisFor[Molecule.Representation] of
    { Prolate - definately increasing with K }
    aAxis: FKinc := 4;
    { Oblate - definately decreasing with K }
    cAxis: FKinc := -4;
    { K is intermediate axis - depends on whether B is closer to A or C }
    bAxis: if B < (A+C)/2 then
             FKinc := -4
           else
             FKinc := 4;
  end;
  Result := inherited FixupProperties;
  if Assigned(Result) then
    Exit;
  { TODO: Derive effective symmetries from transition moments? }
  { Check for consistent statistical weights }
  for i := 0 to Manifold.nSym-1 do begin
    w := Molecule.GetSetting(FirstStatisticalWeight+(RoSyms[i] xor RoSyms[0]));
    j := i + nRotSym;
    while j < 4 do begin
      w1 := Molecule.GetSetting(FirstStatisticalWeight+(RoSyms[j] xor RoSyms[0]));
      if w <> w1 then
        Result := Exception.Create('Inconsistent statistical weights for ' + Name);
      Inc(j, nRotSym);
    end;
  end;
end;


const
  C1Names: array [0..1] of PChar = ('A', nil);
  CiNames: array [0..2] of PChar = ('Ag', 'Au', nil);
  C2Names: array [0..2] of PChar = ('A', 'B', nil);
  CsNames: array [0..2] of PChar = ('A''', 'A"', nil);
  D2Names: array [0..4] of PChar = ('A', 'B1', 'B2', 'B3', nil);
  C2vNames: array [0..4] of PChar = ('A1', 'A2', 'B1', 'B2', nil);
  C2hNames: array [0..4] of PChar = ('Ag', 'Bg', 'Au', 'Bu', nil);
  D2hNames: array [0..8] of PChar = ('Ag', 'B1g', 'B2g', 'B3g', 'Au', 'B1u', 'B2u', 'B3u', nil);

function TAsymmetricTop.GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar;
begin
  if not Assigned(Owner) then begin { Should never happen }
    Result := @C2vNames;
    Exit;
  end;
  if AIndex = LSI then
    case Molecule.PointGroup of
      C1: Result := @C1Names;
      Ci: Result := @CiNames;
      C2: Result := @C2Names;
      Cs: Result := @CsNames;
      D2: Result := @D2Names;
      C2v: Result := @C2vNames;
      C2h: Result := @C2hNames;
      D2h: Result := @D2hNames;
    else
      Raise NotYet;
    end
  else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

constructor TAsymmetricTop.Create(AOwner: TComponent);
begin
  SavedProperties := 'Kmin=-1,Kmax=-1,' + SavedProperties;
  inherited;
end;

class function TAsymmetricTop.OwnerType: TComponentClass;
begin
  Result := TAsymmetricManifold;
end;

class function TAsymmetricTop.NucleusType: TNucleusClass;
begin
  Result := TAsymmetricTopNucleus;
end;

function TAsymmetricTop.Molecule: TAsymmetricMolecule;
begin
  Result := Owner.Owner as TAsymmetricMolecule;
end;

function TAsymmetricTop.OddSpin: TQuantumNumber;
begin
  Result := (S + Molecule.MaxNuclearSpin) and 1; { Closed shell for now }
end;

function TAsymmetricTop.AddOneN(Eigenvalues: TEigenvalues; N, J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer;
var
  K, Khigh, Ka, Kc: TQuantumNumber;
  cSym, Phase, i: Integer;
  function AntiSymmetricWrtAxis(Axis: TabcAxis): Boolean;
  { Given axis, determine from Sym whether we want wavefunction to be symmetric or antisymmetric
    wrt C2 about that axis }
  begin
    with Eigenvalues do
      case Axis of
        aAxis: Result := (cSym and 2) > 0;
        bAxis: Result := (((cSym shr 1) xor cSym) and 1) > 0;
        cAxis: Result := (cSym and 1) > 0;
      end;
  end;
  procedure Add(K: TQuantumNumber);
  var
    BasisState: TAsymmetricTopBasisState;
  begin
    if ((Kmin >= 0) and (Abs(K) < Kmin))
       or ((Kmax >= 0) and (Abs(K) > Kmax)) then
      Exit;
    BasisState := TAsymmetricTopBasisState.Create(Eigenvalues);
    BasisState.State := Self;
    BasisState.K := K;
    BasisState.N := N;
    if K = 0 then
      BasisState.Phase := 0
    else
      BasisState.Phase := Phase;
    BasisState.Ka := Ka;
    BasisState.Kc := Kc;
    BasisState.StateIndex := Eigenvalues.LastStateIndex;
    BasisState.J := J;
    BasisState.HyperfineState := HyperfineState;
  end;
begin
  i := Eigenvalues.Sym;
  if (i xor Symmetry) >= nRotSym then
    Exit; { Special case test for point groups with c/u symmetry - rotational functions are always g,
            so u overall functions must be u vibronic }
  repeat
    cSym := RoSyms[i];
    { Statistical Weight }
    Result := Molecule.GetSetting(FirstStatisticalWeight+(cSym xor RoSyms[0]));
    if Result = 0 then
      Exit;
    if AntiSymmetricWrtAxis(ZAxisFor[Molecule.Representation]) then
      K := 2
    else
      K := 0;
    if AntiSymmetricWrtAxis(XAxisFor[Molecule.Representation]) xor (Minus(N) < 0) then
      Phase := -1
    else
      Phase := 1;
    if Minus(N-K) < 0 then
      Khigh := N-2
    else
      Khigh := N;
    if Kinc < 0 then
      K := Khigh;
    { Work out Ka and Kc for lowest state }
    { Kc is simply J or J-1 to match xo or xe }
    if (Minus(N) < 0) xor ((cSym and 1) <> 0) then
      Kc := N-2
    else
      Kc := N;
    { Ka is 1 for oe or oo }
    if (cSym and 2) <> 0 then
      Ka := 2 { oe or oo }
    else begin { ee or eo - use the requirement that Ka+Kc = J or J+1 }
      Ka := N - Kc;
      if Minus(Ka) < 0 then
        Inc(Ka, 2);
    end;
    while (K <= N) and (K >= 0) do begin
      if (K <> 0) or (Phase = 1) then begin
        { E- matrix has K = 0 missing }
        Add(K);
        Inc(Ka,4);
        Dec(Kc,4);
      end;
      Inc(K,Kinc);
    end;
    Inc(i, nRotSym);
    Inc(Eigenvalues.LastStateIndex);
  until i >= 4;
end;

function TAsymmetricTop.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer;
var
  N: TQuantumNumber;
  StatWt: Integer;
begin
  Result := -1;
  N := Abs(J-S);
  while N <= J+S do begin
    StatWt := AddOneN(Eigenvalues,N,J,HyperfineState);
    if Result < 0 then
      Result := StatWt
    else if Result <> StatWt then
      raise Exception.Create('AddBasis: Inconsistent Statistical Weights in '+Name);
    Inc(N,2);
  end;
end;

function TAsymmetricTop.HWidth(J, Kp, K: TQuantumNumber): Double;
{ Linewidth parameters - as a first stab, go for <Ka^2>, etc. }
{ TODO: This is straight copy of HRot - can this be generalised? }
var
  Izz, Iyy, Ixx, JSq, KSq, IxxpIyy, IxxmIyy: Double;
begin
  if K = Kp then begin
    Izz := GetParameter(FirstWidthPar+Ord(ZAxisFor[Molecule.Representation]));
    Ixx := GetParameter(FirstWidthPar+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(FirstWidthPar+Ord(YAxisFor[Molecule.Representation]));
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    IxxpIyy := Ixx + Iyy + wBar*2;
    Result := Width + Izz*KSq  + IxxpIyy*(JSq - KSq)/2;
  end else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(FirstWidthPar+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(FirstWidthPar+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + wDelta;
    JSq := J*(J+2) / 4;
    Result := IxxmIyy/4 * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else
    Result := 0;
end;

function TAsymmetricTop.HRotA(J, Kp, K: TQuantumNumber): Double;
{ Hamiltonian using Watson's A reduction }
var
  Izz, Iyy, Ixx, JSq, KSq, KpSq, IxxpIyy, IxxmIyy: Double;
begin
  if K = Kp then begin
    Izz := GetParameter(LSP+Ord(ZAxisFor[Molecule.Representation]));
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    IxxpIyy := Ixx + Iyy + BBar*2;
    Result := LJ*Sqr(Sqr(JSq)) + LJJK*Sqr(JSq)*JSq*KSq + LJK*Sqr(JSq)*Sqr(KSq)
              + LKKJ*JSq*KSq*Sqr(KSq) + LK*Sqr(Sqr(KSq));
    Result := Result + HJ*Sqr(JSq)*JSq + HJK*Sqr(JSq)*KSq + HKJ*Jsq*Sqr(KSq) + HK*Sqr(KSq)*KSq;
    Result := Result - DJ*Sqr(JSq) - DJK*JSq*KSq - DK*Sqr(KSq);
    Result := Result + Izz*KSq  + IxxpIyy*(JSq - KSq)/2 + Origin;
  end else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + BDelta;
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    KpSq := Kp*Kp / 4;
    Result := llJ*JSq*Sqr(JSq) + llJK*Sqr(JSq)*(KSq+KpSq)/2 + llKJ*Jsq*(Sqr(KSq)+Sqr(KpSq))/2 + llK*(Sqr(KSq)*KSq+Sqr(KpSq)*KpSq)/2;
    Result := Result + phiJ*Sqr(JSq) + phiJK*Jsq*(KSq+KpSq)/2 + phiK*(Sqr(KSq)+Sqr(KpSq))/2;
    Result := Result + IxxmIyy/4 - deltaJ*JSq - deltaK*(KSq+KpSq)/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else
    Result := 0;
end;

function TAsymmetricTop.HRotS(J, Kp, K: TQuantumNumber): Double;
{ Hamiltonian using Watson's S reduction }
var
  Iyy, Ixx, JSq, IxxmIyy: Double;
begin
  if K = Kp then
    Result := HRotA(J, KP, K)
  else if Abs(K-Kp) = 4 then begin
    Ixx := GetParameter(LSP+Ord(XAxisFor[Molecule.Representation]));
    Iyy := GetParameter(LSP+Ord(YAxisFor[Molecule.Representation]));
    IxxmIyy := Ixx-Iyy + BDelta;
    JSq := J*(J+2) / 4;
    Result := IxxmIyy/4 + (deltaJ + (phiJ + llJ*JSq)*JSq)*JSq;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if Abs(K-Kp) = 8 then begin
    JSq := J*(J+2) / 4;
    Result := (deltaK + (phiJK + llJK*JSq)*JSq)*rl(J, K, Kp)
  end else if Abs(K-Kp) = 12 then begin
    JSq := J*(J+2) / 4;
    Result := (phiK + llKJ*JSq)*rl(J, K, Kp)
  end else if Abs(K-Kp) = 16 then
    Result := llK*rl(J, K, Kp)
  else
    Result := 0;
end;

function TAsymmetricTop.HSR(J, Np, Kp, N, K:TQuantumNumber): Double;
var
  i, rank, q: Integer;
  r, r1, Ksq, Nsq, sgn: Double;
  NN: TQuantumNumber;
  function Gamma(N, S, J: TQuantumNumber):Double;
  begin
    Result := (J*(J+2) - N*(N+2) - S*(S+2))/4;
  end;
  function phi(N,S,J:TQuantumNumber):Double;
  begin
    Result := (N - J + S);
    Result := Result * (N + J + S + 2) * (S + J - N + 2);
    Result := Sqrt(Result * (N + J - S) / (N - 1) / (N + 1) / 16);
  end;
begin
  Result := 0;
  if Abs(Np-N) > 4 then Exit;
  if S > 1 then begin
    { Matrix elements of Spin-Spin Coupling from Hirota eq 2.3.36, 2.3.37 }
    { Adapted taking into account equation 38 of T.J Sears, Comp. Phys. Rep, 2, 1 (1984) }
    { The reduced matrix element <S||T2(S)||S> is also included. Hirota assumes this is 1, which
      is appropriate for a triplet state while Sears assumes a more complex dependence depending
      on the specific make-up of the spins, which will (presumably) also be 1 for a triplet state.
      The values for the lowest few multiplicies are:
        S  <S||T2(S)||S>^2
        1      1
        1.5    6
        2     21
        2.5   56
        3    126
    }
    q := Kp-K;
    case q of
      0: Result := Sqrt(30)*alpha;
      4, -4: Result := Sqrt(6)*beta;
    end;
    if Result <> 0 then
      Result := Result * Minus(N+S+J+Np-Kp) * sixj(J, S, Np, 4, N, S)
                       * Sqrt((N+1)*(Np+1)) * threej(Np, -Kp, 4, q, N, K)
                       * ReducedMatrixElement(4, S);
  end;
  if Abs(Np-N) > 2 then Exit;
  r := Sqrt(S*(S+2)*(S+1)*(N+1)*(Np+1)/4)*Minus(J+S+Np)*sixj(N, S, J, S, Np, 2);
  if Abs(K-Kp) <= 4 then
    for i := 0 to 1 do begin
      { rank = 1 does not contribute }
      rank := 4*i;
      q := (Kp - K) div 2;
      if Abs(2*q) <= rank then begin
        r1 := (Minus(rank)*Sqrt(N*(N+2)*(N+1)/4)*sixj(2, 2, rank, Np, N, N)
               +Sqrt(Np*(Np+2)*(Np+1)/4)*sixj(2, 2, rank, N, Np, Np))*Sqrt(Rank+1)/2;
//        if (i=1) and (Abs(q) = 2) and (N <> Np) then begin
//          NN := Max(N,Np);
//          sgn := (Kp-K) div 4;
//          if NN = Np then begin
//            { <N+1|HSR|N> matrix element wanted - swap }
//            KK := Kp;
//            sgn := -sgn;
//          end else
//            KK := K;
//          r2 := sgn*ds2*(KK*(NN+sgn*KK)+(KK+sgn*4)*(NN+sgn*KK+4))/4;
//          r2 := r2*r*r1*Minus(Np-Kp)*threej(Np,-Kp, rank, 2*q, N, K);
//        end;
//        if (i=1) and (Abs(q) = 2) and (N = Np) then
//          r2 :=  ds2*(Sqr(K)+Sqr(Kp))/4;
        Result := Result + r*r1*Minus(Np-Kp)*threej(Np,-Kp, rank, 2*q, N, K)*TSR[2*i][q];
      end;
      { Seems to be a phase problem with the delta N = +/1 delta K +/- 1 matrix elements,
        compared with acdd. However, it checks OK with explicit evaluation with Maple,
        and ACDD seems to have a spcific problem with these matrix elements, in that they
        have not had their sign changed with respect to W. T. Raynes, JCP 41, 3020 (1964),
        and the ACDD documentation suggests that they should }
    end;
  { Quartic spin rotation - Matrix elements from J. M. Brown and T. J. Sears, JMS, 75, 111 (1979) }
  if QuarticSpinRotation then begin
    Nsq := N*(N+2)/4;
    if Np = N then begin
      if N <> 0 then begin
        if K = Kp then begin
          Ksq := Sqr(K)/4;
          Result := Result + Gamma(N,S,J)/2/Nsq*( DsK*Sqr(Ksq) + (DsNK+DsKN)*Ksq*Nsq + DsN*Sqr(Nsq) );
        end else if Abs(K-Kp) = 4 then
          if Molecule.SReduction then
            Result := Result + ds1/2 * Gamma(N,S,J) * rl(N,K,Kp)
          else
            Result := Result + (ds1{=dsN}*Nsq +  ds2{=dsK}*(Sqr(K)+Sqr(Kp))/8) * Gamma(N,S,J)/2/Nsq * rl(N,K,Kp)
        else if (Abs(K-Kp) = 8) and Molecule.SReduction then
          Result := Result + ds2/2 * Gamma(N,S,J)/Nsq * rl(N,K,Kp);
      end;
    end else { Abs(N-Np) = 2 }
      if K = Kp then begin
        NN := Max(N,Np);
        Ksq := Sqr(K)/4;
        Result := Result - K/2/NN * (DsK*Ksq + DsNK*Sqr(NN)/4)*phi(NN,S,J)*Sqrt(Sqr(NN)-Sqr(K))/2;
      end else if (Abs(K-Kp) = 4) and not Molecule.SReduction then begin
        NN := Max(N,Np);
        sgn := (Kp-K) div 4;
        if NN = Np then begin
          { <N+1|HSR|N> matrix element wanted - swap }
          K := Kp;
          sgn := -sgn;
        end;
        Result := Result - ds2{=dsK} / NN / 2 * phi(NN,S,J)
                         *(K*(NN+sgn*K) + (K+sgn*4)*(NN+sgn*K+4))
                         *Sqrt((NN-sgn*K-2) * (NN+sgn*K+2) * (NN-sgn*K-4) * (NN-sgn*K)) / 16;
      end else if (Abs(K-Kp) = 8) and Molecule.SReduction then begin
        NN := Max(N,Np);
        sgn := (Kp-K) div 8;
        if NN = Np then begin
          { <N+1|HSR|N> matrix element wanted - swap }
          K := Kp;
          sgn := -sgn;
        end;
        Result := Result - sgn * ds2 / NN * phi(NN,S,J)
                           *Sqrt((NN-sgn*K) * (Sqr(NN)-Sqr(K+2*sgn)) * (Sqr(NN)-Sqr(K+4*sgn))
                                 * (Sqr(NN)-Sqr(K+6*sgn)) * (NN-sgn*K-8) / 256 )
      end;
  end;
end;

function TAsymmetricTop.HDip(TransitionMoment: TTransitionMoment; Jp, Kp, J, K: TQuantumNumber): Double;
var
  q: TQuantumNumber;
  Phase: Integer;
begin
  Result := TransitionMoment.Strength;
  Phase := 0;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).Component;
    { To match symmetries, treat + and - q as actually T(k,|q|) +- T(k,-|q|) }
    if (Kp < K) and (q < 0) then
      Phase := 2;
  end else if TransitionMoment is TCartesianTransitionMoment then begin
    case Molecule.xyzAxis(TCartesianTransitionMoment(TransitionMoment).Axis) of
      zAxis: q := 0;
      yAxis: q := 2;
      xAxis: begin
               q := 2;
               if K < Kp then
                 Phase := 2;
             end;
    end;
    if q <> 0 then
      Result := Result / Sqrt(2.0);
    { TODO: Check for complex matrix elements }
  end else
    raise NotYet;
  if Abs(q) = Abs(Kp-K) then
    Result := Result * Sqrt((J+1.0)*(Jp+1)/(TTransitionMoment(TransitionMoment).Rank+1))
              *Minus(J-K+q+Phase)*threej(Jp, Kp, TransitionMoment.Rank, K-Kp, J, -K)
  else
    Result := 0;
end;

function TAsymmetricTop.HyperfineMatrixElement(A, B: TBasisState; Signp,
  Sign: Integer; MNode: TMNode; F: TQuantumNumber): Double;
var
  Bra, Ket: TAsymmetricTopBasisState;
  Nuc: TAsymmetricTopNucleus;
  RankWanted: TQuantumNumber;
  JP, J, Np, N, Kp, K: TQuantumNumber;
  t: Double;
begin
  Bra := A as TAsymmetricTopBasisState;
  Ket := B as TAsymmetricTopBasisState;
  Nuc := MNode as TAsymmetricTopNucleus;
  JP := Bra.J;
  J := Ket.J;
  NP := Bra.N;
  N := Ket.N;
  KP := SignP*Bra.K;
  K := Sign*Ket.K;
  Result := 0;
  RankWanted := F mod 10;
  if RankWanted = 2 then begin
    { Magnetic Dipole }
    if (Np = N) and (Kp=K) then
      Result := Result + Minus(N+S+JP+2)
                * sixj(S, Jp, N, J, S, 2) * Nuc.aF;
    if Abs(K-Kp) <= 4 then begin
      t := ninej(Np, N, 4, S, S, 2, Jp, J, 2); { FPC Code generation problem? }
      t := t * threej(Np, -Kp, 4, Kp-K, N, K);
      Result := Result + Sqrt(30 * (N+1.0)*(NP+1.0)) * t
                         * Minus(Np-Kp) * Nuc.TT[(Kp-K) div 2];
    end;
    Result := Result * Sqrt((J+1.0)*(JP+1.0)) * ReducedMatrixElement(2, S) * ReducedMatrixElement(2, Nuc.Spin);
  end else if RankWanted = 4 then begin
    { Quadrupole }
  end;
  if IsNan(Result) then
    raise Exception.Create('Help');
end;

function TAsymmetricTop.RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double;
var
  Bra, Ket: TAsymmetricTopBasisState;
  K, Kp: TQuantumNumber;
  H: function (J, Kp, K: TQuantumNumber): Double of object;
begin
  Bra := A as TAsymmetricTopBasisState;
  Ket := B as TAsymmetricTopBasisState;
  Result := 0;
  case TMatElemType(Flag) of
    meHrot:  if Molecule.SReduction then
               H := {$IFDEF FPC}@{$ENDIF}HrotS
             else
               H := {$IFDEF FPC}@{$ENDIF}HrotA;
    meWidth: H := {$IFDEF FPC}@{$ENDIF}HWidth;
  else
    raise NotYet;
  end;
  if Bra.J <> Ket.J then
    raise Exception.Create('TAsymmetricTop.RotationalMatrixElement');
  Kp := SignA*Bra.K;
  K := SignB*Ket.K;
  if Bra.N = Ket.N then
    Result := H(Bra.N, Kp, K);
  if (Flag = Ord(meHrot)) { TODO: Spin dependent width model? }
     and (S <> 0) and (Abs(Bra.N-Ket.N) <= 4) then
    Result := Result + HSR(Bra.J, Bra.N, Kp, Ket.N, K);
  { TODO: C1 symmetry requires all four matrix elements or explicit test for identical phase? }
end;


function TAsymmetricTop.RotationalTransitionMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Double;
var
  Bra, Ket: TAsymmetricTopBasisState;
  TransitionMoment: TTransitionMoment;
begin
  Bra := A as TAsymmetricTopBasisState;
  Ket := B as TAsymmetricTopBasisState;
  TransitionMoment := MNode as TTransitionMoment;
  Result := HDip(TransitionMoment, Bra.N, SignA*Bra.K, Ket.N, SignB*Ket.K);
  if S > 0 then { TODO: Scope for pushing this one level up, as it will be the same for all K combinations }
    Result := Result * Minus(Bra.N+S+Ket.J+TransitionMoment.Rank) * Sqrt((Bra.J+1)*(Ket.J+1)) * sixj(Bra.N, Bra.J, S, Ket.J, Ket.N, TransitionMoment.Rank);
end;

function TAsymmetricTop.RotationalPerturbationMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Double;
var
  Bra, Ket: TAsymmetricTopBasisState;
  P: TAsymmetricTopPerturbation;
begin
  if Flag <> Ord(meHrot) then
    raise NotYet;
  Bra := A as TAsymmetricTopBasisState;
  Ket := B as TAsymmetricTopBasisState;
  P := MNode as TAsymmetricTopPerturbation;
  if Bra.J <> Ket.J then
    raise Exception.Create('TAsymmetricTop.PerturbationMatrixElement');
  Result := P.MatrixElement(Bra.J, Bra.N, SignA*Bra.K, Ket.N, SignB*Ket.K)
end;

function TAsymmetricTop.HasWidth: Boolean;
var
  i: Integer;
begin
  Result := inherited HasWidth;
  if not Result then
    for i := FirstWidthPar to nParameters do
      if Parameters[i] <> 0 then begin
        Result := True;
        Exit;
      end;
end;

const
  AsymmetricTopSettingNames: array[LSI..LSI+2] of string = (
'Symmetry',
'Minimum |K| to use; -1 (default) for no lower limit.',
'Maximum |K| to use; -1 (default) for no upper limit.'
);

function TAsymmetricTop.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopSettingNames))
     and (AIndex <= High(AsymmetricTopSettingNames)) then
    Result := AsymmetricTopSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 AsymmetricTopParameterNames: array [LSP..FirstWidthPar+4] of string = (
'A rotational constant',
'B rotational constant',
'C rotational constant',
'Average of rotational constants perpendicular to K axis - Alternate form for rotational constants',
'Difference between rotational constants perpendicular to K axis - Alternate form for rotational constants',
'K^4 Quartic Centrifugal Distortion',
'J(J+1)K^2 Quartic Centrifugal Distortion',
'J^2(J+1)^2 Quartic Centrifugal Distortion',
'deltaK - A/S reduction dependent off diagonal Quartic Centrifugal Distortion',
'deltaJ - A/S reduction dependent off diagonal Quartic Centrifugal Distortion',
'K^6 Sextic Centrifugal Distortion',
'J(J+1)K^4 Sextic Centrifugal Distortion',
'J^2(J+1)^2K^2 Sextic Centrifugal Distortion',
'J^3(J+1)^3 Sextic Centrifugal Distortion',
'phiK - A/S reduction dependent off diagonal Sextic Centrifugal Distortion',
'phiJK - A/S reduction dependent off diagonal Sextic Centrifugal Distortion',
'phiJ - A/S reduction dependent off diagonal Sextic Centrifugal Distortion',
    { Octic Centrifugal Distortion }
'K^8 Octic Centrifugal Distortion',
'J(J+1)K^6 Octic Centrifugal Distortion',
'J^2(J+1)^2K^4 Octic Centrifugal Distortion',
'J^3(J+1)^3K^2 Octic Centrifugal Distortion',
'J^4(J+1)^4 Octic Centrifugal Distortion',
'llK - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
'llKJ - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
'llJK - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
'llJ - A/S reduction dependent off diagonal Octic Centrifugal Distortion',
    { Spin-Rotation Parameters }
'aa spin-rotation Parameter',
'bb spin-rotation Parameter',
'cc spin-rotation Parameter',
'(eab + eba)/2',
'(eac + eca)/2',
'(ebc + ebc)/2',
'(DeltasN in A reduction) quartic spin rotation parameter',
'(DeltasNK in A reduction) quartic spin rotation parameter',
'(DeltasKN in A reduction) quartic spin rotation parameter',
'(DeltasK in A reduction) quartic spin rotation parameter',
'(deltasN in A reduction) quartic spin rotation parameter',
'(deltasK in A reduction) quartic spin rotation parameter',
'Diagonal spin-spin coupling constant (=D/3)',
'Off-Diagonal spin-spin coupling constant (=E)',
'multiplier of <Ja^2> for linewidth',
'multiplier of <Jb^2> for linewidth',
'multiplier of <Jc^2> for linewidth',
'multiplier of <J(J+1)-K^2> for linewidth',
'multiplier of <J+^2+J-^2>/4 for linewidth'
);

function TAsymmetricTop.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopParameterNames))
     and (AIndex <= High(AsymmetricTopParameterNames)) then
    Result := AsymmetricTopParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TAsymmetricTop.zIndex: Integer;
begin
  Result := LSP + Ord(zAxisFor[Molecule.Representation]);
end;

{ TAsymmetricManifold }

class function TAsymmetricManifold.OwnerType: TComponentClass;
begin
  Result := TAsymmetricMolecule;
end;

function TAsymmetricManifold.FirstSym(J: TQuantumNumber):Integer;
begin
  Result := 0;
end;

function TAsymmetricManifold.NextSym(J: TQuantumNumber; Sym: Integer):Integer;
begin
  Result := Sym+1;
  if Result >= nSym then
    Result := -1;
end;

function TAsymmetricManifold.nSym: Integer;
begin
  Result := TAsymmetricMolecule(Owner).LastSym+1;
end;

function TAsymmetricManifold.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case AIndex of
    0: Result := TAsymmetricTop;
    1: Result := TAsymmetricTopPerturbation;
  else
    Result := nil;
  end;
end;

{ TAsymmetricMolecule }

function TAsymmetricMolecule.GetPointGroup(Index: integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

function TAsymmetricMolecule.GetRepresentation(Index: integer): TRepresentation;
begin
  Result := TRepresentation(GetSetting(Index));
end;

procedure TAsymmetricMolecule.SetPointGroup(Index: integer; const AValue: TPointGroup);
begin
  SetSetting(Index, Ord(AValue));
end;

procedure TAsymmetricMolecule.SetRepresentation(Index: integer; const AValue: TRepresentation);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TAsymmetricMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := 'PointGroup=C1,Representation=Ir,eeWt=1,eoWt=1,oeWt=1,ooWt=1,C2xAxis=c,' + SavedProperties;
  inherited Create(AOwner);
end;

function TAsymmetricMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  { TODO: Copy to individual matrix element }
  if FakeSym then
    { Check for equal Statistical Weights }
    { TODO: may need modulating by individual state symmetries? }
    Result := GetSetting(FirstStatisticalWeight+Sym1)
              = GetSetting(FirstStatisticalWeight+Sym2)
  else if PseudoC2v then
    Result := True
  else if not TransitionMoment.Magnetic and (Minus(TransitionMoment.Rank) < 0) then
    { This is essentially testing for a parity change }
    Result := (OddRankSym xor Sym1 xor Sym2) = 0
  else
    Result := Sym1 = Sym2;
end;

function TAsymmetricMolecule.FixupProperties: Exception;
begin
  if PseudoC2v then
    FOddRankSym := eoSym(1 {=A2})
  else
    case PointGroup of
      C1:  FOddRankSym := 0;
      Ci:  FOddRankSym := 1;
      C2:  FOddRankSym := 0;
      Cs:  FOddRankSym := 1;
      D2:  FOddRankSym := 0;
      C2v: FOddRankSym := eoSym(1 {=A2});
      C2h: FOddRankSym := 2; { Au }
      D2h: FOddRankSym := 4; { Au }
    else
      raise NotYet;
    end;
  Result := Inherited FixupProperties;
end;

function TAsymmetricMolecule.GetAxis(const Index: Integer): TabcAxis;
begin
  Result := TabcAxis(GetSetting(Index));
end;

procedure TAsymmetricMolecule.SetAxis(const Index: Integer; const AValue: TabcAxis);
begin
  SetSetting(Index, Ord(AValue));
end;

function TAsymmetricMolecule.GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar;
const
  AxisNames: array [0..3] of PChar = ('a', 'b', 'c', nil);
begin
  if AIndex in [LMI+7, LMI+8] then
    Result := @AxisNames
  else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

const
                                                   {Ir, IIr, IIIr, Il, IIl, IIIl}
  JAdjustMask: array[TRepresentation] of Integer = ( 1,   3,    2,  1,   3,    2);

function TAsymmetricMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
  if Mixture.JAdjustSym and not Odd(OddSpin) { TODO: Implement for odd spin } and (Minus(J) < 0) then
    if PseudoC2v then
      Result := Result xor JAdjustMask[Representation]
    else
      case PointGroup of
        D2, C2v, D2h: Result := Result xor JAdjustMask[Representation];
        { Three cases for Cs:
          1. The J alternation flips bewteen the two C2v symmetries that are in the same matrix - don't alternate
          2. The J alternation flips to the other matrix for both symmetries - alternate
          3. The J alternation flips one symmetry but not the other to the other matrix - In this case
             neither choice is likely to make sense, so alternate, so user has choice }
        Cs, C2h, C2: if JAdjustMask[Representation] <> SymWrt[C2zAxis] then
                       Result := Result xor 1;
        C1, Ci: { All matrices the same }
      end;
end;

function TAsymmetricMolecule.LastSym: Integer;
begin
  if PseudoC2v then
    Result := 3
  else
    case PointGroup of
      C1: Result := 0;
      Ci,C2,Cs: Result := 1;
      D2,C2v,C2h: Result := 3;
      D2h: Result := 7;
    else
      raise NotYet;
    end;
end;

const
  AsymmetricMoleculeSettingNames: array[LMI..LMI+10] of string = (
'Point group - C1, Ci, C2, Cs, D2, C2v, C2h or D2h',
'Representation - Ir, IIr, IIIr, Il, IIl or IIIl',
'Use Watson''s S reduction (as opposed to A).',
'Statistical weight for ee levels',
'Statistical weight for eo levels',
'Statistical weight for oe levels',
'Statistical weight for oo levels',
'a, b, or c; C2 axis in C2v; out of plane axis in Cs',
'a, b, or c; out of plane axis in C2v',
'Ignore any operators breaking C2v symmetry.',
'Turn off transition moment test.'
);

function TAsymmetricMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricMoleculeSettingNames))
     and (AIndex <= High(AsymmetricMoleculeSettingNames)) then
    Result := AsymmetricMoleculeSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

class function TAsymmetricMolecule.ManifoldType: TMNodeClass;
begin
  Result := TAsymmetricManifold;
end;

{ TAsymmetricTopBasisState }

function TAsymmetricTopBasisState.BasisLabel: string;
begin
  Result := inherited BasisLabel + ' ';
  if TAsymmetricTop(State).S <> 0 then
    Result := Result + State.Manifold.Mixture.QuantumNumberToStr(N) + ' ';
  Result := Result + State.Manifold.Mixture.QuantumNumberToStr(K);
end;

function TAsymmetricTopBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey;
  if TAsymmetricTop(State).S <> 0 then
    Result := Result + ' N';
  Result := Result + ' K' + AxisNames[ZAxisFor[TAsymmetricMolecule(State.Manifold.Molecule).Representation]];
end;

function TAsymmetricTopBasisState.StateLabel: string;
begin
  Result := Format('%s %2s %2s %2s', [inherited StateLabel, State.Manifold.Mixture.QuantumNumberToStr(N), State.Manifold.Mixture.QuantumNumberToStr(Ka), State.Manifold.Mixture.QuantumNumberToStr(Kc)]);
  if TAsymmetricTop(State).S <> 0 then
    Result := Result + Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(J)]);
end;

function TAsymmetricTopBasisState.StateLabelKey: string;
begin
  Result := inherited StateLabelKey;
  if TAsymmetricTop(State).S <> 0 then
    Result := Result + ' N Ka Kc J'
  else
    Result := Result + ' J Ka Kc'
end;

function TAsymmetricTopBasisState.TrackIndex: Integer;
begin
 Result := Abs(K);
 if K < 0 then Inc(K);
end;

function TAsymmetricTopBasisState.TxName(Upper, Lower: TBasisState): string;
var
  Molecule: TAsymmetricMolecule;
  U, L: TAsymmetricTopBasisState;
  US, LS: TAsymmetricTop;
  fU, fL: string;
  function Fnumber(State: TAsymmetricTopBasisState):string;
  var
    S: TQuantumNumber;
  begin
    S := (State.State as TAsymmetricTop).S;
    if S > 0 then
      Result := IntToStr((State.N - State.J + S + 2) div 2)
    else
      Result := '';
  end;
begin
  Molecule := State.Manifold.Molecule as TASymmetricMolecule;
  U := Upper as TASymmetricTopBasisState;
  US := U.State as TASymmetricTop;
  L := Lower as TASymmetricTopBasisState;
  LS := L.State as TASymmetricTop;
  Result := LowerCase(DeltaJName(Abs(U.K)-Abs(L.K)))+DeltaJName(U.N-L.N);
  if (US.S > 0) or (LS.S > 0) then begin
    { Add spin label }
    fU := Fnumber(U);
    fL := Fnumber(L);
    Result := Result + fU;
    if fU <> fL then { Add second spin label if different }
      Result := Result + fL;
  end;
  Result := Result + AtLeast(L.Ka, 2) + AtLeast(L.Kc, 2);
  Result := Result + Format('(%2s)',[Molecule.QuantumNumberToStr(L.N)]);
  { TODO: Hyperfine }
end;

function TAsymmetricTopBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := 'deltaK DeltaN';
  if Upper.State.Manifold.MaxSpin > 0 then
    Result := Result + ' Fn''';
  if Lower.State.Manifold.MaxSpin > 0 then
    Result := Result + ' Fn"';
  Result := Result + ' Ka" Kc" (J")';
end;

{ TAsymmetricTopPerturbation }

function TAsymmetricTopPerturbation.MatrixElement(J, Np, Kp, N, K: TQuantumNumber): Double;
var
  xPower, yPower, zPower: Integer;
  dK: Integer;
  Kpp, KppMax: TQuantumNumber;
begin
  xPower := GetSetting(LPI+1+Ord(XAxisFor[Molecule.Representation]));
  yPower := GetSetting(LPI+1+Ord(YAxisFor[Molecule.Representation]));
  zPower := GetSetting(LPI+1+Ord(ZAxisFor[Molecule.Representation]));
  dK := Abs(K-Kp) div 2;
  Result := 0;
  if (N=Np) and (dK <= xPower + yPower + pPower) and not Odd(dK+xPower+yPower+pPower) then begin
    { Actually work out [Jz,[Jy,Jx]+]+ }
    if pPower > 0 then begin
      if (xPower or yPower) <> 0 then
        raise EReportException.Create('AsymmetricTopPerturbation: pPower can''t be used with xPower or yPower');
      if dK = pPower then
        Result := rl(N, Kp, K)
      else
        Result := 0;
    end else if xPower > 0 then
      if yPower > 0 then begin
        KppMax := Min(Kp+2*yPower, K+2*xPower);
        Kpp := Max(Kp-2*yPower, K-2*xPower);
        while Kpp <= KppMax do begin
          Result := Result + rly(yPower,N,Kp,Kpp)*rlx(xPower,N,Kpp,K);
          Inc(Kpp,2);
        end;
        KppMax := Min(Kp+2*xPower, K+2*yPower);
        Kpp := Max(Kp-2*xPower, K-2*yPower);
        while Kpp <= KppMax do begin
          Result := Result + rlx(xPower,N,Kp,Kpp)*rly(yPower,N,Kpp,K);
          Inc(Kpp,2);
        end;
      end else
        Result := rlx(xPower,N,Kp,K)
    else if yPower > 0 then
      Result := rly(yPower,N,Kp,K)
    else Result := 1;
    if zPower > 0 then
      if (xPower or yPower or pPower) <> 0 then
        Result := Result * (IntPower(K,zPower) + IntPower(Kp,zPower))
      else
        Result := Result * IntPower(K,zPower);
    if JPower > 0 then
      Result := Result * IntPower(N*(N+2), JPower div 2);
    Result := Result * Value / IntPower(2,zPower + JPower);
//    if (xPower = 1) and (yPower = 0) and (zPower = 1) then begin
//      Result := Value * (K + Kp)*rl(N, K, Kp);
  end;
end;

function TAsymmetricTopPerturbation.Molecule: TAsymmetricMolecule;
begin
  Result := Owner.Owner as TAsymmetricMolecule;
end;

function TAsymmetricTopPerturbation.PairName: string;
  function Show(Letter: Char; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Letter;
    else
      Result := Letter + IntToStr(Power)
    end;
  end;
begin
  Result := Show('J',JPower)
            + Show('a',aPower) + Show('b',bPower) + Show('c',cPower)
            + Show('+', pPower);
end;

const
  AsymmetricTopPerturbationSettingNames: array[LPI..LPI+4] of string = (
'Twice power of J(J+1)',
'Power of Ja',
'Power of Jb',
'Power of Jc',
'Power of J+-'
);

function TAsymmetricTopPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopPerturbationSettingNames))
     and (AIndex <= High(AsymmetricTopPerturbationSettingNames)) then
    Result := AsymmetricTopPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TAsymmetricTopPerturbation.SetAxisPower(Axis: TxyzAxis; AValue: Integer);
begin
  SetSetting(LPI+1+Ord(abcAxis(Molecule.Representation, Axis)), AValue);
end;

{ TAsymmetricTopNucleus }

const
  AsymmetricTopNucleusParameterNames: array[1..6] of string = (
'Nuclear spin - electron spin Fermi Contact interaction (I.S).',
'Tzz = -(Txx+Tyy) component of Nuclear spin - electron spin dipole dipole interaction.',
'Txx-Tyy component of Nuclear spin - electron spin dipole dipole interaction.',
'ab component of Nuclear spin - electron spin dipole dipole interaction.',
'ac component of Nuclear spin - electron spin dipole dipole interaction.',
'bc component of Nuclear spin - electron spin dipole dipole interaction.'
);

function TAsymmetricTopNucleus.FixupProperties: Exception;
var
  Txy, Txz, Tyz: Double;
  Representation: TRepresentation;
  function efind(Axis1, Axis2: TabcAxis):Double;
  var
    s: string;
  begin
    if Axis1 < Axis2 then
      s := AxisNames[Axis1] + AxisNames[Axis2]
    else
      s := AxisNames[Axis2] + AxisNames[Axis1];
    Result := GetParameter(IndexOfParameter('T'+s))
  end;
begin
  Representation := (Molecule as TAsymmetricMolecule).Representation;
  Txy := efind(XAxisFor[Representation], YAxisFor[Representation]);
  Txz := efind(XAxisFor[Representation], ZAxisFor[Representation]);
  Tyz := efind(YAxisFor[Representation], ZAxisFor[Representation]);

  TT[ 0] := Tzz/2;
  TT[ 1] := -(Txz + Tyz)/Sqrt(6);
  TT[-1] :=  (Txz - Tyz)/Sqrt(6);
  TT[ 2] :=  (Txxmyy + Txy)/Sqrt(24);
  TT[-2] :=  (Txxmyy - Txy)/Sqrt(24);
  Result:=inherited FixupProperties;
end;

function TAsymmetricTopNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(AsymmetricTopNucleusParameterNames))
     and (AIndex <= High(AsymmetricTopNucleusParameterNames)) then
    Result := AsymmetricTopNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

initialization
  RegisterClasses([TAsymmetricMolecule, TAsymmetricManifold, TAsymmetricTop, TAsymmetricTopPerturbation, TAsymmetricTopNucleus]);
  TSpecies.AddMoleculeType(TAsymmetricMolecule);
end.

