unit SymT;
{ Symmetric Top }

{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses CLasses, SysUtils, MNode, Mixture, QuantumNumber;

const
  WTI = LMI+1; { Index of first statistical weight }

type

  TPointGroup = (C3v=0, C4v, C5v, C6v, C7v, C8v, D3h, D4h, D5h, D6h, D7h, D8h, D2d, D3d, D4d, D5d, D6d, D7d, D8d);

  TCharacterTableEntry = record
    Name: AnsiString;
    g, rho, eta: Integer;
  end;

  TCharacterTable = record
    GroupName: string;
    Size: Integer;
    n: Integer;         { Highest order rotation axis }
    gKmult: Integer;    { multiplier of k in expression for g}
    rhoKmult: Integer;  { multiplier of k in expression for rho}
    etaKmult: Integer;  { multiplier of k in expression for eta}
    DipoleSym: Integer; { Symmetry of space fixed dipole moment component }
    Names: array of PChar;
    IReps: array of TCharacterTableEntry;
    Degenerate: array of Boolean;
    Group: array of Integer; { Statistical weight group number for each irep }
    GroupNames: array of string;
  end;

  TSymmetricMolecule = class;

  { TSymmetricTop }

  TSymmetricTop = class (TState)
  private
   FMolecule: TSymmetricMolecule;
   FProlate: Boolean;
  public
    property Prolate: Boolean read FProlate;
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState:THyperfineBasis): Integer; override;
    function GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar; override;
    function HRot(J, Kp, lp, K, l: TQuantumNumber): Double;
    function HDip(TransitionMoment: TTransitionMoment; Jp, Kp, lp, J, K, l: TQuantumNumber): Double;
    // function MatrixElement(ABra, AKet: TBasisState; MatElemType: TMatElemType): Double; override;
    // function TransitionMatrixElement(TransitionMoment: TTransitionMoment; ABra, AKet: TBasisState): Double; override;

    function RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double; override;
    function RotationalTransitionMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Double; override;

    property Molecule: TSymmetricMolecule read FMolecule;
    function OddSpin: TQuantumNumber; override;
    class function OwnerType: TComponentClass; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Symmetry: Integer index LSI read GetSetting write SetSetting;
    property B: Double index LSP   read GetParameter write SetParameter;
    property C: Double index LSP+1 read GetParameter write SetParameter;
    { Quartic Centrifugal Distortion }
    property DJ:      Double index LSP+2 read GetParameter write SetParameter;
    property DJK:     Double index LSP+3 read GetParameter write SetParameter;
    property DK:      Double index LSP+4 read GetParameter write SetParameter;
    { Coriolis }
    property zeta:    Double index LSP+5 read GetParameter write SetParameter;
    property etaJ:    Double index LSP+6 read GetParameter write SetParameter;
    property etaK:    Double index LSP+7 read GetParameter write SetParameter;
    { l doubling }
    property qplus:   Double index LSP+ 8 read GetParameter write SetParameter;
    property qminus:  Double index LSP+ 9 read GetParameter write SetParameter;
    property r:       Double index LSP+10 read GetParameter write SetParameter;
    property DqJ:     Double index LSP+11 read GetParameter write SetParameter;
    property DqK:     Double index LSP+12 read GetParameter write SetParameter;
    { Sextic Centrifugal Distortion }
    property HJ:    Double index LSP+13 read GetParameter write SetParameter;
    property HJK:   Double index LSP+14 read GetParameter write SetParameter;
    property HKJ:   Double index LSP+15 read GetParameter write SetParameter;
    property HK:    Double index LSP+16 read GetParameter write SetParameter;
  end;

  TSymmetricTopBasisState = class(TBasisState)
  public
    K, l: TQuantumNumber;
    function klLabel: string;
    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel: string; override; { State name, assuming energy ordering preserved }
    function StateLabelKey: string; override; { Explanation of State Label }
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as qRK(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
  end;

  TSymmetricManifold = class(TManifold)
  protected
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
  public
    class function OwnerType: TComponentClass; override;
    function FirstSym(J: TQuantumNumber):Integer; override;
    function NextSym(J: TQuantumNumber; Sym: Integer):Integer; override;
    function nSym: Integer; override; { Number of different symmetries }
  end;

  TSymmetricMolecule = class(TMolecule)
  private
    FCharacterTable: TCharacterTable;
    function GetPointGroup(const Index: Integer): TPointGroup;
    procedure SetPointGroup(const Index: Integer; const AValue: TPointGroup);
  protected
    function GetSettingNames(AIndex: Integer): string; override;
  public
    property CharacterTable: TCharacterTable read FCharacterTable;
    constructor Create(AOwner: TComponent); override;
    class function ManifoldType: TMNodeClass; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    function SymName(Sym: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property PointGroup:     TPointGroup index LMI read GetPointGroup write SetPointGroup;
    property Wt0:            Integer  index WTI+0  read GetSetting write SetSetting;
    property Wt1:            Integer  index WTI+1  read GetSetting write SetSetting;
    property Wt2:            Integer  index WTI+2  read GetSetting write SetSetting;
    property Wt3:            Integer  index WTI+3  read GetSetting write SetSetting;
    property Wt4:            Integer  index WTI+4  read GetSetting write SetSetting;
    property Wt5:            Integer  index WTI+5  read GetSetting write SetSetting;
    property Wt6:            Integer  index WTI+6  read GetSetting write SetSetting;
    property Wt7:            Integer  index WTI+7  read GetSetting write SetSetting;
    property Wt8:            Integer  index WTI+8  read GetSetting write SetSetting;
    property Wt9:            Integer  index WTI+9  read GetSetting write SetSetting;
    property Wt10:           Integer  index WTI+10 read GetSetting write SetSetting;
    property Wt11:           Integer  index WTI+11 read GetSetting write SetSetting;
    property Wt12:           Integer  index WTI+12 read GetSetting write SetSetting;
    property Wt13:           Integer  index WTI+13 read GetSetting write SetSetting;
  end;

const
  MaxnStatWeight = 14; { Maximum number of statistical weights }

implementation

uses nj;

{ This implements Table 12-14 and the surrounding logic of
  P.R. Bunker and Per Jensen, Molecular Symmetry and Spectroscopy, 2nd Edition.
  See also http://www.mpip-mainz.mpg.de/~gelessus/group.html
}

const

  CharacterTableNames: array [TPointGroup] of PChar = ('C3v', 'C4v', 'C5v', 'C6v', 'C7v', 'C8v', 'D3h', 'D4h', 'D5h', 'D6h', 'D7h', 'D8h', 'D2d', 'D3d', 'D4d', 'D5d', 'D6d', 'D7d', 'D8d');

  C3vCharacterTable: array [0..2] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  C5vCharacterTable: array [0..3] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  C7vCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  C4vCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:3; rho:0; eta:0),
   (Name:'B2'; g:3; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  C6vCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:3; rho:0; eta:0),
   (Name:'B2'; g:3; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0));

  C8vCharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D3hCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E''';  g:1; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E"';   g:1; rho:0; eta:1));

  D5hCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1));

  D7hCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1'''; g:0; rho:0; eta:0),
   (Name:'A2'''; g:0; rho:1; eta:0),
   (Name:'E1'''; g:1; rho:0; eta:0),
   (Name:'E2'''; g:2; rho:0; eta:0),
   (Name:'E3'''; g:3; rho:0; eta:0),
   (Name:'A1"';  g:0; rho:0; eta:1),
   (Name:'A2"';  g:0; rho:1; eta:1),
   (Name:'E1"';  g:1; rho:0; eta:1),
   (Name:'E2"';  g:2; rho:0; eta:1),
   (Name:'E3"';  g:3; rho:0; eta:1));

  D4hCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:2; rho:0; eta:0),
   (Name:'B2g'; g:2; rho:1; eta:0),
   (Name:'Eg';  g:1; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:2; rho:0; eta:1),
   (Name:'B2u'; g:2; rho:1; eta:1),
   (Name:'Eu';  g:1; rho:0; eta:1));

  D6hCharacterTable: array [0..11] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:3; rho:0; eta:0),
   (Name:'B2g'; g:3; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:3; rho:0; eta:1),
   (Name:'B2u'; g:3; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

  D8hCharacterTable: array [0..13] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'B1g'; g:4; rho:0; eta:0),
   (Name:'B2g'; g:4; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'E3g'; g:3; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'B1u'; g:4; rho:0; eta:1),
   (Name:'B2u'; g:4; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1),
   (Name:'E3u'; g:3; rho:0; eta:1));

  D3dCharacterTable: array [0..5] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'Eg';  g:1; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'Eu';  g:1; rho:0; eta:1));

  D5dCharacterTable: array [0..7] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1));

  D7dCharacterTable: array [0..9] of TCharacterTableEntry =
  ((Name:'A1g'; g:0; rho:0; eta:0),
   (Name:'A2g'; g:0; rho:1; eta:0),
   (Name:'E1g'; g:1; rho:0; eta:0),
   (Name:'E2g'; g:2; rho:0; eta:0),
   (Name:'E3g'; g:3; rho:0; eta:0),
   (Name:'A1u'; g:0; rho:0; eta:1),
   (Name:'A2u'; g:0; rho:1; eta:1),
   (Name:'E1u'; g:1; rho:0; eta:1),
   (Name:'E2u'; g:2; rho:0; eta:1),
   (Name:'E3u'; g:3; rho:0; eta:1));

  D2dCharacterTable: array [0..4] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:2; rho:0; eta:0),
   (Name:'B2'; g:2; rho:1; eta:0),
   (Name:'E';  g:1; rho:0; eta:0));

  D4dCharacterTable: array [0..6] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:4; rho:0; eta:0),
   (Name:'B2'; g:4; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0));

  D6dCharacterTable: array [0..8] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:6; rho:0; eta:0),
   (Name:'B2'; g:6; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0));

  D8dCharacterTable: array [0..10] of TCharacterTableEntry =
  ((Name:'A1'; g:0; rho:0; eta:0),
   (Name:'A2'; g:0; rho:1; eta:0),
   (Name:'B1'; g:8; rho:0; eta:0),
   (Name:'B2'; g:8; rho:1; eta:0),
   (Name:'E1'; g:1; rho:0; eta:0),
   (Name:'E2'; g:2; rho:0; eta:0),
   (Name:'E3'; g:3; rho:0; eta:0),
   (Name:'E4'; g:4; rho:0; eta:0),
   (Name:'E5'; g:5; rho:0; eta:0),
   (Name:'E6'; g:6; rho:0; eta:0),
   (Name:'E7'; g:7; rho:0; eta:0));

var
  CharacterTables: array [TPointGroup] of TCharacterTable;

{ TSymmetricMolecule }

function TSymmetricMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
var
  Gdipole, G1, G2: TCharacterTableEntry;
  g: Integer;
begin
  if not TransitionMoment.Magnetic and  (Minus(TransitionMoment.Rank) < 0) then begin
    Result := False;
    { Does Sym1 * DipoleSym = Sym2? }
    { En Match }
    { TODO: Speed up? Is Gdipole.g always 0? }
    Gdipole := CharacterTable.IReps[CharacterTable.DipoleSym];
    G1 := CharacterTable.IReps[Sym1];
    G2 := CharacterTable.IReps[Sym2];
    g := Gdipole.g + G1.g;
    if g < 0 then
      g := g - g*CharacterTable.n; { Want result of mod on next line to be positive }
    g := g mod CharacterTable.n;
    if g > CharacterTable.n div 2 then
      Dec(g,CharacterTable.n);
    if Abs(g) <> G2.g then Exit;
    { g/u '/'' Match }
    if Odd(Gdipole.eta + G1.eta + G2.eta) then Exit;
    if Odd(Gdipole.rho + G1.rho + G2.rho) then Exit; { TODO: Check }
    Result := True;
  end else
    Result := Sym1 = Sym2;
end;

constructor TSymmetricMolecule.Create(AOwner: TComponent);
var
  i: Integer;
begin
  SavedProperties := 'PointGroup=C3v,' + SavedProperties;
  for i := 0 to MaxnStatWeight-1 do
    SavedProperties := SavedProperties + Format('wt%d=1,', [i]);
  inherited Create(AOwner);
end;

function TSymmetricMolecule.GetSettingNames(AIndex: Integer): string;
begin
  Result := inherited GetSettingNames(AIndex);
  if (AIndex >= WTI) and (AIndex < WTI+MaxnStatWeight) then
    if AIndex <= WTI+High(CharacterTable.GroupNames) then
      Result := CharacterTable.GroupNames[AIndex-WTI]
    else
      Result := '(' + Result + ')';
end;

function TSymmetricMolecule.GetPointGroup(const Index: Integer): TPointGroup;
begin
  Result := TPointGroup(GetSetting(Index));
end;

procedure TSymmetricMolecule.SetPointGroup(const Index: Integer; const AValue: TPointGroup);
begin
  FCharacterTable := CharacterTables[AValue];
  if PointGroup = AValue then Exit;
  SetSetting(Index, Ord(AValue));
  PerformMNodeNotification(Self, mnPropNameChange, Index);
end;

function TSymmetricMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym;
  if Mixture.JAdjustSym and (Minus(J) < 0) and (CharacterTable.IReps[Sym].Name[1] <> 'E') then
    if CharacterTable.IReps[Sym].rho = 1 then
      Dec(Result)
    else
      Inc(Result);
end;

function TSymmetricMolecule.SymName(Sym: Integer): string;
begin
  Result := CharacterTable.Names[Sym];
end;

function TSymmetricMolecule.LastSym: Integer;
begin
  Result := CharacterTable.Size-1;
end;

function TSymmetricMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = LMI then
    Result := 'Point group of molecule.'
  else if (AIndex >= WTI) and (AIndex < WTI+MaxnStatWeight) then
    if AIndex <= WTI+High(CharacterTable.GroupNames) then
      Result := 'Staistical weight for ' + CharacterTable.GroupNames[AIndex-WTI] + ' rotational levels.'
    else
      Result := 'Unused'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

class function TSymmetricMolecule.ManifoldType: TMNodeClass;
begin
  Result := TSymmetricManifold;
end;

{ TSymmetricManifold }

function TSymmetricManifold.FirstSym(J: TQuantumNumber): Integer;
begin
  Result := 0;
end;

function TSymmetricManifold.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case AIndex of
    0: Result := TSymmetricTop;
    { TODO: 1: Result := TSymmetricTopPerturbation; }
  else
    Result := nil;
  end;
end;

function TSymmetricManifold.NextSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Sym+1;
  if Result >= nSym then
    Result := -1;
end;

function TSymmetricManifold.nSym: Integer;
begin
  Result := TSymmetricMolecule(Owner).CharacterTable.Size;
end;

class function TSymmetricManifold.OwnerType: TComponentClass;
begin
  Result := TSymmetricMolecule;
end;

{ TSymmetricTop }

function SameStatisticalWeightGroup(PointGroup: TPointGroup; sym, sym1: Integer):Boolean;
begin
  Result := True;
  { For each symmetric top point group, one of the generating operations corresponds to
    a permutation/inversion, so the symmetry of the nuclear+rovibronic wavefunction is not selected
    by this operation. Symmetries that only differ in this operation are therefore in the same
    statistical weight group }
  if not (PointGroup in [D2d, D4d, D6d, D8d]) then { R+ }
    Result := Result and (CharacterTables[PointGroup].IReps[sym].g = CharacterTables[PointGroup].IReps[sym1].g);
  if not (PointGroup in [D3h, D4h, D5h, D6h, D7h, D8h, D3d, D5d, D7d]) then { R'+ }
    Result := Result and (CharacterTables[PointGroup].IReps[sym].eta = CharacterTables[PointGroup].IReps[sym1].eta);
  if not (PointGroup in [C3v, C4v, C5v, C6v, C7v, C8v]) then { R- }
    Result := Result and (CharacterTables[PointGroup].IReps[sym].rho = CharacterTables[PointGroup].IReps[sym1].rho);
end;

procedure DetermineStatisticalWeightGroups(PointGroup: TPointGroup);
var
  sym, sym1, nGroups: Integer;
begin
  with CharacterTables[PointGroup] do begin
    SetLength(Group, Size);
    SetLength(GroupNames, Size); { Initial guess - contract later }
    nGroups := 0;
    FillChar(Group[0], SizeOf(Group), 0);
    for sym := 0 to Size-1 do { For each irreducible representation }
      if Group[sym] = 0 then begin { if not assigned to statistical weight group }
        GroupNames[nGroups] := 'wt(' + Names[sym];
        Group[sym] := nGroups+1;
        for sym1 := sym+1 to Size-1 do
          if SameStatisticalWeightGroup(PointGroup, sym, sym1) then begin
            Group[sym1] := nGroups+1;
            GroupNames[nGroups] := GroupNames[nGroups] + ',' + Names[sym1];
          end;
        GroupNames[nGroups] := GroupNames[nGroups] + ')';
        Inc(nGroups);
      end;
    SetLength(GroupNames,nGroups);
    //Write(GroupName, ': ', Names[DipoleSym], ' : ');
    //for i := 0 to nGroups-1 do
    //  Write(' ', GroupNames[i]);
    //Writeln;
  end;
end;

function FindIrep(PointGroup: TPointGroup; g, rho, eta:Integer):Integer;
var
  i: Integer;
begin
  with CharacterTables[PointGroup] do
    for i := 0 to Size-1 do
      if (Ireps[i].g = g) and (Ireps[i].rho = rho) and (Ireps[i].eta = eta) then begin
        Result := i;
        Exit;
      end;
   raise Exception.Create('FindIrep');
end;

procedure DetermineDipoleSymmetry(PointGroup: TPointGroup);
begin
  with CharacterTables[PointGroup] do
    if PointGroup in [C3v, C4v, C5v, C6v, C7v, C8v] then
      DipoleSym := FindIrep(PointGroup, 0, 1, 0)
    else if PointGroup in [D2d, D4d, D6d, D8d] then
      DipoleSym := FindIrep(PointGroup, n div 2, 0, 0)
    else
      DipoleSym := FindIrep(PointGroup, 0, 0, 1);
end;

procedure AddTables;
var
  p: TPointGroup;
  {.$DEFINE CHECK}
  {$IFDEF CHECK}
  procedure SetStatWtSets(PointGroup: TPointGroup; DipoleName: string; GroupsDef: array of PChar);
  var
    i, p, sym: Integer;
    s, Irep: string;
    function IrepIndex(Name: string): Integer;
    var
      j: Integer;
    begin
      with CharacterTables[PointGroup] do
        for j := 0 to Size-1 do
          if Names[j] = Name then begin
            Result := j;
            Exit;
          end;
      raise Exception.Create('SymT:3');
    end;
  begin
    with CharacterTables[PointGroup] do begin
      {$IFDEF CHECK}
        if Length(Group) <> Size then raise NotYet;
        if Length(GroupNames) <> Length(GroupsDef) then raise NotYet;
      {$ELSE}
        SetLength(Group, Size);
        SetLength(GroupNames, Length(GroupsDef));
        FillChar(Group[0], SizeOf(Group[0])*Size, 0);
      {$ENDIF}
      for i := 0 to High(GroupsDef) do begin
        s := GroupsDef[i];
        {$IFDEF CHECK}
          if GroupNames[i] <> 'wt(' + s + ')' then raise NotYet;
        {$ELSE}
          GroupNames[i] := 'wt(' + s + ')';
        {$ENDIF}
        while s <> '' do begin
          Irep := s;
          p := Pos(',',s);
          if p > 0 then begin
            Delete(Irep, p, Length(Irep));
            Delete(s,1,p);
          end else
            s := '';
          sym := IrepIndex(Irep);
          {$IFDEF CHECK}
          if Group[sym] <> i+1 then raise NotYet;
          {$ELSE}
          if Group[sym] <> 0 then
            raise Exception.Create('SymT:1')
          else
            Group[sym] := i+1;
          {$ENDIF}
        end;
      end;
      { Check for unset groups }
      for sym := 0 to Size-1 do
        if Group[sym] = 0 then
          raise Exception.Create('SymT:2');
      {$IFDEF CHECK}
      if DipoleSym <> IrepIndex(DipoleName) then raise NotYet;
      {$ELSE}
      DipoleSym := IrepIndex(DipoleName);
      {$ENDIF}
    end;
  end;
  {$ENDIF}

  procedure A(T: array of TCharacterTableEntry; Order:Integer; rhomult: Integer; gMult:Integer = -1; etamult: Integer = 0);
  var
    i: Integer;
  begin
    with CharacterTables[p] do begin
      GroupName := CharacterTableNames[p];
      Size := Length(T);
      n := Order;
      gKmult := gMult;
      rhoKmult := rhomult;
      etaKmult := etamult;
      Setlength(IReps, Size);
      Setlength(Names, Size+1);
      SetLength(Degenerate,Size);
      Names[Size] := nil;
      for i := 0 to Size-1 do begin
        Names[i] := PChar(T[i].Name);
        IReps[i] := T[i];
        Degenerate[i] := Names[i][1] = 'E';
      end;
      DetermineDipoleSymmetry(p);
      DetermineStatisticalWeightGroups(p);
      if p < High(TPointGroup) then
        Inc(p);
    end;
  end;
begin
  p := Low(TPointGroup);
  { Must be in ordinal order }
  A(C3vCharacterTable,3,1); A(C4vCharacterTable,4,1); A(C5vCharacterTable,5,1); A(C6vCharacterTable,6,1); A(C7vCharacterTable,7,1); A(C8vCharacterTable,8,1);
  A(D3hCharacterTable,3,0,-1,1); A(D4hCharacterTable,4,0); A(D5hCharacterTable,5,0,-1,1); A(D6hCharacterTable,6,0); A(D7hCharacterTable,7,0,-1,1); A(D8hCharacterTable,8,0);
  A(D2dCharacterTable, 4,0,-3);  A(D3dCharacterTable,3,0); A(D4dCharacterTable, 8,0,-5);  A(D5dCharacterTable,5,0);
  A(D6dCharacterTable,12,0,-7);  A(D7dCharacterTable,7,0); A(D8dCharacterTable,16,0,-9);

  {$IFDEF CHECK}
  SetStatWtSets(C3v, 'A2',  ['A1,A2', 'E']);
  SetStatWtSets(D3h, 'A1"', ['A1'',A1"', 'A2'',A2"', 'E'',E"']);
  SetStatWtSets(D6h, 'A1u', ['A1g,A1u', 'A2g,A2u', 'B1g,B1u', 'B2g,B2u', 'E1g,E1u', 'E2g,E2u']);
  {$ENDIF}
end;

class function TSymmetricTop.OwnerType: TComponentClass;
begin
  Result := TSymmetricManifold;
end;

function TSymmetricTop.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; HyperfineState: THyperfineBasis): Integer;
var
  EVsym, EVRSym: TCharacterTableEntry;
  l, K, Kinc: TQuantumNumber;
  CharacterTable: TCharacterTable;
  procedure Consider(K, l: Integer);
  var
    BasisState: TSymmetricTopBasisState;
    g, rho, eta, Phase: Integer;
  begin
    { Does EVsym * Sym(K) = EVRSym? }
    { En Match }
    g := CharacterTable.gKmult*K div 2 + EVsym.g;
    if g < 0 then
      g := g - g*CharacterTable.n; { Want result of mod on next line to be positive }
    g := g mod CharacterTable.n;
    if g > CharacterTable.n div 2 then
      Dec(g,CharacterTable.n);
    if Abs(g) <> EVRsym.g then Exit;
    { g/u '/'' Match }
    eta := K*CharacterTable.etaKmult div 2 + EVsym.eta;
    if Odd(eta+EVRSym.eta) then Exit;
    if (g = 0) or (2*g = CharacterTable.n) then begin
      { A1 or A2 / B1 or B2 }
      rho := (J+K*CharacterTable.rhoKmult) div 2 + EVsym.rho;
      if Odd(rho+EVRSym.rho) then begin
        Phase := -1;
        if (l=0) and (K=0) then
          Exit; { |J00>-|J00> }
      end else begin
        Phase := 1;
        if (l=0) and (K=0) then
          Phase := 0; { |J00>+|J00> }
      end
    end else begin
      { E - two levels, but only work out the energy of one of them }
      Phase := 0;
      { Choose the one with G > 0, preserving sign of kl }
      { There should be no matrix elements off diagonal in G }
      { This also fixes transition matrix elements, as these are also diagonal in G }
      if g < 0 then begin
        K := -K;
        l := -l;
      end;
    end;
    BasisState := TSymmetricTopBasisState.Create(Eigenvalues);
    BasisState.State := Self;
    BasisState.J := J;
    BasisState.K := K;
    BasisState.l := l;
    BasisState.Phase := Phase;
    BasisState.StateIndex := Eigenvalues.LastStateIndex;
    BasisState.HyperfineState := HyperfineState;
  end;
begin
  CharacterTable := Molecule.CharacterTable;
  Result := Molecule.GetSetting(WTI+CharacterTable.Group[Eigenvalues.Sym]-1);
  if Result = 0 then
    Exit;
  EVsym := CharacterTable.IReps[Symmetry];
  if EVsym.Name[1] = 'E' then
    l := 2
  else
    l := 0;
  EVRsym := CharacterTable.IReps[Eigenvalues.Sym];
  if Prolate then begin
    K := 0;
    Kinc := 2
  end else begin
    Kinc := -2;
    K := J;
  end;
  while (K <= J) and (K >= 0) do begin
    Consider(K,l);
    if (l<>0) and (K>0) then
      Consider(-K,l);
    Inc(K,Kinc);
  end;
  Inc(Eigenvalues.LastStateIndex);
end;

function TSymmetricTop.GetSettingsNames(AIndex: Integer; var LowestValue: Integer): PPChar;
begin
  if AIndex = LSI then
    if Molecule = nil then { Should never happen }
      Result := @CharacterTables[D3h].Names[0]
    else
      Result := @CharacterTables[Molecule.PointGroup].Names[0]
  else
    Result := inherited GetSettingsNames(AIndex,LowestValue);
end;

constructor TSymmetricTop.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TSymmetricMolecule;
  inherited;
end;

function TSymmetricTop.FixupProperties: Exception;
begin
  FProlate := C > B; { Precomputation not a speed up, but work(ed) round a FPC PPC bug }
  Result:=inherited FixupProperties;
end;

{ TSymmetricTopBasisState }

function TSymmetricTopBasisState.BasisLabel: string;
begin
  Result := Format('%s %2s', [inherited BasisLabel, State.Manifold.Mixture.QuantumNumberToStr(K)]);
  if l <> 0 then
    Result := Result + Format(' %s', [State.Manifold.Mixture.QuantumNumberToStr(l)])
end;

function TSymmetricTopBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey + '  K l';
end;

function TSymmetricTopBasisState.klLabel: string;
var
  kl: Integer;
  CharacterTable: TCharacterTable;
begin
  if l <> 0 then begin
    if K = 0 then
      Result := ' '
    else begin
      kl := K*l;
      CharacterTable := TSymmetricTop(State).Molecule.CharacterTable;
      if CharacterTable.IReps[TSymmetricTop(State).Symmetry].g
         mod 3 = 2 then
        { this is sgn(k*sin(2 pi Gve /3)) as per Hoy and Mills 1973 }
        kl := -kl;
      if kl > 0 then
        Result := '+'
      else
        Result := '-'
    end;
  end else
    Result := '';
end;

function TSymmetricTopBasisState.StateLabel: string;
var
  CharacterTable: TCharacterTable;
begin
  Result := Format('%s %2s %2s', [inherited StateLabel,
                                  State.Manifold.Mixture.QuantumNumberToStr(J),
                                  State.Manifold.Mixture.QuantumNumberToStr(Abs(K))]);
  CharacterTable := TSymmetricTop(State).Molecule.CharacterTable;
  if l<>0 then
    Result := Result + ' ' + klLabel;
  Result := Result + Format(' %*s', [Length(CharacterTable.Ireps[Sym].Name), TSymmetricTop(State).Molecule.SymName(Sym)]);
end;

function TSymmetricTopBasisState.StateLabelKey: string;
begin
  Result := Format('%s J K (kl) Sym', [inherited StateLabelKey]);
end;

function TSymmetricTopBasisState.TrackIndex: Integer;
begin
  Result := Abs(K); { TODO: May want to include l also }
end;

function TSymmetricTopBasisState.TxName(Upper, Lower: TBasisState): string;
var
  Molecule: TSymmetricMolecule;
  U, LL: TSymmetricTopBasisState;
  US, LS: TSymmetricTop;
begin
  Molecule := State.Manifold.Molecule as TSymmetricMolecule;
  U := Upper as TSymmetricTopBasisState;
  US := U.State as TSymmetricTop;
  LL := Lower as TSymmetricTopBasisState;
  LS := LL.State as TSymmetricTop;
  { TODO: Abs in next line may need consideration }
  Result := LowerCase(DeltaJName(Abs(U.K)-Abs(LL.K)))+DeltaJName(U.J-LL.J);
  if LL.l <> 0 then
    Result := Result + Format('%3s', [klLabel+Molecule.QuantumNumberToStr(Abs(LL.K))])
  else
    Result := Result + Format('%2s', [Molecule.QuantumNumberToStr(Abs(LL.K))]);
  Result := Result + Format('(%2s)',[Molecule.QuantumNumberToStr(LL.J)]);
  { TODO: Hyperfine, Open Shell }
end;

function TSymmetricTopBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := 'delta|K| deltaJ K"(J")';
end;

function TSymmetricTop.OddSpin: TQuantumNumber;
begin
  Result := 0; { Closed shell for now }
end;

function TSymmetricTop.HRot(J, Kp, lp, K, l: TQuantumNumber): Double;
var
  JSq, KSq: Double;
begin
  if (K = Kp) and (l = lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    Result := Origin + B*JSq + (C-B)*KSq
              - DJ*Sqr(JSq) - DJK*JSq*KSq - DK*Sqr(KSq)
              + HJ*Sqr(JSq)*JSq + HJK*Sqr(JSq)*KSq + HKJ*Jsq*Sqr(KSq) + HK*Sqr(KSq)*KSq;
    Result := Result + (-2*C*zeta + etaj*JSq + etak*KSq) * l*K/4;
  end else if (Abs(K-Kp) = 4) and (K-Kp = l-lp) then begin
    JSq := J*(J+2) / 4;
    KSq := K*K / 4;
    { TODO: Similar to Cartwright and Mills; JMS 34 415 (1970); qplus = -q as used in NH3 Rydberg work }
    { Aliev and Papousek suggest qt+ has extra factors involving v and l }
    Result := (qplus + DqJ*JSq + DqK*KSq)/2; { TODO: DqK does not look right }
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if (Abs(K-Kp) = 4) and (K-Kp = -(l-lp)) then begin
    JSq := J*(J+2) / 4;
    Result := qminus/2;
    Result := Result * Sqrt((JSq - K*(K+Kp)/8)*(JSq - Kp*(K+Kp)/8));
  end else if (Abs(K-Kp) = 2) and (2*(K-Kp) = -(l-lp)) then begin
    JSq := J*(J+2) / 4;
    Result := r * Sqrt((JSq - K*Kp/4)) * (K+Kp) / 2;
  end else
    Result := 0;
end;

function TSymmetricTop.RotationalMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Double;
var
  Bra, Ket: TSymmetricTopBasisState;
  H: function (J, Kp, lp, K, l: TQuantumNumber): Double of object;
  Kp, lp, K, l: TQuantumNumber;
begin
  Bra := A as TSymmetricTopBasisState;
  Ket := B as TSymmetricTopBasisState;
  Result := 0;
  case TMatElemType(Flag) of
    meHrot:  H := {$IFDEF FPC}@{$ENDIF}Hrot
  else
    raise NotYet;
  end;
  if Bra.J <> Ket.J then
    raise Exception.Create('TSymmetricTop.RotationalMatrixElement');
  Kp := SignA*Bra.K;
  lp := SignA*Bra.l;
  K := SignB*Ket.K;
  l := SignB*Ket.l;
  { TODO: may require all four matrix elements or explicit test for identical phase? }
  Result := H(Bra.J, Kp, lp, K, l);
  { TODO: S <> 0 terms }
end;

{
function TSymmetricTop.MatrixElement(ABra, AKet: TBasisState; MatElemType: TMatElemType): Double;
var
  Bra, Ket: TSymmetricTopBasisState;
  H: function (J, Kp, lp, K, l: TQuantumNumber): Double of object;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  Result := 0;
  if (Bra.State <> Ket.State) or (Bra.Sym <> Ket.Sym) or (Bra.J <> Ket.J) then
    Exit;
  case MatElemType of
    meHrot:  H := Hrot;
    meWidth: raise NotYet;
  else
    raise NotYet;
  end;
  Result := H(Bra.J, Bra.K, Bra.l, Ket.K, Ket.l);
  if Bra.Phase = 0 then begin
    if Ket.Phase <> 0 then
      Result := (Result + Ket.Phase*H(Bra.J, Bra.K, Bra.l, -Ket.K, -Ket.l))/Sqrt(2.0);
  end else if Ket.Phase = 0 then
    Result := (Result + Bra.Phase*H(Bra.J, -Bra.K, -Bra.l, Ket.K, Ket.l))/Sqrt(2.0)
  else
    // All 4 here to permit general symmetry to work - could be speeded up with specific test
    Result := (Result + Ket.Phase*H(Bra.J, Bra.K, Bra.l, -Ket.K, -Ket.l)
                      + Bra.Phase*H(Bra.J, -Bra.K, -Bra.l, Ket.K, Ket.l)
                      + Bra.Phase*Ket.Phase*H(Bra.J, -Bra.K, -Bra.l, -Ket.K, -Ket.l)
               )/2;
end;
}

function TSymmetricTop.HDip(TransitionMoment: TTransitionMoment; Jp, Kp, lp, J, K, l: TQuantumNumber): Double;
var
  q: TQuantumNumber;
  Phase: Integer;
begin
  Result := 0;
  Phase := 0;
  if TransitionMoment is TSphericalTransitionMoment then begin
    q := TSphericalTransitionMoment(TransitionMoment).Component;
    { If the other sign of q is required, apply symmtery operation }
    if q <> K-Kp then
      phase := TransitionMoment.Rank + q*Molecule.CharacterTable.rhoKmult;
  end else
    raise NotYet;
  if Abs(q) = Abs(Kp-K) then
    Result := Sqrt((J+1.0)*(Jp+1)/(TransitionMoment.Rank+1))
              * Minus(J-K+q+Phase) * threej(Jp, Kp, TransitionMoment.Rank, K-Kp, J, -K) * TransitionMoment.Strength;
end;

function TSymmetricTop.RotationalTransitionMatrixElement(A, B: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Double;
var
  Bra, Ket: TSymmetricTopBasisState;
  TransitionMoment: TTransitionMoment;
  Kp, lp, K, l: TQuantumNumber;
begin
  Bra := A as TSymmetricTopBasisState;
  Ket := B as TSymmetricTopBasisState;
  TransitionMoment := MNode as TTransitionMoment;
  Kp := SignA*Bra.K;
  lp := SignA*Bra.l;
  K := SignB*Ket.K;
  l := SignB*Ket.l;
  Result := HDip(TransitionMoment, Bra.J, Kp, lp, Ket.J, K, l);
end;

{
function TSymmetricTop.TransitionMatrixElement(TransitionMoment: TTransitionMoment; ABra, AKet: TBasisState): Double;
var
  Bra, Ket: TSymmetricTopBasisState;
begin
  Bra := ABra as TSymmetricTopBasisState;
  Ket := AKet as TSymmetricTopBasisState;
  Result := HDip(TransitionMoment, Bra.J, Bra.K, Bra.l, Ket.J, Ket.K, Ket.l);
  if Bra.Phase = 0 then begin
    if Ket.Phase <> 0 then
      Result := (Result + Ket.Phase*HDip(TransitionMoment, Bra.J, Bra.K, Bra.l, Ket.J, -Ket.K, -Ket.l))/Sqrt(2.0);
  end else if Ket.Phase = 0 then
    Result := (Result + Bra.Phase*HDip(TransitionMoment, Bra.J, -Bra.K, -Bra.l, Ket.J, Ket.K, Ket.l))/Sqrt(2.0)
  else begin
    Result :=  Result + Ket.Phase*HDip(TransitionMoment, Bra.J, Bra.K, Bra.l, Ket.J, -Ket.K, -Ket.l);
    Result :=  Result + Bra.Phase*HDip(TransitionMoment, Bra.J, -Bra.K, -Bra.l, Ket.J, Ket.K, Ket.l);
    Result :=  Result + Bra.Phase*Ket.Phase*HDip(TransitionMoment, Bra.J, -Bra.K, -Bra.l, Ket.J, -Ket.K, -Ket.l);
    Result := Result / 2;
  end;
end;
}

const
  SymmetricTopSettingNames: array[LSI..LSI] of string = (
'Symmetry'
);

function TSymmetricTop.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopSettingNames))
     and (AIndex <= High(SymmetricTopSettingNames)) then
    Result := SymmetricTopSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 SymmetricTopParameterNames: array [LSP..LSP+16] of string = (
'Rotational perpendicular to symmetry axis.',
'Rotational constant about symmetry axis.',
'J^2(J+1)^2 Quartic Centrifugal Distortion',
'J(J+1)K^2 Quartic Centrifugal Distortion',
'K^4 Quartic Centrifugal Distortion',
'Coriolis coupling constant.',
'J(J+1) dependence of Coriolis coupling constant.',
'K^2 dependence of Coriolis coupling constant.',
'l doubling constant.',
'l doubling constant.',
'l doubling constant.',
'Centrifugal distortion of l doubling constant.',
'Centrifugal distortion of l doubling constant.',
'J^3(J+1)^3 Sextic Centrifugal Distortion',
'J^2(J+1)^2K^2 Sextic Centrifugal Distortion',
'J(J+1)K^4 Sextic Centrifugal Distortion',
'K^6 Sextic Centrifugal Distortion'
);

function TSymmetricTop.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SymmetricTopParameterNames))
     and (AIndex <= High(SymmetricTopParameterNames)) then
    Result := SymmetricTopParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

initialization
  AddTables;
  RegisterClasses([TSymmetricMolecule, TSymmetricManifold, TSymmetricTop {, TSymmetricTopPerturbation}]);
  TSpecies.AddMoleculeType(TSymmetricMolecule);
end.
