unit MatrixVec; { Unit Matrix is part of FPC run time library }
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes, Vector;

type
  TMatrix = class ({IFDEF MAKEFROMCOLVECTS TGVectorN $ELSE} TColumnVector) { Dimensions 1..RowCount, 1..ColCount }
  private
    FRowCount: Longint;
    FColCount: Longint;
    function GetValid(ARow, ACol:Longint):Boolean;
    procedure SetValid(ARow, ACol:Longint; IsValid:Boolean);
    function GetValue(ARow, ACol:Longint):Double;
    procedure SetValue(ARow, ACol:Longint; v:Double); {$IFDEF INLINE} inline; {$ENDIF}
    function GetPointer(ARow, ACol:Longint):Pointer; {$IFDEF INLINE} inline; {$ENDIF}
    procedure SetRowCount(n: Longint);
    procedure SetColCount(n: Longint);
    procedure SetCell(ARow, ACol:Longint; v:Double; IsValid:Boolean);
    {$IFDEF MAKEFROMCOLVECTS}
    function GetColumn(ACol:Longint):TColumnVector;
    procedure CheckSizes;
    {$ENDIF}
  public
    constructor Create(Rows, Cols: Longint);
    {$IFDEF MAKEFROMCOLVECTS}
    destructor Destroy; override;
    procedure Sqr;
    procedure Add(AMatrix: TMatrix);
    {$ENDIF}
    procedure SwapColumns(Col1, Col2: Longint);
    procedure Assign(Source: TPersistent); override;
    {$IFNDEF MAKEFROMCOLVECTS}
    function Index(ARow, ACol: Longint): Integer; {$IFDEF INLINE} inline; {$ENDIF}
    {$ENDIF}
    function GetCell(ARow, ACol:Longint; out v:Double):Boolean;
    property Item[ARow, ACol: Longint]:Pointer read GetPointer;
    property RowCount: Longint read FRowCount write SetRowCount;
    property ColCount: Longint read FColCount write SetColCount;
    property Valid[ARow, ACol:Longint]:Boolean read GetValid write SetValid;
    property Value[ARow, ACol:Longint]:Double read GetValue write SetValue; default;
    {$IFDEF MAKEFROMCOLVECTS}
    property Column[ACol:Longint]:TColumnVector read GetColumn;
    {$ENDIF}
    procedure PostMultiply(M: TMatrix);
  end;

implementation

uses {$IFNDEF DESIGNPACKAGE} Blas, {$ENDIF} MathSup, SysUtils;

constructor TMatrix.Create(Rows, Cols: Longint);
begin
  FRowCount := Rows;
  FColCount := Cols;
  {$IFDEF MAKEFROMCOLVECTS}
  inherited Create(SizeOf(Pointer));
  Offset := 1;
  CheckSizes;
  {$ELSE}
  inherited Create(Rows*Cols);
  {$ENDIF}
end;

{$IFDEF MAKEFROMCOLVECTS}
destructor TMatrix.Destroy;
begin
  FColCount := 0;
  CheckSizes;
  inherited Destroy;
end;
{$ENDIF}

function TMatrix.Index(ARow, ACol: Longint): Integer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  {$IFOPT R+}
  if (ARow <= 0) or (ACol <= 0) or (ARow > RowCount) then
    raise Exception.CreateFmt('TMatrix[%d,%d]: Bound check failure (%d)', [ARow, ACol, RowCount]);
  {$ENDIF}
  Result := (ACol-1)*RowCount+ARow;
end;

function TMatrix.GetCell(ARow, ACol:Longint; out v:Double):Boolean;
var
  p: ^Double;
begin
  p := Item[ARow, ACol];
  Result := not IsNanOrInf(p^);
  System.Move(p^, v, SizeOf(Double));  { So that specific Invalid values can be seen }
end;

function TMatrix.GetPointer(ARow, ACol:Longint):Pointer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  {$IFDEF MAKEFROMCOLVECTS}
  Result := TColumnVector(inherited Item[ACol]^).Item[ARow];
  {$ELSE}
  Result := inherited Item[Index(ARow, ACol)];
  {$ENDIF}
end;

procedure TMatrix.SetCell(ARow, ACol:Longint; v:Double; IsValid:Boolean);
var
  p: ^Double;
begin
  p := GetPointer(ARow, ACol);
  { Note: To Set a specific invalid value, set IsValid := True
    and v to the required value }
  if IsValid then
    p^ := v
    // May need to use: System.Move(v, p^, SizeOf(Double))
    // if compiler generates floating point load/store
  else
    SetNan(p^);
end;

{$IFDEF MAKEFROMCOLVECTS}
function TMatrix.GetColumn(ACol:Longint):TColumnVector;
begin
  Result := TColumnVector(inherited Item[ACol]^);
end;
{$ENDIF}

function TMatrix.GetValid(ARow, ACol:Longint):Boolean;
var
  v: Double;
begin
  Result := GetCell(ARow, ACol, v);
end;

procedure TMatrix.SetValid(ARow, ACol:Longint; IsValid:Boolean);
begin
  SetCell(ARow, ACol, 0, IsValid);
end;

function TMatrix.GetValue(ARow, ACol:Longint):Double;
begin
  {$IFDEF MAKEFROMCOLVECTS}
  Result := PDouble(PChar(TColumnVector(inherited Item[ACol]^).Base)+(ARow-1)*8)^;
  {$ELSE}
  {$IFOPT R+}
  if (ARow <= 0) or (ACol <= 0) or (ARow > RowCount) then
    raise Exception.CreateFmt('TMatrix[%d,%d]: Bound check failure', [ARow, ACol]);
  {$ENDIF}
  Result := PDouble(inherited Item[(ACol-1)*RowCount+ARow])^;
  {$ENDIF}
end;

procedure TMatrix.SetValue(ARow, ACol:Longint; v:Double); {$IFDEF FPCINLINE} inline; {$ENDIF}
begin;
  SetCell(ARow, ACol, v, True);
end;

procedure TMatrix.SetRowCount(n: Longint);
var
  c, OldRows: Longint;
begin
  if n <> RowCount then begin
    OldRows := RowCount;
    FRowCount := n;
    {$IFDEF MAKEFROMCOLVECTS}
      CheckSizes;
    {$ELSE}
      if (RowCount < OldRows) and (RowCount > 0) then
        for c := 2 to ColCount do
          System.Move(inherited Item[(c-1)*OldRows+1]^, Item[1,c]^, OldRows*Sizeof(Double));
      Size := ColCount * RowCount;
      if RowCount > OldRows then begin
        { Matrix is Expanding - Shift Each row into position }
        for c := ColCount downto 2 do
          System.Move(inherited Item[(c-1)*OldRows+1]^, Item[1,c]^, OldRows*Sizeof(Double));
        { Clear resulting gaps }
        for c := 1 to ColCount do
          ClearRange(Index(OldRows+1, c), RowCount-OldRows);
      end;
    {$ENDIF}
  end;
end;

procedure TMatrix.SetColCount(n: Longint);
begin
  if n <> ColCount then begin
    FColCount := n;
    {$IFDEF MAKEFROMCOLVECTS}
      CheckSizes;
    {$ELSE}
      Size := ColCount * RowCount;
    {$ENDIF}
  end;
end;

{$IFDEF MAKEFROMCOLVECTS}
procedure TMatrix.CheckSizes;
var
  n, j: Longint;
begin
  if FColCount > Size then begin
    n := Size;
    Size := FColCount;
    for j := n+1 to Size do begin
      TColumnVector(inherited Item[j]^) := TColumnVector.Create(FRowCount);
    end;
  end else if FColCount < Size then begin
    for j := Size downto FColCount+1 do begin
      TColumnVector(inherited Item[j]^).Free;
    end;
    Size := FColCount;
  end;
  if (FColCount > 0) and (TColumnVector(inherited Item[1]^).Size <> FRowCount) then
    for j := 1 to Size do
      TColumnVector(inherited Item[j]^).Size := RowCount;
end;
{$ENDIF}

procedure TMatrix.Assign(Source: TPersistent);
var
  s: TMatrix;
  {$IFDEF MAKEFROMCOLVECTS}
  c: Longint;
  {$ENDIF}
begin
  if Source is TMatrix then begin
    s := TMatrix(Source);
    FRowCount := s.RowCount;
    FColCount := s.ColCount;
    {$IFDEF MAKEFROMCOLVECTS}
    CheckSizes;
    for c := 1 to ColCount do
      Column[c].Assign(s.Column[c]);
    {$ELSE}
    Size := RowCount*ColCount;
    System.Move(s.FirstItem^, FirstItem^, RowCount*ColCount*Sizeof(Double));
    {$ENDIF}
  end else
    inherited Assign(Source);
end;

procedure TMatrix.SwapColumns(Col1, Col2: Integer);
var
  {$IFDEF MAKEFROMCOLVECTS}
    t: TColumnVector;
  {$ELSE}
    i: Longint;
    v: Double;
  {$ENDIF}
begin
  if Col1 <> Col2 then begin
  {$IFDEF MAKEFROMCOLVECTS}
    t := Column[Col1];
    TColumnVector(inherited Item[Col1]^) := Column[Col2];
    TColumnVector(inherited Item[Col2]^) := t;
  {$ELSE}
    {TODO: Speed up}
    for i := 1 to RowCount do begin
      System.Move(Item[i, Col1]^, v, sizeof(Double));
      System.Move(Item[i, Col2]^, Item[i, Col1]^, sizeof(Double));
      System.Move(v, Item[i, Col2]^, sizeof(Double));
    end;
  {$ENDIF}
  end;
end;

{$IFNDEF DESIGNPACKAGE}
{$DEFINE BLAS}
{$ENDIF}
{$IFDEF OFFBLAS}
function _dgemm_(TRANSA, TRANSB: PChar;
                var M, N, K: Integer;
                var ALPHA, A: Double;
                var LDA: Integer;
                var B: Double;
                var LDB: Integer;
                var BETA, C: Double;
                var LDC: Integer;
                TRANSAlen, TRANSBlen: Integer):Integer; cdecl; external 'lapack';
{$ENDIF}
{ Note that ATLAS seems to be slower than blas in this application }

procedure TMatrix.PostMultiply(M: TMatrix);
{ Overwrite matrix with matrix * M }
var
{$IFDEF BLAS}
  mm, N: Integer;
  alpha, beta: Double;
  A: PDouble;
  No: Char;
{$ELSE}
  W: array of Double;
  WM: TMatrix;
  Sum: Double;
  i, ii, iii: Integer;
  PW, PWM, PM, PE: PDouble;
{$ENDIF}
begin
  if ColCount <> M.RowCount then
    raise Exception.Create('TMatrix.PostMultiply:Size mismatch');
  if ColCount <> M.ColCount then
    raise Exception.Create('TMatrix.PostMultiply:Size change');
  {$IFDEF BLAS}
  mm := RowCount;
  N  := ColCount;
  Size := Size * 2; { Allocate workspace }
  A := PDouble(PChar(Base) + (Size div 2)*SizeOf(Double));
  System.Move(Base^, A^, (Size div 2)*SizeOf(Double));
  alpha := 1;
  beta := 0;
  No := 'N';
  dgemm_(No, No, mm, N, N, alpha, A^, mm, PDouble(M.Base)^, N, beta, PDouble(Base)^, mm, 1, 1);
  Size := Size div 2;
  {$ELSE}
  {$DEFINE ALT2}
  {$IFDEF ALT2}
  { Closest to blas code, but still slower, at least on g77/FPC }
  WM := TMatrix.Create(RowCount, ColCount);
  WM.Assign(Self);
  Zero;
  PM := M.FirstItem;
  for ii := 1 to ColCount do begin
    for iii := 1 to ColCount do begin
      {$IFDEF SIMPLE}
      Sum := M[iii,ii];
      {$ELSE}
      Sum := PM^;
      Inc(PM);
      {$ENDIF}
      if Sum <> 0 then begin
        {$IFDEF SIMPLE}
        for i := 1 to RowCount do
          Value[i,ii] := Value[i,ii] + Sum * WM[i,iii];
        {$ELSE}
        PW := Item[1,ii];
        PWM := WM.Item[1,iii];
        PE := PDouble(PChar(PWM) + RowCount*SizeOf(Double));
        while Cardinal(PWM) < Cardinal(PE) do begin
          PW^ := PW^ + Sum*PWM^;
          Inc(PW); Inc(PWM);
        end;
        {$ENDIF}
      end;
    end;
  end;
  WM.Free;
  {$ELSE}
  {$IFDEF ALT}
  SetLength(W, ColCount);
  for i := 1 to RowCount do begin
    for ii := 1 to ColCount do
      W[ii-1] := Value[i,ii];
    for ii := 1 to ColCount do begin
      Sum := 0;
      PW := @W[0];
      PM := M.Item[1,ii];
      for iii := 1 to ColCount do begin
        {$IFDEF SIMPLE}
        Sum := Sum + W[iii-1] * M[iii,ii];
        {$ELSE}
        Sum := Sum + PW^*PM^;
        Inc(PW); Inc(PM);
        {$ENDIF}
      end;
      Value[i,ii] := Sum;
    end;
  end;
  {$ELSE}
  SetLength(W, ColCount+1);
  for i := 1 to RowCount do begin
    for ii := 1 to ColCount do begin
      Sum := 0;
      for iii := 1 to ColCount do
        Sum := Sum + Value[i, iii] * M[iii,ii];
      W[ii] := Sum;
    end;
    for ii := 1 to ColCount do
      Value[i,ii] := W[ii];
  end;
  {$ENDIF}
  {$ENDIF}
  {$ENDIF}
end;

end.
