unit StackDump;

interface

uses Classes, SysUtils;

function StackWatch(Enabled: Boolean):Boolean;

var
  VersionInfo: string;
  ExpectedException: ExceptClass = nil;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Dialogs, Forms, StackTrace;

type
  TStackWatcher = class
  private
    PreviousExceptObjProc: Pointer;
    PreviousRaiseExceptionProc: Pointer;
  public
    S: TCallStack;
    LogFile: Text;
    LogFileName: string;
    FileNameSet, FileIsOpen: Boolean;
    StackFilled: Boolean;
    SavedLastError: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure OnException(Sender: TObject; E: Exception);
  end;

var
  StackWatcher:TStackWatcher;

type
  TExceptObjProc = function (P: PExceptionRecord): Exception;

function InterceptGetExceptionObject(P: PExceptionRecord): Exception;
begin
  Result := TExceptObjProc(StackWatcher.PreviousExceptObjProc)(P);
  if not StackWatcher.StackFilled then begin
    StackWatcher.SavedLastError := GetLastError;
    FillCallStack(StackWatcher.S, 0);
    StackWatcher.StackFilled := True;
  end;
end;

type
  TRaiseException = procedure (dwExceptionCode, dwExceptionFlags, nNumberOfArguments: DWORD; lpArguments: PDWORD); stdcall;

procedure InterceptRaiseException(dwExceptionCode, dwExceptionFlags, nNumberOfArguments: DWORD; lpArguments: PDWORD); stdcall;
begin
  if (dwExceptionCode <> $0EEDFAE2 { cDelphiTerminate } ) and not StackWatcher.StackFilled then begin
    StackWatcher.SavedLastError := GetLastError;
    FillCallStack(StackWatcher.S, 0);
    StackWatcher.StackFilled := True;
  end;
  TRaiseException(StackWatcher.PreviousRaiseExceptionProc)(dwExceptionCode, dwExceptionFlags, nNumberOfArguments, lpArguments);
end;

{ TStackWatcher }

constructor TStackWatcher.Create;
var
  p: Integer;
begin
  PreviousExceptObjProc := ExceptObjProc;
  ExceptObjProc := @InterceptGetExceptionObject;
  PreviousRaiseExceptionProc := RaiseExceptionProc;
  RaiseExceptionProc := @InterceptRaiseException;
  Application.OnException := OnException;
  LogFileName := Application.ExeName;
  FileNameSet := not FileIsReadOnly(ExtractFilePath(LogFileName)); { Can't write to directory }
  if FileNameSet then begin
    p := LastDelimiter('.'+ PathDelim + DriveDelim, LogFileName);
    if (p <= 0) or (LogFileName[p] <> '.') then
      p := Length(LogFileName);
    LogFileName := Copy(LogFileName, 1, p-1) + '.log';
    FileNameSet := not (FileExists(LogFileName) and FileIsReadOnly(LogFileName));
  end;
  if not FileNameSet then
    LogFileName := ExtractFileName(Application.ExeName) + '.log';
  // ShowMessage(Format('%s %d %s', [Application.ExeName, Ord(FileNameSet), LogFileName]));
  FileIsOpen := False;
  StackFilled := False;
end;

destructor TStackWatcher.Destroy;
begin
  Application.OnException := nil;
  ExceptObjProc := PreviousExceptObjProc;
  RaiseExceptionProc := PreviousRaiseExceptionProc;
  if FileIsOpen then
    CloseFile(LogFile);
  inherited;
end;

procedure TStackWatcher.OnException(Sender: TObject; E: Exception);
var
  Msg: string;
  LastError: Integer;
  function IO_OK:Boolean;
  begin
    { Note: Need to clear IOResult always }
    Result := (IOResult = 0) and FileIsOpen;
    if not Result then begin
      FileNameSet := False;
      FileIsOpen := False;
    end;
  end;
begin
  if DebugHook <> 0 then
    Exit; { Don't show dialog box, as already caught by debugger }
  Msg := E.Message;
  {$I-}
  if E is ExpectedException then
  else
  if (E is EFCreateError) or (E is EFOpenError) then begin
    LastError := SavedLastError;
    if LastError <> 0 then
      Msg := Msg + ': ' + SysErrorMessage(LastError);
    { No stack dump for simple file system errors }
  end else if StackFilled and (LogFileName <> '') then begin
    if not FileNameSet and not FileIsReadOnly(GetCurrentDir) then begin
      LogFileName := GetCurrentDir + PathDelim + LogFileName;
      FileNameSet := True;
    end;
    if FileNameSet then begin
      if not FileIsOpen then begin
        AssignFile(LogFile, LogFileName);
        if FileExists(LogFileName) then
          Append(LogFile)
        else
          Rewrite(LogFile);
        FileIsOpen := True;
      end;
      Writeln(LogFile, '');
      Writeln(LogFile, '********* Unexpected exception detected - ' + DateTimeToStr(Now) + ' *********');
      Writeln(LogFile, VersionInfo);
      Writeln(LogFile, 'Exception code: ' + E.ClassName);
      if IO_OK then
        Writeln(LogFile, 'Exception address: ' + TextualDebugInfoForAddress(Cardinal(ExceptAddr)));
      Writeln(LogFile, #13#10'Call stack (oldest call at bottom):');
      if IO_OK then
        Writeln(LogFile, CallStackTextualRepresentation(S, ''));
      Writeln(LogFile, '*****************************************************************');
      Writeln(LogFile, '');
      Flush(LogFile);
      if IO_OK then
        Msg := Msg + ' - Logged to ' + LogFileName
      else begin
        Msg := Msg + ' - not logged to ' + LogFileName;
        LogFileName := 'stack.log';
      end;
    end else
      Msg := Msg + ' - log not saved to ' + GetCurrentDir;
    StackFilled := False;
  end else
    Msg := Msg + ' (No Stack info)';
  {$I+}
  Application.MessageBox(PChar(Msg), PChar(Application.Title), MB_OK + MB_ICONSTOP);
end;

function StackWatch(Enabled: Boolean):Boolean;
begin
  Result := Assigned(StackWatcher);
  if Enabled and not Assigned(StackWatcher) then
    StackWatcher := TStackWatcher.Create
  else if not Enabled and Assigned(StackWatcher) then
    FreeAndNil(StackWatcher);
end;

initialization

finalization
  StackWatch(False);
end.
