unit EditLog;

interface
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

uses
  {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
  {$IFDEF LCL_} lMessages, {$ENDIF}
  {$IFNDEF CLX_} Graphics, Controls, Forms, Dialogs,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, Qt, Types, {$ENDIF}
  Classes, SysUtils, Log, SrchStrs;

type

  TPlaybackEvent = procedure (Sender:TObject; b:Byte) of object;

  TEditLog = class(TLog)
  private
    FRecording: Boolean;
    FRecordLock: Integer; { To prevent double recording if a high level routine
                            calls a lower level one that is also recorded }
    FMacro, FPlayBackStream: TStream;
    FRecord: TStream;
    FOnPlayback: TPlaybackEvent;
    FBookMark: TPoint;
    FLocalKeyMacro: Boolean;
    FTabTranslatedCaretPos: Integer;
    FCaretWidth: Integer;
    {$IFNDEF CLX_}
    procedure WMSetFocus(var Message: {$IFDEF VCL_}TWMSetFocus{$ENDIF}{$IFDEF LCL_}TLMSetFocus{$ENDIF});
      message {$IFDEF VCL_}WM_SETFOCUS{$ENDIF}{$IFDEF LCL_} LM_SETFOCUS {$ENDIF};
    procedure WMKillFocus(var Message: {$IFDEF VCL_}TWMSetFocus{$ENDIF}{$IFDEF LCL_}TLMKillFocus{$ENDIF});
      message {$IFDEF VCL_}WM_KILLFOCUS{$ENDIF}{$IFDEF LCL_} LM_KILLFOCUS {$ENDIF};
    {$ENDIF}
    procedure UpdateCaretPos;
    procedure SetRecording(Recrd: Boolean);
    procedure SetSelText(const Value: String);
    function  GetCaretPos: TPoint;
    procedure SetBookMark(const Value: TPoint);
  protected
    FLastCaretSize: Integer; 
    {$IFDEF CLX_}
    function EventFilter(Sender: QObjectH; Event: QEventH): Boolean; override;
    {$ENDIF}
    procedure SetCaretPos(Pos: TPoint); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure InternalPlayback(b:Byte); virtual;
    procedure InternalRecord(b:Byte);
    procedure RecordBytes(const Buffer; Count: Longint);
    procedure RecordString(const s: string);
    function  NextPlayBackString: string;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddToRecord(b:Byte);
    procedure PlayBackStream(Stream:TStream);
    procedure PlayBack;
    property Recording:Boolean read FRecording write SetRecording;
    property Macro:TStream read FMacro;

    procedure ChangeCaretPos(x, y: Integer; ExtendSelection: Boolean);
    procedure PreviousChar(Selecting: Boolean);
    procedure NextChar(Selecting: Boolean);
    procedure PreviousWord(Selecting: Boolean);
    procedure NextWord(Selecting: Boolean);
    procedure StartOfLine(Selecting: Boolean);
    procedure StartOfText(Selecting: Boolean);
    procedure EndOfLine(Selecting: Boolean);
    procedure EndOfText(Selecting: Boolean);
    procedure PreviousLine(Selecting: Boolean);
    procedure NextLine(Selecting: Boolean);
    function Search(const Target: string; Options:TFindOptions): Boolean; override;
    procedure SelectAll; override;
    procedure FindCaretTarget(Line: Integer; ExtendSelection: Boolean);

    procedure BeginUpdate; override;
    procedure EndUpdate; override;

    procedure InsertNewLine;
    procedure DeletePrevious;
    procedure DeleteNext;
    procedure InsertString(const NewString: string);
    procedure SetSelTextBuf(Buffer: PChar);
    procedure InsertTextBuf(Buffer: PChar);

    function Replace(const Target, Replacement: string; Options:TFindOptions):Integer;
    function IncrementNumberAtCursor(Increment:Integer): Boolean;

    property SelText write SetSelText;

    property  BookMark: TPoint read FBookMark write SetBookMark;
    procedure SetBookMarkHere;
    procedure ExchangeCursorAndBookMark;
    procedure GotoBookMark;
    procedure RemoveTabs;
    function LineWithoutTabs(Line: Integer): string;

  published
    procedure ClearSelection; override;
    procedure CopyToClipboard; override;
    procedure PasteFromClipboard; override;
    function CanPasteFromClipboard: Boolean; override;

    property OnPlayback: TPlaybackEvent read FOnPlayback write FOnPlayback;
    property LocalKeyMacro: Boolean read FLocalKeyMacro write FLocalKeyMacro default False;
  end;

var
  EditLogMacro: TStream;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

uses
    {$IFNDEF CLX_} ClipBrd, {$ELSE} QClipBrd, QCaret, {$ENDIF}
    {$IFDEF LCL_} LCLType, LCLIntf, {$ENDIF}
    Huge, UCompat, WCompat;

{$IFDEF SINGLE}
{$IFDEF VER80}
{$R EditLog.d16}
{$ELSE}
{$R EditLog.d32}
{$ENDIF}

procedure Register;
begin
  RegisterComponents('CMW', [TEditLog]);
end;

{$ENDIF}

{$DEFINE KEY_COMPAT} { Define this to avoid use of keybd_event calls - must be used under linux }
{$IFDEF UNIX}
{$IFNDEF KEY_COMPAT}
{$DEFINE KEY_COMPAT}
{$ENDIF}
{$ENDIF}

const
  ExternalMark:TShiftState = [ssShift, ssAlt, ssCtrl, ssLeft, ssRight, ssMiddle, ssDouble];
  InternalMark:TShiftState = [ssShift, ssAlt, ssCtrl, ssLeft, ssRight, ssDouble];
  {$IFDEF KEY_COMPAT}
  KeyMark:     TShiftState = [ssShift, ssAlt, ssCtrl, ssRight, ssMiddle, ssDouble];
  {$ENDIF}

type
  PlayBackAction = (pbIncrement, pbSearch, pbReplace, pbCopy, pbClearSelection, pbPaste,
                    pbDeletePrevious, pbDeleteNext, pbSelectAll,
                    pbSetBookMarkHere, pbExchangeCursorAndBookMark, pbGotoBookMark, pbRemoveTabs);

constructor TEditLog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  {$IFDEF LCL_}
  FCompStyle:= csFixed; { Force control type that can get the focus, and have a caret }
  {$ENDIF}
  FCaretWidth :=
  {$IFDEF MSWINDOWS}
    GetSystemMetrics(SM_CXBORDER);
  {$ELSE}
    1;
  {$ENDIF}
  Lines.SetPointVar := {$IFDEF FPC}@{$ENDIF}SetCaretPos;
  Lines.GetPointVar := {$IFDEF FPC}@{$ENDIF}GetCaretPos;
  AutoScroll := False;
end;

destructor TEditLog.Destroy;
begin
  {$IFDEF CLX_}
  DestroyCaret(Canvas);
  {$ENDIF}
  FMacro.Free;
  FMacro := nil;
  FRecord.Free;
  FRecord := nil;
  inherited Destroy;
end;


procedure TEditLog.InternalPlayback(b:Byte);
var
  i: Integer;
  s1, s2: string;
  Options: TFindOptions;
begin
  case PlayBackAction(b) of
    pbIncrement: begin
                   FPlayBackStream.Read(i, SizeOf(i));
                   IncrementNumberAtCursor(i);
                 end;
    pbSearch:   begin
                   FPlayBackStream.Read(Options, SizeOf(Options));
                   s1 := NextPlayBackString;
                   Search(s1, Options);
                 end;
    pbReplace:   begin
                   FPlayBackStream.Read(Options, SizeOf(Options));
                   s1 := NextPlayBackString;
                   s2 := NextPlayBackString;
                   Replace(s1, s2, Options);
                 end;
    pbCopy:                      CopyToClipBoard;
    pbClearSelection:            ClearSelection;
    pbPaste:                     PasteFromClipBoard;
    pbDeletePrevious:            DeletePrevious;
    pbDeleteNext:                DeleteNext;
    pbSelectAll:                 SelectAll;
    pbSetBookMarkHere:           SetBookMarkHere;
    pbExchangeCursorAndBookMark: ExchangeCursorAndBookMark;
    pbGotoBookMark:              GotoBookMark;
  end;
end;


procedure TEditLog.SetRecording(Recrd: Boolean);
begin
  if Recrd <> Recording then begin
    if Assigned(FPlayBackStream) and Recrd then
      raise Exception.Create('Can''t record while playing back');
    FRecording := Recrd;
    if Recording then begin
      FRecord.Free;
      FRecord := TMemoryStream.Create;
    end else begin
      if LocalKeyMacro then begin
        FMacro.Free;
        FMacro := FRecord;
      end else begin
        EditLogMacro.Free;
        EditLogMacro := FRecord;
      end;
      FRecord := nil;
    end;
  end;
  FRecordLock := 0;
end;

{$IFNDEF KEY_COMPAT}
type
  RecordItem = Word;

const
  ShiftCount = 8;
  ShiftMask = $ff;

{$IFDEF VER80}
procedure Keybd_Event; far; external 'USER';
{$ENDIF}

procedure KeyBd(Key: integer; Down: Boolean);
var
{$ifdef WIN32}
  Scan: Byte;
  Flg: Dword;
{$else}
  Scan: Integer;
  Flg: Integer;
{$endif}
begin
  Scan := MapVirtualKey(Key,0);  { find VK }
{$IFDEF WIN32}
  if Down then
    Flg := 0
  else
    Flg := KEYEVENTF_KEYUP;
  keybd_event(Key, Scan, Flg, 0);
{$ELSE}
  if not Down then Key := Integer(Key or $8000);
  asm
    push bx
    mov bx, Scan
    mov ax, Key
  end;
  Keybd_Event;
  asm
    pop bx
  end;
{$ENDIF}
end;

{$IFDEF UNIX}
{ Not used - use different architecture for Linux }
function GetKeyState(Key:Word):Integer;
var
  Down: Boolean;
begin
  case Key of
    VK_SHIFT: Down := ssShift in Application.KeyState;
    VK_CONTROL: Down := ssCtrl in Application.KeyState;
    VK_MENU: Down := ssAlt in Application.KeyState;
  else
    raise Exception.Create('GetKeyState?');
  end;
  if Down then
    Result := -1
  else
    Result := 0;
end;
{$ENDIF}

procedure TEditLog.PlayBackStream(Stream:TStream);
var
  Key: Word;
  Shift: TShiftState;
  Sh, Al, Ct: Boolean;
  ss: Byte;
  LastPlayBackStream: TStream;
  procedure SetShiftKey(Key: Word; ToBeSet:Boolean);
  begin
    if (GetKeyState(Key) < 0) <> ToBeSet then begin
      KeyBd(Key, ToBeSet);
      Application.ProcessMessages;
    end;
  end;
begin
  if (Stream <> nil) and (Stream <> FPlayBackStream) then begin
    LastPlayBackStream := FPlayBackStream;
    try
     FPlayBackStream := Stream;
     Stream.Seek(0,0);
     Sh := GetKeyState(VK_SHIFT) < 0;
     Al := GetKeyState(VK_MENU) < 0;
     Ct := GetKeyState(VK_CONTROL) < 0;
     while Stream.Read(Key,2) = 2 do begin
       SetFocus;
       Application.ProcessMessages;
       ss := (Key shr ShiftCount) and ShiftMask;
       Shift := TShiftState(ss);
       Key := Key and ShiftMask;
       if Shift = InternalMark then
         InternalPlayback(Key)
       else if Shift = ExternalMark then begin
         if Assigned(FOnPlayback) then OnPlayBack(Self, Key);
       end else begin
         SetShiftKey(VK_SHIFT, ssShift in Shift);
         SetShiftKey(VK_CONTROL, ssCtrl in Shift);
         SetShiftKey(VK_MENU, ssAlt in Shift);
         KeyBd(Key, True);
         Application.ProcessMessages;
         KeyBd(Key, False);
       end;
       Application.ProcessMessages;
     end;
     SetShiftKey(VK_SHIFT, Sh);
     SetShiftKey(VK_CONTROL, Ct);
     SetShiftKey(VK_MENU, Al);
    finally
     FPlayBackStream := LastPlayBackStream;
    end;
  end;
end;
{$ENDIF}

{$IFDEF KEY_COMPAT}
type
  RecordItem = Integer;

const
  ShiftCount = 24;   { TODO: May need more bits for shift state for LCL_ }
  ShiftMask = $ffff;

procedure TEditLog.PlayBackStream(Stream:TStream);
var
  Key: Word;
  Item: RecordItem;
  Shift: TShiftState;
  ss: {$IFNDEF LCL_} Byte {$ELSE} Integer {$ENDIF};
  LastPlayBackStream: TStream;
  ch: Char;
begin
  if (Stream <> nil) and (Stream <> FPlayBackStream) then begin
    LastPlayBackStream := FPlayBackStream;
    try
     FPlayBackStream := Stream;
     Stream.Seek(0,0);
     while Stream.Read(Item,4) = 4 do begin
       ss := (Item shr ShiftCount) and $ff;
       Shift := TShiftState(ss);
       Key := Item and ShiftMask;
       if Shift = InternalMark then
         InternalPlayback(Byte(Key))
       else if Shift = ExternalMark then begin
         if Assigned(FOnPlayback) then OnPlayBack(Self, Byte(Key));
       end else if Shift = KeyMark then begin
         ch := Char(Key);
         KeyPress(ch);
       end else
         KeyDown(Key, Shift);
     end;
    finally
     FPlayBackStream := LastPlayBackStream;
    end;
  end;
end;
{$ENDIF}

procedure TEditLog.AddToRecord(b:Byte);
var
  k: RecordItem;
begin
  if Recording and (FRecordLock = 0) then begin
    k := ({$IFNDEF LCL_}Byte{$ELSE}Integer{$ENDIF}(ExternalMark) shl ShiftCount) or b;
    FRecord.Write(k, SizeOf(k));
  end;
end;

procedure TEditLog.InternalRecord(b:Byte);
var
  k: RecordItem;
begin
  if Recording and (FRecordLock = 0) then begin
    k := ({$IFNDEF LCL_}Byte{$ELSE}Integer{$ENDIF}(InternalMark) shl ShiftCount) or b;
    FRecord.Write(k, SizeOf(k));
  end;
end;

procedure TEditLog.RecordBytes(const Buffer; Count: Longint);
begin
  if Recording and (FRecordLock = 0) then begin
    FRecord.Write(Buffer, Count);
  end;
end;

procedure TEditLog.RecordString(const s: string);
var
  {$IFDEF VER80}
  l: Byte;
  {$ELSE}
  l: Integer;
  {$ENDIF}
begin
  l := Length(s);
  RecordBytes(l, SizeOf(l));
  if l > 0 then
    RecordBytes(s[1], l);
end;

function TEditLog.NextPlayBackString: string;
var
  {$IFDEF VER80}
  l: Byte;
  {$ELSE}
  l: Integer;
  {$ENDIF}
begin
  FPlayBackStream.Read(l, SizeOf(l));
  SetLength(Result, l);
  if l > 0 then
    FPlayBackStream.Read(Result[1], l);
end;

procedure TEditLog.PlayBack;
begin
  if LocalKeyMacro then
    PlayBackStream(FMacro)
  else
    PlayBackStream(EditLogMacro)
end;

{$IFNDEF CLX_}
procedure TEditLog.WMSetFocus(var Message: {$IFDEF VCL_}TWMSetFocus{$ENDIF}{$IFDEF LCL_}TLMSetFocus{$ENDIF});
begin
  inherited {$IFDEF FPC} WMSetFocus(Message) {$ENDIF};
  FLastCaretSize := CharSize.y;
  CreateCaret(Handle, HBitmap(0), FCaretWidth*2, CharSize.y);
  UpdateCaretPos;
  ShowCaret(Handle);
end;

procedure TEditLog.WMKillFocus(var Message: {$IFDEF VCL_}TWMSetFocus{$ENDIF}{$IFDEF LCL_}TLMKillFocus{$ENDIF});
begin
  inherited {$IFDEF FPC} WMKillFocus(Message) {$ENDIF};
  DestroyCaret{$IFDEF LCL_}(Handle){$ENDIF};
end;
{$ELSE}
function TEditLog.EventFilter(Sender: QObjectH; Event: QEventH): Boolean;
begin
  Result := inherited EventFilter(Sender, Event);
  case QEvent_type(Event) of
    QEventType_FocusIn:
      begin
        CreateCaret(Canvas, nil, FCaretWidth*2, CharSize.y);
        UpdateCaretPos;
        ShowCaret;
      end;
    QEventType_FocusOut: DestroyCaret;
  end;
end;
{$ENDIF}

procedure TEditLog.UpdateCaretPos;
begin
  if Focused then begin
    FTabTranslatedCaretPos := TabCharPos(CaretPos.x, CaretPos.y);
    {$IFNDEF CLX_}
    if FLastCaretSize <> CharSize.y then begin
      CreateCaret(Handle, HBitmap(0), FCaretWidth*2, CharSize.y);
      FLastCaretSize := CharSize.y; { TODO : check font change ok }
    end;
    {$ELSE}
    ChangeCaretHeight(CharSize.y);
    {$ENDIF}
    {$IFDEF LCL_} LCLIntf. {$ENDIF}
    {$IFDEF VCL_} Windows. {$ENDIF}
    {$IFDEF CLX_} QCaret. {$ENDIF}
    SetCaretPos(PaintRect.Left + (FTabTranslatedCaretPos-HorzScrollBar.Position)*CharSize.x-FCaretWidth,
                PaintRect.Top + (CaretPos.y-VertScrollBar.Position)*CharSize.y);
  end;
end;

procedure TEditLog.SetCaretPos(Pos: TPoint);
begin
  inherited SetCaretPos(Pos);
  if not Lines.Updating then begin
    MakeVisible(CaretPos);
    UpdateCaretPos;
  end;
end;

function  TEditLog.GetCaretPos: TPoint;
begin
  Result := CaretPos;
end;

procedure TEditLog.BeginUpdate;
begin
  Inc(FRecordLock);
  inherited BeginUpdate;
end;

procedure TEditLog.EndUpdate;
begin
  inherited EndUpdate;
  if not Lines.Updating then
    SetCaretPos(CaretPos);
  Dec(FRecordLock);
end;

procedure TEditLog.DeletePrevious;
var
  s: string;
begin
  InternalRecord(Byte(pbDeletePrevious));
  try
    BeginUpdate;
    if AreaSelected then
      ClearSelection
    else if CaretPos.y < Lines.Count then begin
      if CaretPos.x > 0 then begin
        s := Lines[CaretPos.y];
        if CaretPos.x >= Length(s) then begin
          Lines[CaretPos.y] := Copy(s, 1, Length(s)-1);
          CaretPos := Point(Length(s)-1, CaretPos.y);
        end else begin
          Lines[CaretPos.y] := Concat(Copy(s, 1, CaretPos.x-1), Copy(s, CaretPos.x+1, Length(s)));
          CaretPos := Point(CaretPos.x-1, CaretPos.y);
        end;
      end else if CaretPos.y > 0 then begin
        s := Lines[CaretPos.y-1];
        Lines[CaretPos.y-1] := Concat(s, Lines[CaretPos.y]);
        Lines.Delete(CaretPos.y);
        CaretPos := Point(Length(s), CaretPos.y-1);
      end;
    end else
      PreviousChar(False);
    CaretTargetIsHere;
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.DeleteNext;
begin
  InternalRecord(Byte(pbDeleteNext));
  try
    BeginUpdate;
    if AreaSelected then
      ClearSelection
    else if not ( (CaretPos.y >= Lines.Count)
                  or ( (CaretPos.y = Lines.Count-1)
                       and (CaretPos.x >= Length(Lines[CaretPos.y])) ) ) then begin
      NextChar(False);
      DeletePrevious;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.InsertNewLine;
var
  s: string;
begin
  try
    BeginUpdate;
    if AreaSelected then
      ClearSelection;
    if CaretPos.y >= Lines.Count then
      Lines.Add('')
    else begin
      s := Lines[CaretPos.y];
      if CaretPos.x >= Length(s) then
        Lines.Insert(CaretPos.y+1, '')
      else begin
        Lines[CaretPos.y] := Copy(s, 1, CaretPos.x);
        Lines.Insert(CaretPos.y+1, Copy(s, CaretPos.x+1, Length(s)));
      end;
    end;
    ChangeCaretPos(0, CaretPos.y+1, False);
    CaretTargetIsHere;
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.InsertString(const NewString: string);
var
  s: string;
begin
  try
    BeginUpdate;
    if AreaSelected then
      ClearSelection;
    if CaretPos.y >= Lines.Count then
      Lines.Add(NewString)
    else begin
      s := Lines[CaretPos.y];
{$IFDEF VER80}
      if Length(s) + Length(NewString) > 255 then
        raise Exception.Create('Line longer than 255 characters');
{$ENDIF}
      if CaretPos.x >= Length(s) then
        Lines[CaretPos.y] := Concat(s, NewString)
      else
        Lines[CaretPos.y] := Concat(Copy(s, 1, CaretPos.x), NewString, Copy(s, CaretPos.x+1, Length(s)));
    end;
    ChangeCaretPos(CaretPos.x + Length(NewString), CaretPos.y, False);
    CaretTargetIsHere;
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.SetSelTextBuf(Buffer: PChar);
begin
  try
    BeginUpdate;
    if AreaSelected then
      ClearSelection;
    InsertTextBuf(Buffer);
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.InsertTextBuf(Buffer: PChar);
var
  p, pStart: PChar;
  Line: string;
  n: Integer;
  c: Char;
  procedure ResetLine;
  begin
    pStart := p;
    SetLength(Line, 0);
    n := 0;
  end;
  procedure AddLine;
  var
    s: string;
  begin
    if n > 0 then begin
      SetLength(s, n);
      hmemcpy(@s[1], pStart, n);
      InsertString(s);
    end;
  end;
begin
  try
    BeginUpdate;
    p := Buffer;
    ResetLine;
    while Ord(p^) <> 0 do begin
      case p^ of
        #13, #10: begin
               c := p^;
               AddLine;
               {$IFNDEF VER80}
                 Inc(p);
               {$ELSE}
                 HugeInc(Pointer(p), 1);
               {$ENDIF}
               if (p^ <> c) and ((p^ = #10) or (p^ = #13)) then
                 {$IFNDEF VER80}
                   Inc(p);
                 {$ELSE}
                   HugeInc(Pointer(p), 1);
                 {$ENDIF}
               InsertNewLine;
               ResetLine;
             end;
      else
        {$IFDEF VER80}
        if n = 255 then begin
          AddLine;
          InsertNewLine;
          ResetLine;
        end;
        {$ENDIF}
        Inc(n);
        {$IFNDEF VER80}
          Inc(p);
        {$ELSE}
          HugeInc(Pointer(p), 1);
        {$ENDIF}
      end;
    end;
    if n > 0 then
      AddLine;
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.SelectAll;
begin
  InternalRecord(Byte(pbSelectAll));
  inherited SelectAll;
end;

procedure TEditLog.CopyToClipboard;
begin
  InternalRecord(Byte(pbCopy));
  inherited CopyToClipboard;
end;

procedure TEditLog.ClearSelection;
begin
  InternalRecord(Byte(pbClearSelection));
  inherited ClearSelection;
end;

procedure TEditLog.PasteFromClipboard;
var
  GHandle: TGHandle;
  // i, j: Integer;
  // buff: string[255];
begin
  InternalRecord(Byte(pbPaste));
  GHandle := nil;
  if ClipboardHasText then
    try
      BeginUpdate;
    { for i := 0 to Clipboard.FormatCount-1 do begin
        j := GetClipboardFormatName(Clipboard.Formats[i], @buff[1], 250);
        if j = 0 then
          case Clipboard.Formats[i] of
            1: buff := 'CF_TEXT';
            7: buff := 'CF_OEMTEXT';
           13: buff := 'CF_UNICODETEXT';
           16: buff := 'CF_LOCALE';
          else
            buff := '()';
          end
        else
          SetLength(buff, j);
        InsertString(Format(' %d=%s ', [Clipboard.Formats[i],buff]));
        GHandle := TGHandle.Create;
        GHandle.SetFromClipboard(Clipboard.Formats[i]);
        InsertString(Format('%d',[GHandle.Size]));
        for j := 0 To GHandle.Size do
          InsertString(Format('%4d',[Ord((PChar(GHandle.Base)+j)^)]));
        InsertNewline;
        GHandle.Free;
      end;
      InsertNewLine; }
      GHandle := TGHandle.Create;
      GHandle.SetFromClipboard(CF_TEXT);
      SetSelTextBuf(GHandle.Item[0]);
    { InsertNewLine;
      InsertString(Format('%d',[GHandle.Size]));
      for i := 0 To GHandle.Size do
        InsertString(Format('%4d',[Ord((PChar(GHandle.Base)+i)^)])); }
    finally
      EndUpdate;
      GHandle.Free;
    end;
end;

procedure TEditLog.SetSelText(const Value: string);
{$IFDEF VER80}
var
  Buffer: array[0..255] of Char;
{$ENDIF}
begin
{$IFDEF VER80}
  SetSelTextBuf(StrPCopy(Buffer, Value));
{$ELSE}
  SetSelTextBuf(PChar(Value));
{$ENDIF}
end;

procedure TEditLog.KeyPress(var Key: Char);
{$IFDEF KEY_COMPAT}
var
  k: RecordItem;
{$ENDIF}
begin
  inherited KeyPress(Key);
  case Key of
    Char(Ord('B') and 31): SetBookMarkHere;
    Char(Ord('Q') and 31): GotoBookMark;
  else
    if (Ord(Key) >= 32) and (Ord(Key) <> 127) then begin
      {$IFDEF KEY_COMPAT}
      if Recording then begin
        k := ({$IFNDEF LCL_}Byte{$ELSE}Integer{$ENDIF}(KeyMark) shl ShiftCount ) or Ord(Key);
        FRecord.Write(k, Sizeof(k));
      end;
      {$ENDIF}
      InsertString(Key);
    end else
      Exit;
  end;
  Key := #0;
end;

procedure TEditLog.ChangeCaretPos(x, y: Integer; ExtendSelection: Boolean);
begin
  if ExtendSelection then
    ExtendSelectionTo(Point(x,y), True)
  else begin
   CaretPos := Point(x, y);
   Anchor := CaretPos;
   AreaSelected := False;
  end;
end;

procedure TEditLog.PreviousChar(Selecting: Boolean);
begin
  if (CaretPos.x = 0) or (CaretPos.y >= Lines.Count) then
    ChangeCaretPos(MaxInt, CaretPos.y-1, Selecting)
  else
    ChangeCaretPos(CaretPos.x-1, CaretPos.y, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.PreviousWord(Selecting: Boolean);
var
  p: TPoint;
  Line: string;
begin
  p := CaretPos;
  if (p.x = 0) or (p.y >= Lines.Count) then
    PreviousChar(Selecting)
  else if (p.y >= 0) then begin
    Line := Lines[p.y];
    while p.x > 0 do begin
      Dec(p.x);
      if IsInWord(Line[p.x+1]) then
        Break;
    end;
    while (p.x > 0) and IsInWord(Line[p.x]) do
      Dec(p.x);
    ChangeCaretPos(p.x, p.y, Selecting);
    CaretTargetIsHere;
  end;
end;

procedure TEditLog.NextChar(Selecting: Boolean);
var
  l: Integer;
begin
  if (CaretPos.y >= 0) and (CaretPos.y < Lines.Count) then
    l := Length(Lines[CaretPos.y])
  else
    l := 0;
  if CaretPos.x >= l then
    ChangeCaretPos(0, CaretPos.y+1, Selecting)
  else
    ChangeCaretPos(CaretPos.x+1, CaretPos.y, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.NextWord(Selecting: Boolean);
var
  p: TPoint;
  Line: string;
  l: Integer;
begin
  p := CaretPos;
  if (p.y >= 0) and (p.y < Lines.Count) then
    Line := Lines[p.y]
  else
    Line := '';
  l := Length(Line);
  if (p.x >= l) then begin
    p.x := 0;
    Inc(p.y);
  end else begin
    { Next Word }
    while (p.x < l) and IsInWord(Line[p.x+1]) do
      Inc(p.x);
    while (p.x < l) and not IsInWord(Line[p.x+1]) do
      Inc(p.x);
  end;
  ChangeCaretPos(p.x, p.y, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.StartOfLine(Selecting: Boolean);
begin
  ChangeCaretPos(0, CaretPos.y, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.StartOfText(Selecting: Boolean);
begin
  ChangeCaretPos(0, 0, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.EndOfLine(Selecting: Boolean);
begin
  ChangeCaretPos(MaxInt, CaretPos.y, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.EndOfText(Selecting: Boolean);
begin
  ChangeCaretPos(MaxInt, Lines.Count-1, Selecting);
  CaretTargetIsHere;
end;

procedure TEditLog.PreviousLine(Selecting: Boolean);
begin
  FindCaretTarget(CaretPos.y-1, Selecting);
end;

procedure TEditLog.NextLine(Selecting: Boolean);
begin
  FindCaretTarget(CaretPos.y+1, Selecting);
end;

procedure TEditLog.KeyDown(var Key: Word; Shift: TShiftState);
var
  k: RecordItem;
  FOnKeyDown: TKeyEvent;
  NormalKey: Boolean;
begin
  {$IFDEF MSWINDOWS}
  {$IFNDEF KEY_COMPAT}
  if Recording and not (ssAlt in Shift) then
    case Key of
      VK_LBUTTON, VK_RBUTTON, VK_MBUTTON, VK_CANCEL,
      VK_SHIFT, VK_CONTROL, VK_MENU, VK_PAUSE:
      { MAYDO: This does not handle MBCS input; would need to handle/record KeyString();
               This also requires QWidget::setKeyCompression(False) (the default) }
    else
      k := (Byte(Shift) shl ShiftCount ) or Key;
      FRecord.Write(k, Sizeof(k));
    end;
  {$ENDIF}
  {$ENDIF}
  { Note - do not call inherited - use this instead }
  FOnKeyDown := OnKeyDown;
  if Assigned(FOnKeyDown) then FOnKeyDown(Self, Key, Shift);
  { Temporarily disable recording, so action not recorded twice }
  Inc(FRecordLock);
  NormalKey := False;
  case Key of
    VK_HOME:  if ssCtrl in Shift then
                StartOfText(ssShift in Shift)
              else
                StartOfLine(ssShift in Shift);
    VK_END:   if ssCtrl in Shift then
                EndOfText(ssShift in Shift)
              else
                EndOfLine(ssShift in Shift);
    VK_PRIOR: if ssCtrl in Shift then
                FindCaretTarget(VertScrollBar.Position, ssShift in Shift)
              else
                FindCaretTarget(CaretPos.y-VertScrollBar.PageSize+2, ssShift in Shift);
    VK_NEXT:  if ssCtrl in Shift then
                FindCaretTarget(VertScrollBar.Position+VertScrollBar.PageSize-1, ssShift in Shift)
              else
                FindCaretTarget(CaretPos.y+VertScrollBar.PageSize-2, ssShift in Shift);
    VK_UP:    if ssCtrl in Shift then
                VertScrollBar.Position := VertScrollBar.Position - 1
              else
                PreviousLine(ssShift in Shift);
    VK_DOWN:  if ssCtrl in Shift then
                VertScrollBar.Position := VertScrollBar.Position + 1
              else
                NextLine(ssShift in Shift);
    VK_LEFT:  if ssCtrl in Shift then
                PreviousWord(ssShift in Shift)
              else
                PreviousChar(ssShift in Shift);
    VK_RIGHT: if ssCtrl in Shift then
                NextWord(ssShift in Shift)
              else
                NextChar(ssShift in Shift);
    VK_RETURN:if ssAlt in Shift then
                Redo
              else
                InsertNewline;
    VK_BACK:  if ssAlt in Shift then
                Undo
              else
                DeletePrevious;
    VK_DELETE:if ssShift in Shift then
                CutToClipboard
              else
                DeleteNext;
    VK_INSERT:if ssCtrl in Shift then
                CopyToClipboard
              else if ssShift in Shift then
                PasteFromClipboard;
    VK_TAB:   if ssShift in Shift then begin
                ChangeCaretPos(CaretPos.x - (CaretPos.x - 1) mod TabSpacing + 1, CaretPos.y, False);
                CaretTargetIsHere;
              end else
                InsertString(Char(9));
    Ord('B'): if Shift = [ssShift, ssCtrl] then
                ExchangeCursorAndBookMark
              else
                NormalKey := True;
    Ord('Z'): if Shift = [ssCtrl, ssShift] then
                Redo
              else
                Exit;
    VK_F11:   if Shift = [] then
                IncrementNumberAtCursor(1)
              else if Shift = [ssShift] then
                IncrementNumberAtCursor(-1)
              else
                NormalKey := True;
    VK_F12:   if Shift = [] then
                PlayBack
              else if Shift = [ssShift] then begin
                Recording := not Recording;
                Key := 0;
                Exit;
              end else
                NormalKey := True;
  else
    NormalKey := True;
  end;
  Dec(FRecordLock);
  if not NormalKey then begin
    {$IFDEF KEY_COMPAT}
    if Recording then begin
      k := ({$IFNDEF LCL_}Byte{$ELSE}Integer{$ENDIF}(Shift) shl ShiftCount ) or Key;
      FRecord.Write(k, Sizeof(k));
    end;
    {$ENDIF}
    Key := 0; { Indicate processing }
  end;
end;

function TEditLog.Search(const Target: string; Options:TFindOptions): Boolean;
begin
  InternalRecord(Byte(pbSearch));
  RecordBytes(Options, SizeOf(Options));
  RecordString(Target);
  Result := inherited Search(Target, Options);
end;

function TEditLog.Replace(const Target, Replacement: string; Options:TFindOptions): Integer;
var
  Found: Boolean;
  p1: TPoint;
begin
  Result := 0;
  if Length(Target) = 0 then Exit;
  InternalRecord(Byte(pbReplace));
  RecordBytes(Options, SizeOf(Options));
  RecordString(Target);
  RecordString(Replacement);
  Found := AnsiCompareText(SelText, Target) = 0;
  if not Found then
    Found := Search(Target, Options);
  if (not Found) and (frReplace in Options) then
    ShowMessage('Cannot find "' + Target + '".')
  else
    try
      BeginUpdate;
      while Found do begin
        p1 := SelRect.TopLeft;
        SelText := Replacement;
        SelRect := Rect(p1.x, p1.y, CaretPos.x, CaretPos.y);
        if not (frDown in Options) then
          CaretPos := p1;
        CaretTargetIsHere;
        Inc(Result);
        if not (frReplaceAll in Options) then Break;
        Found := Search(Target, Options);
      end;
    finally
      EndUpdate;
    end;
end;

function TEditLog.IncrementNumberAtCursor(Increment:Integer): Boolean;
var
  s, e: Integer;
  n: Longint;
  line, newn: string;
  function IsDigit(c: Char): Boolean;
  begin
    Result := (c >= '0') and (c <= '9');
  end;
  function IsSpace(c: Char): Boolean;
  begin
    Result := (c = ' ') or (c = #9);
  end;
begin
  InternalRecord(Byte(pbIncrement));
  RecordBytes(Increment, SizeOf(Increment));
  Result := False;
  try
    BeginUpdate;
    if (CaretPos.y < 0) or (CaretPos.y >= Lines.Count) then Exit;
    s := CaretPos.x + 1;
    line := Lines[CaretPos.y];
    e := s;
    while (s > 1) and IsDigit(line[s-1]) do Dec(s);
    Result := s < e;
    if (s > 1) and (line[s-1] = '-') then Dec(s);
    if (e <= Length(line)) and (s = e) and (line[e] = '-') then Inc(e);
    while (e <= Length(line)) and IsDigit(line[e]) do begin
      Inc(e);
      Result := True;
    end;
    if Result then
      try
        n := StrToInt(Copy(line, s, e-s))+Increment;
        newn := IntToStr(n);
        if (Length(newn) > e-s) and (s > 1) and IsSpace(line[s-1]) then
          Dec(s)
        else if Length(newn) < e-s then
          newn := ' ' + newn;
        Lines[CaretPos.y] := Concat(Copy(line, 1, s-1), newn, Copy(line, e, Length(line)));
        CaretPos := Point(s + Length(newn) - 1, CaretPos.y);
        CaretTargetIsHere;
      except
        Result := False;
      end;
  finally
    EndUpdate;
  end;
end;

procedure TEditLog.SetBookMark(const Value: TPoint);
begin
  FBookMark := Value;
end;

procedure TEditLog.ExchangeCursorAndBookMark;
var
  t: Tpoint;
begin
  InternalRecord(Byte(pbExchangeCursorAndBookMark));
  t := CaretPos;
  CaretPos := BookMark;
  BookMark := t;
end;

procedure TEditLog.GotoBookMark;
begin
  InternalRecord(Byte(pbGotoBookMark));
  CaretPos := BookMark;
end;

procedure TEditLog.SetBookMarkHere;
begin
  InternalRecord(Byte(pbSetBookMarkHere));
  BookMark := CaretPos;
end;

function TEditLog.CanPasteFromClipboard: Boolean;
begin
  Result := ClipboardHasText or inherited CanPasteFromClipboard;
end;

procedure TEditLog.FindCaretTarget(Line: Integer; ExtendSelection: Boolean);
var
  NewPos: TPoint;
begin
  NewPos.x := CaretTarget;
  NewPos.y := Line;
  MakeSane(NewPos);
  if NewPos.y <> CaretPos.y then begin
    NewPos.x := UnTabCharPos(CaretTarget, NewPos.y);
    ChangeCaretPos(NewPos.x, NewPos.y, ExtendSelection);
  end;
end;

procedure TEditLog.RemoveTabs;
var
  P, EndPoint, OrgCursor: TPoint;
  s, e, cp, tp, nExtra: Integer;
  L: string;
  FoundTab: Boolean;
begin
  InternalRecord(Byte(pbRemoveTabs));
  try
    BeginUpdate;
    if AreaSelected then begin
      P := SelRect.TopLeft;
      EndPoint := SelRect.BottomRight;
    end else begin
      P := Point(0, 0);
      EndPoint := Point(0, Lines.Count);
    end;
    OrgCursor := CaretPos;
    OrgCursor.x := TabCharPos(OrgCursor.x, OrgCursor.y);
    while P.y <= EndPoint.y  do begin
      s := P.x;
      if P.y = EndPoint.y then
        e := EndPoint.x
      else
        e := Length(Lines[p.y]);
      if e > s then begin
        L := Lines[P.y];
        cp := Pos(#9, L);
        tp := cp; { Logical position - must be same as character number before first tab }
        FoundTab := False;
        while (cp > 0) and (cp <= e) do begin
          nExtra := TabSpacing - (tp-1) mod TabSpacing - 1;
          if cp > s then begin
            L := Concat(Copy(L, 1, cp-1), StringOfChar(' ', nExtra+1), Copy(L, cp+1, e-cp+1));
            Inc(cp, nExtra);
            Inc(e, nExtra);
            FoundTab := True;
          end;
          Inc(tp, nExtra);
          { Move to next tab }
          repeat
            Inc(cp);
            Inc(tp);
          until (cp > e) or (L[cp] = #9);
        end;
        if FoundTab then Lines[P.y] := L;
      end;
      Inc(P.y);
      P.x := 0;
    end;
    CaretPos := Point(UnTabCharPos(OrgCursor.x,OrgCursor.y), OrgCursor.y);
  finally
    EndUpdate;
  end;
end;

function TEditLog.LineWithoutTabs(Line: Integer): string;
var
  cp, nExtra: Integer;
begin
  Result := Lines[Line];
  cp := Pos(#9, Result);
  while cp > 0 do begin
    nExtra := TabSpacing - (cp-1) mod TabSpacing - 1;
    Result := Concat(Copy(Result, 1, cp-1), StringOfChar(' ', nExtra+1), Copy(Result, cp+1, Length(Result)-cp));
    cp := Pos(#9, Result);
  end;
end;

procedure TEditLog.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button =  mbMiddle then begin
    CaretPos := ConvertMouse(X, Y);
    AreaSelected := False;
    {$IFDEF UNIX}
    SetSelText(PrimarySelection.AsText);
    {$ELSE}
    PasteFromClipboard;
    {$ENDIF}
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

end.
